(***********************************************************************)
(*                                                                     *)
(*                         Fast Escape Analysis                        *)
(*                                                                     *)
(*                  Bruno Blanchet, ENS Ulm - INRIA Rocquencourt       *)
(*                  Alain Deutsch, INRIA Rocquencourt                  *)
(*                                                                     *)
(*  Copyright 1995 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

open Types
open Typedtree
open Format
open Asttypes
open Fsolver
open Fenv
open Fenvir
open Ftrtype

let in_let_rec = ref false

(* Analysis of patterns *)

let enrich_env_with_pattern boundinfo id_info env pattern =
(* var_list -> (bool * Store.op list) 
            -> env -> Typedtree.pattern -> env *) 
   let rec enrich env pattern = match pattern.pat_desc with
               Tpat_any -> env
            |  Tpat_var(ident) -> 
                  let ty = TranslType.expr env pattern.pat_type in
                  if NewType.functional ty then
                     let info = IdentInfo.build ty id_info (!in_let_rec) in
                     boundinfo := info :: !boundinfo;
                     Envir.update ident info env
                  else env
            |  Tpat_alias(pat,ident) ->
                  let ty = TranslType.expr env pattern.pat_type in
                  if NewType.functional ty then
                     let info = IdentInfo.build ty id_info (!in_let_rec) in
                     boundinfo := info :: !boundinfo;
                     Envir.update ident info (enrich env pat)
                  else enrich env pat
            |  Tpat_constant _ -> env
            |  Tpat_tuple patlist -> List.fold_left enrich env patlist
            |  Tpat_construct(constr,patlist) -> List.fold_left enrich
                 env patlist
            |  Tpat_record recpatlist -> List.fold_left 
                  (fun env -> function (_,pat) -> enrich env pat) env recpatlist
            |  Tpat_or (pat1,pat2) -> env (* An or pattern cannot declare any variable *)
   in enrich env pattern

let rec enrich_env_useless env pattern =
    match pattern.pat_desc with
       Tpat_any -> env
    |  Tpat_var(ident) -> 
         let ty = TranslType.expr env pattern.pat_type in
         if NewType.functional ty then
            Envir.update ident (IdentInfo.empty ty (!in_let_rec)) env
         else env
    |  Tpat_alias(pat,ident) ->
         let ty = TranslType.expr env pattern.pat_type in
         if NewType.functional ty then
            Envir.update ident (IdentInfo.empty ty (!in_let_rec))
              (enrich_env_useless env pat)
         else enrich_env_useless env pat
    |  Tpat_constant _ -> env
    |  Tpat_tuple patlist -> 
                  List.fold_left enrich_env_useless env patlist
    |  Tpat_construct(constr,patlist) -> 
                  List.fold_left enrich_env_useless env patlist
    |  Tpat_record recpatlist -> List.fold_left 
                  (fun env -> function (_,pat) -> enrich_env_useless env pat) 
                  env recpatlist
    |  Tpat_or(_,_) -> env (* An or pattern cannot declare any variable *)

let enrich_var_env_with_pattern analpar env pattern =
(* var_env_analres -> var_env -> Typedtree.pattern -> var_env *) 
   let rec enrich env pattern = 
           match pattern.pat_desc with
               Tpat_any -> env
            |  Tpat_var(ident) -> VarEnv.update ident analpar env
            |  Tpat_alias(pat,ident) -> VarEnv.update ident analpar
                 (enrich env pat)
            |  Tpat_constant _ -> env
            |  Tpat_tuple patlist -> List.fold_left enrich env patlist
            |  Tpat_construct(constr,patlist) -> List.fold_left enrich
                 env patlist
            |  Tpat_record recpatlist -> List.fold_left 
                  (fun env -> function (_,pat) -> enrich env pat) env recpatlist
            |  Tpat_or (pat1,pat2) -> env (* An or pattern cannot declare any variable *)
   in enrich env pattern

let rec set_unknown_tail_call boundinfo =
   List.iter (fun info -> info.IdentInfo.unknown_tail_call <- true) !boundinfo

let begin_scc boundinfo = 
   List.iter (fun info -> info.IdentInfo.current_scc <- true) !boundinfo

let end_scc boundinfo = 
   List.iter (fun info -> info.IdentInfo.current_scc <- false) !boundinfo

let noregister = fun _ -> ()

let update tail typ e = 
(* The tail flag may be only an approximation : 
some tail calls may not be detected
So if an expression is estimated not being in tail
position we mustn't put a letstack', because if this expression becomes
tail, there will be a bug. I keep this for safety *)
      let op = Context.eval e.e_node in
      e.e_stackop <- Store.get_norm op;
      e.e_stackty <- typ;
      e.e_stacktailop <- Store.get_tail tail op;
      e.e_stacktailty <- typ

let update_stack tail typ e =
      e.e_stacktailop <- Store.get_tail tail (Context.eval e.e_node);
      e.e_stacktailty <- typ

(* Determine tail, tail_call, ... *)

let get_func_info env rator tyrator =(* find information about the function *)
    let escapeinfo =
      match rator.exp_desc with
         (* Normally, a primitive is declared in the .ml file, then its
            escape information is stored in the environment at the time
            of the declaration, but if the primitive is only declared in
            the .mli file, the escape information must be retrieved
            here *)
         Texp_ident(_,{val_type=val_type;
                       val_kind=Val_prim {Primitive.prim_name=prim_name;
                          Primitive.prim_native_name = prim_native_name}}) -> 
           (** We should use prim_native_name when it is not empty **)
           let prim_name = if prim_native_name = "" then prim_name else prim_native_name in
           (* The user cannot redefine a library primitive *)
           let val_type = TranslType.expr env val_type in
           IdentInfo.lookup_primitive prim_name val_type
       | Texp_ident(path,_) ->
           (match Envir.lookup_ident path env with
               None -> IdentInfo.empty tyrator (!in_let_rec)
             | Some(rator_info) -> rator_info)
       | Texp_function(_) ->
	   (* This case should never happen, since we beta reduce before *)
           Message.debug 0 (fun () -> print_string 
               "Strange : this function should have been beta reduced");
           IdentInfo.empty tyrator (!in_let_rec)
       | _ -> IdentInfo.empty tyrator (!in_let_rec)
    in (escapeinfo.IdentInfo.tupled,
        !Control.preserve_tail_calls or 
          ((escapeinfo.IdentInfo.current_scc or
            escapeinfo.IdentInfo.unknown_tail_call) &
            !Control.preserve_tail_recursion),
	!Control.preserve_tail_calls or 
          (escapeinfo.IdentInfo.unknown_tail_call &
            !Control.preserve_tail_recursion),
	if escapeinfo.IdentInfo.tupled then 1 else IdentInfo.arity escapeinfo)

let rec equations_for_clauses env clauses tye tyinfo tail tupled (*register*) =
  List.map (fun (pattern,body) -> 
    let body = tail_for_expression body tye tyinfo 
        (enrich_env_useless env pattern) tail tupled in
    (pattern, body)) clauses

and eq_for_exp_new_type env e tail tupled =
  let tye = TranslType.expr env e.exp_type in
  let tyinfo = Lattice.typeinfo tye in
  tail_for_expression e tye tyinfo env tail tupled

and tail_for_expression e tye tyinfo env tail tupled =

(* gives back e' : expression
   e : Typedtree.expression
   tye : NewType.t_type_expr;
   tyinfo : int option None : e not level preserving; Some top : top(e)=top
   env : Envir.t = Var --> IdentInfo.t *)

let node = Context.newnode() in

let e = Simplify.simplify e in

let annotate_desc tailcalls unk_tailcalls d =
  {e_desc = d;
   e_loc = e.exp_loc;
   e_type = e.exp_type;
   e_env = e.exp_env;
   e_tail = tail;
   e_tailcalls = tailcalls;
   e_unknown_tailcalls = unk_tailcalls;
   e_node = node;
   e_tye = tye; (* champ ajoute' !! *)
   e_type_info = tyinfo;
   e_immop = Store.opnlp;
   e_stackop = Store.opnlp;
   e_stackty = None;
   e_stacktailop = Store.opnlp;
   e_stacktailty = None
   }

in

match e.exp_desc with
  Texp_ident(i,v) -> annotate_desc false false (Eexp_ident(i,v))

| Texp_constant(c) -> annotate_desc false false (Eexp_constant(c))

| Texp_let(rec_flag, stack_flag, bindings, body) ->
     let first_rec = (not (!in_let_rec)) & (rec_flag = Recursive) in
     if first_rec then in_let_rec := true;
     let boundinfo = ref [] in
     let (newenv, ltupled) = List.fold_left
       (fun (newenv, ltupled) (pattern,expression) ->
	 let parop = IdentInfo.according_exp expression in
	 (enrich_env_with_pattern boundinfo parop newenv pattern,
	  (fst parop :: ltupled)))
       (env, []) bindings in
     let bindings' = 
       List.map2 (fun (pattern, e) tupled  -> 
	 let e' = eq_for_exp_new_type
	     (if rec_flag = Recursive then newenv else env) e false tupled in
         if e'.e_unknown_tailcalls then set_unknown_tail_call boundinfo;
         (pattern, e'))
         bindings (List.rev ltupled)
     in
     if first_rec then 
       begin
	 in_let_rec := false;
	 end_scc boundinfo
       end;
     let inlinesupinfo = !Simplify.inline_sup_info in
     if (!Control.inlining_sup > 0) then
        List.iter (fun b -> Simplify.inline_sup_info :=
          Simplify.put_inline_info
          (!Simplify.inline_sup_info) rec_flag b) bindings;
     let body' = tail_for_expression body tye tyinfo newenv tail tupled in
     Simplify.inline_sup_info := inlinesupinfo;
     annotate_desc body'.e_tailcalls body'.e_unknown_tailcalls (Eexp_let(rec_flag, 
        stack_flag, bindings', body'))

| Texp_function(clauses) ->
    let (_, tyres) = NewType.extract_types_of_parameters false tye 1 in
    let clauses =
      if tupled then
	let tyinfo = Lattice.typeinfo tyres in
	List.map (fun (pattern,body) -> 
	  let body =
	    match pattern.pat_desc with
              Tpat_tuple(pat_list) -> 
		tail_for_expression body tyres tyinfo 
		  (enrich_env_useless env pattern) true false
	    | Tpat_any ->
		tail_for_expression body tyres tyinfo env true false
	    | _ -> Misc.fatal_error "equations_for_tupled_clauses" 
               (*not tupled clause*)
	  in
	  (pattern, body)) clauses
      else
        equations_for_clauses env clauses tyres (Lattice.typeinfo tyres)
          true tupled
    in annotate_desc
      (List.exists (fun (_,e) -> e.e_tailcalls) clauses)
      (List.exists (fun (_,e) -> e.e_unknown_tailcalls) clauses)
      (Eexp_function(clauses))

| Texp_apply({exp_desc = Texp_ident(i,({val_kind=Val_prim 
     {Primitive.prim_name = prim_name}} as v))} as rator, [arg1; arg2])
   when (prim_name = "%sequand" or prim_name = "%sequor") -> 
     let arg1 = tail_for_expression arg1 tye tyinfo env false false 
     and arg2 = tail_for_expression arg2 tye tyinfo env tail false in
     (* arg1 and arg2 are of type bool, like the whole expression e *)
      annotate_desc arg2.e_tailcalls arg2.e_unknown_tailcalls (Eexp_apply(
  {e_desc = Eexp_ident(i,v);
   e_loc = rator.exp_loc;
   e_type = rator.exp_type;
   e_env = rator.exp_env;
   e_tail = false;
   e_tailcalls = false;
   e_unknown_tailcalls = false;
   e_node = node;
   e_tye = tye;
   e_type_info = None;
   e_immop = Store.opnlp;
   e_stackop = Store.opnlp;
   e_stackty = None;
   e_stacktailop = Store.opnlp;
   e_stacktailty = None
   }, [arg1; arg2]))

| Texp_apply(rator,rands) ->
    let tyrands = List.map (fun rand -> TranslType.expr env rand.exp_type) rands in
    let tyrator = List.fold_right (fun ty1 ty2 ->
       Subst.annotate_htbl (NewType.ETarrow(ty1,ty2)) ty2.NewType.e_top true)
       tyrands tye in
    let tyratorinfo = Lattice.typeinfo tyrator in
    let (tupled, rec_tail_call, unk_tail_call, arity) = 
      get_func_info env rator tyrator in
    (* Test: arity < List.length rands because in this case, the called
       function returns a function which is then applied. The returned function
       may be extracted from a data structure => we have no escape information
       about it, and it may indeed cause recursion *)
    let tail_call = tail & (rec_tail_call or (arity < List.length rands & !Control.preserve_tail_recursion)) in
    let unk_tail_call = tail & (unk_tail_call or (arity < List.length rands & !Control.preserve_tail_recursion)) in
    let applrands = function (lparam, ltyparam) -> 
         List.map2 (fun param typaram -> 
             let tyinfo = Lattice.typeinfo typaram in
             tail_for_expression param typaram tyinfo env 
               false false) lparam ltyparam
    in 
    let rator' =
      tail_for_expression rator tyrator tyratorinfo env false false in
    annotate_desc tail_call unk_tail_call (
          if tupled then 
             match rands, tyrands with
                [{exp_desc = Texp_tuple(exp_list)}], [tyrand] -> 
                     let ty_exp_list = match tyrand with
                        {NewType.e_desc = NewType.ETtuple(ty_exp_list)} -> 
                           ty_exp_list
                      | _ -> Message.error "eq_for_exp" (fun () -> print_string
                           "tuple has not a tuple type") in
                     Eexp_tuple_apply(rator', 
                        (applrands (exp_list, ty_exp_list)))
              | _ -> Eexp_apply(rator', applrands (rands,tyrands))
          else
             Eexp_apply(rator', applrands (rands,tyrands)))

| Texp_match(expression, clauses) -> 
   let expression' = eq_for_exp_new_type env expression false false
   and clauses' = equations_for_clauses env clauses tye tyinfo tail tupled in
   annotate_desc
     (List.exists (fun (_,e) -> e.e_tailcalls) clauses')
     (List.exists (fun (_,e) -> e.e_unknown_tailcalls) clauses')
     (Eexp_match(expression', clauses'))

| Texp_try(expression, clauses) -> 
   let expression' = tail_for_expression expression tye tyinfo env false tupled
   and clauses' = equations_for_clauses env clauses tye tyinfo tail tupled in
   annotate_desc
     (List.exists (fun (_,e) -> e.e_tailcalls) clauses')
     (List.exists (fun (_,e) -> e.e_unknown_tailcalls) clauses')
     (Eexp_try(expression', clauses'))

| Texp_tuple(args) -> 
   let args = List.map (fun arg -> 
        eq_for_exp_new_type env arg false tupled) args
   in annotate_desc false false (Eexp_tuple(args))

| Texp_construct(constructor_desc, args) -> 
   let args = List.map (fun arg ->
        eq_for_exp_new_type env arg false tupled) args
   in annotate_desc false false (Eexp_construct(constructor_desc, args))

| Texp_record(field_descriptions) -> 
     let field_descriptions = List.map (fun (f, init) -> 
        (f, eq_for_exp_new_type env init  false tupled)) field_descriptions
     in annotate_desc false false (Eexp_record(field_descriptions))

| Texp_field(expr,desc) -> 
     let expr = eq_for_exp_new_type env expr false tupled in
     annotate_desc false false (Eexp_field(expr,desc))

| Texp_setfield(lhs,desc,rhs) ->
     let lhs = eq_for_exp_new_type env lhs false false in
     let rhs = eq_for_exp_new_type env rhs false false in
     annotate_desc false false (Eexp_setfield(lhs,desc,rhs))

| Texp_array(args) -> 
    let args =
      match args with
        [] -> []
      | (arg1::_) ->
         let tyarg = TranslType.expr env arg1.exp_type in
         let tyinfo = Lattice.typeinfo tyarg in
            List.map (fun arg -> tail_for_expression arg tyarg tyinfo
            env false tupled) args
    in 
    annotate_desc false false (Eexp_array(args))

| Texp_ifthenelse(predicate, consequent, alternative) ->
     let predicate = eq_for_exp_new_type env predicate false false
     and consequent = tail_for_expression consequent tye tyinfo env tail tupled
     and alternative = match alternative with None -> None | Some(e) -> 
        Some (tail_for_expression e tye tyinfo env tail tupled)
     in 
     annotate_desc
       (match alternative with
         None -> consequent.e_tailcalls
       | Some e -> consequent.e_tailcalls or e.e_tailcalls) 
       (match alternative with
         None -> consequent.e_unknown_tailcalls
       | Some e -> consequent.e_unknown_tailcalls or e.e_unknown_tailcalls) 
       (Eexp_ifthenelse(predicate, consequent, alternative))

| Texp_sequence(e1,e2) ->
     let e1 = eq_for_exp_new_type env e1 false false
     and e2 = tail_for_expression e2 tye tyinfo env tail tupled in
     annotate_desc e2.e_tailcalls e2.e_unknown_tailcalls (Eexp_sequence(e1,e2))

| Texp_while(e1,e2) ->
     let e1 = eq_for_exp_new_type env e1 false false
     and e2 = eq_for_exp_new_type env e2 false false in
     annotate_desc false false (Eexp_while(e1,e2))

| Texp_for(ident,e1,e2,direction,body) ->
     let tye1 = TranslType.expr env e1.exp_type in
     let tyinfo = Lattice.typeinfo tye1 in
     let e1 = tail_for_expression e1 tye1 tyinfo env false false
     and e2 = tail_for_expression e2 tye1 tyinfo env false false
     and body = eq_for_exp_new_type env body false false in
     annotate_desc false false (Eexp_for(ident,e1,e2,direction,body))

| Texp_when(cond,expr) ->
     let expr = tail_for_expression expr tye tyinfo env false false
     and cond = eq_for_exp_new_type env cond tail false in
     annotate_desc cond.e_tailcalls cond.e_unknown_tailcalls (Eexp_when(cond,expr))

| Texp_send (expr, m) -> (* TO DO : Be more precise for objects *)
     let expr = eq_for_exp_new_type env expr false false in
     annotate_desc false false (Eexp_send(expr,m))
| Texp_new t ->
     annotate_desc false false (Eexp_new t)
| Texp_instvar(n_self, n_field) ->
     annotate_desc false false (Eexp_instvar(n_self, n_field))
| Texp_setinstvar(n_self, n_field, expr) ->
     let expr = eq_for_exp_new_type env expr false false in
     annotate_desc false false (Eexp_setinstvar(n_self, n_field, expr))
| Texp_override(n_self, modifs) ->
     annotate_desc false false (Eexp_override(n_self, 
        List.map (fun (n_field, expr) ->
         (n_field, eq_for_exp_new_type env expr false false)) modifs))

let eq_for_exp_new_type = Profile.f4 "tail_for_expression" eq_for_exp_new_type

(* Real escape analysis *)

let falseemp = (false, [])

let get_func_info env rator tyrator tail =(* find information about the function *)
    let escapeinfo =
      match rator.e_desc with
         (* Normally, a primitive is declared in the .ml file, then its
            escape information is stored in the environment at the time
            of the declaration, but if the primitive is only declared in
            the .mli file, the escape information must be retrieved
            here *)
         Eexp_ident(_,{val_type=val_type;
                       val_kind=Val_prim {Primitive.prim_name=prim_name;
                          Primitive.prim_native_name = prim_native_name}}) -> 
           (** We should use prim_native_name when it is not empty **)
           let prim_name = if prim_native_name = "" then prim_name else prim_native_name in
           (* The user cannot redefine a library primitive *)
           let val_type = TranslType.expr env val_type in
           IdentInfo.lookup_primitive prim_name val_type
       | Eexp_ident(path,_) ->
           (match Envir.lookup_ident path env with
               None -> IdentInfo.empty tyrator (!in_let_rec)
             | Some(rator_info) -> rator_info)
       | Eexp_function(_) ->
	   (* This case should never happen, since we beta reduce before *)
           Message.debug 0 (fun () -> print_string 
               "Strange : this function should have been beta reduced");
           IdentInfo.empty tyrator (!in_let_rec)
       | _ -> IdentInfo.empty tyrator (!in_let_rec)
    in (IdentInfo.polymorphic_instantiation tail escapeinfo tyrator, 
        escapeinfo.IdentInfo.tupled)

let rec equations_for_expression e env var_env sublist sublist_stack (lp : bool * Store.op list): unit =

(* e : expression
   env : Envir.t = Var --> IdentInfo.t
   var_env : VarEnv.t = Var --> Context.t
   sublist : expression list ref
   sublist_stack : expression list ref
   lp : Store.op list 


   tail = e.e_tail
   tye = e.e_tye : NewType.t_type_expr;
   tyinfo = e.e_type_info
   tyinfo : int option None : e not level preserving; Some top : top(e)=top
   node = e.e_node
   *)

let linkafter (* closure contains: e env var_env *) e' op lp =
    let tyinfo = e'.e_type_info in
    let sublist = ref [] in
    let sublist_stack = ref [] in
    equations_for_expression e' env var_env sublist sublist_stack lp;
    List.iter (update e'.e_tail tyinfo) !sublist;
    List.iter (update_stack e'.e_tail tyinfo) !sublist_stack;
    Context.link e.e_node e'.e_node op

in

let equations_for_clauses (* closure contains: e env var_env *) clauses lp register =
   let var_analpar = ref [] in
   List.iter (fun (pattern,body) -> 
      let sublist = ref [] in
      let sublist_stack = ref [] in
      equations_for_expression body
         (enrich_env_useless env pattern)
         (enrich_var_env_with_pattern var_analpar var_env pattern)
         sublist sublist_stack lp;
             (* I don't know what the parameter p1 does to his parameters
                assume the worst => they escape *)
      List.iter (update body.e_tail body.e_type_info) !sublist;
      List.iter (update_stack body.e_tail body.e_type_info) !sublist_stack;
      Context.link e.e_node body.e_node Context.opid) clauses;
   register var_analpar;
   ()

in

let linkbefore (* closure contains: e env var_env sublist sublist_stack *) e' op body_tailcalls lp =
        if body_tailcalls then
        begin
           let sublist' = ref [] in
           equations_for_expression e' env var_env sublist' sublist_stack lp;
           List.iter (fun subexpr -> 
              if !Control.move_lets then
              begin
                 subexpr.e_stackop <- Store.get_norm (Context.eval subexpr.e_node);
                 subexpr.e_stackty <- e'.e_type_info
              end;
              (* If we cannot move lets, letstack is simply forbidden
                 because of the tail call *)
              sublist_stack:= subexpr :: !sublist_stack) !sublist';
           e'.e_immop <- Store.get_norm op;
           (* e'.e_stackop will be left Store.opnlp => no letstack *)
           sublist_stack:= e'::!sublist_stack;
           Context.link e.e_node e'.e_node op
        end
        else
        begin
           equations_for_expression e' env var_env sublist sublist_stack lp;
           e'.e_immop <- Store.get_norm op;
           sublist:= e'::!sublist;
           Context.link e.e_node e'.e_node op
        end

in

match e.e_desc with
  Eexp_ident(Path.Pident(name),({val_type = type_expr})) ->
     (match VarEnv.lookup name var_env with
       None -> ()
     | Some  nodelist -> 
          nodelist := (e.e_node, e.e_tye, TranslType.expr env type_expr) :: !nodelist)

| Eexp_ident(i,v) -> ()

| Eexp_constant(c) -> ()

| Eexp_let(rec_flag, stack_flag, bindings, body) ->
     let boundinfo = ref [] in
     let (newenv, newvar_env, ldata) = List.fold_left
       (fun (newenv, newvar_env, ldata) (pattern,expression) ->
          let anal = ref []
          and parop = IdentInfo.according_Eexp expression in
          (enrich_env_with_pattern boundinfo parop newenv pattern,
           enrich_var_env_with_pattern anal newvar_env pattern,
           (anal, parop)::ldata))
	 (env,var_env,[]) bindings in
     equations_for_expression body newenv 
        newvar_env sublist sublist_stack lp;
     Context.link e.e_node body.e_node Context.opid;
     if rec_flag = Recursive then
       let lanalop = 
	 List.map2 (fun (pattern,expression) (anal,parop) -> 
           let op = Context.newop "let"
           and sublist' = ref []
           and sublist_stack' = ref [] in
           equations_for_expression expression newenv 
             newvar_env sublist' sublist_stack' parop;
           List.iter (update false expression.e_type_info) !sublist';
           List.iter (update_stack false expression.e_type_info) !sublist_stack';
              (* sublist_stack' should be empty *)
           Context.link e.e_node expression.e_node op;
           expression.e_immop <- Store.get_norm op;
           if body.e_tailcalls then
             sublist_stack := expression::!sublist_stack
              (* No letstack if there is an unknown tail call *)
           else
             sublist := expression::!sublist;
           (op, anal)) bindings (List.rev ldata)
       in 
       List.iter (fun (op, anal) -> Context.register_anal op anal) lanalop
     else
       List.iter2 (fun (pattern, e) (anal,parop) -> 
         linkbefore e (Context.register_anal0 anal) body.e_tailcalls parop)
         bindings (List.rev ldata)

| Eexp_function(clauses) ->
    let (tupled, list_par) = lp in
    (match list_par with
      op :: lp -> 
         if tupled then
	   let lp = list_par in
           (* An analysis for each parameter *)
	   let var_analpar = List.map (fun _ -> ref []) lp in
	   List.iter (fun (pattern,body) -> 
	     let sublist = ref [] in
	     let sublist_stack = ref [] in
	     begin
	       match pattern.pat_desc with
		 Tpat_tuple(pat_list) -> 
		   let new_var_env = List.fold_left2 (fun var_env var_analpar pat ->
                     enrich_var_env_with_pattern var_analpar var_env pat) 
                       var_env var_analpar pat_list in
		(* The number of patterns must correspond to the number of
                   operations *)
		   equations_for_expression body 
		     (enrich_env_useless env pattern) new_var_env
		     sublist sublist_stack falseemp
	       | Tpat_any ->
		   equations_for_expression body 
		     env var_env sublist sublist_stack falseemp
	       | _ -> Misc.fatal_error "equations_for_tupled_clauses" 
               (*not tupled clause*)
	     end;
	     List.iter (update body.e_tail body.e_type_info) !sublist;
	     List.iter (update_stack body.e_tail body.e_type_info) !sublist_stack;
	     Context.link e.e_node body.e_node Context.opid) clauses;
	   List.iter2 Store.register_anal lp var_analpar

         else
           equations_for_clauses clauses (tupled,lp) (Store.register_anal op)
    | [] ->
         (* It's useless to compute information about the parameters *)
         equations_for_clauses clauses (tupled,[]) noregister)

| Eexp_apply({e_desc = Eexp_ident(i,({val_kind=Val_prim 
     {Primitive.prim_name = prim_name}} as v))} as rator, [arg1; arg2])
   when (prim_name = "%sequand" or prim_name = "%sequor") -> 
     linkafter arg1 Context.opzero falseemp;
     linkafter arg2 Context.opid falseemp
     (* arg1 and arg2 are of type bool, like the whole expression e *)

| Eexp_apply(rator,rands) ->
    let (lop,tupled) = get_func_info env rator rator.e_tye e.e_tail in
    let rec applrands = function
	([],  _) -> ()
      | (lparam, []) -> 
          List.iter (fun param -> 
            linkbefore param (Context.optop param.e_type_info) e.e_tailcalls falseemp) lparam
      | (param::lparam, op::lop) -> 
          linkbefore param op e.e_tailcalls falseemp;
	  applrands (lparam, lop)
    in 
    if e.e_tail then
      linkbefore rator Context.optail e.e_tailcalls falseemp
          (* rator tail-escapes *)
    else
      linkbefore rator Context.opid e.e_tailcalls falseemp;
    if tupled then 
      match rands with
        [{e_desc = Eexp_tuple(exp_list)}] -> 
          Message.error "eq_for_exp" (fun _ -> print_string "Unexpected tuple")
      | _ ->
	  let op = Context.newop "tuplesup" in
          List.iter (Context.addeq op) lop;
          applrands (rands, [op])
    else
      applrands (rands, lop)

| Eexp_tuple_apply(rator,rands) ->
    let (lop,tupled) = get_func_info env rator rator.e_tye e.e_tail in
    let rec applrands = function
	([],  _) -> ()
      | (lparam, []) -> 
          List.iter (fun param -> 
            linkbefore param (Context.optop param.e_type_info) e.e_tailcalls falseemp) lparam
      | (param::lparam, op::lop) -> 
          linkbefore param op e.e_tailcalls falseemp;
	  applrands (lparam, lop)
    in 
    if e.e_tail then
      linkbefore rator Context.optail e.e_tailcalls falseemp
          (* rator tail-escapes *)
    else
      linkbefore rator Context.opid e.e_tailcalls falseemp;
    if tupled then 
          applrands (rands, lop)
    else
      Message.error "eq_for_exp" (fun _ -> print_string "Expecting tupled")

| Eexp_match(expression, clauses) -> 
   let op = Context.newop "match" in
   linkafter expression op falseemp;
   equations_for_clauses clauses lp (Context.register_anal op)

| Eexp_try(expression, clauses) -> 
   linkafter expression Context.opid lp;
   equations_for_clauses clauses lp noregister

| Eexp_tuple(args) -> 
    List.iter (fun arg -> 
      linkbefore arg (Context.optopinter (Lattice.topof arg.e_tye))
	false lp) args

| Eexp_construct(_, args) -> 
    List.iter (fun arg ->
      linkbefore arg (Context.optopinter (Lattice.topof arg.e_tye))
	false lp) args

| Eexp_record(field_descriptions) -> 
    List.iter (fun (f, init) -> 
      linkbefore init (Context.optopinter (Lattice.topof init.e_tye))
	false lp) field_descriptions

| Eexp_field(expr,_) -> 
    linkbefore expr (match expr.e_type_info with
         None -> Context.optop None
       | Some _ -> Context.opid) false lp 

| Eexp_setfield(lhs,_,rhs) ->
    linkbefore lhs Context.opzero false falseemp;
    linkafter rhs (Context.optop rhs.e_type_info) falseemp

| Eexp_array(args) -> 
    begin
    match args with
      [] -> ()
    | (arg1::_) ->
        let tytop = Lattice.topof arg1.e_tye in
        List.iter (fun arg -> linkbefore arg
            (Context.optopinter tytop) false lp) args
    end

| Eexp_ifthenelse(predicate, consequent, alternative) ->
    linkafter predicate Context.opzero falseemp;
    linkafter consequent Context.opid lp;
    begin
      match alternative with 
	None -> ()
      | Some(e) -> linkafter e Context.opid lp
    end

| Eexp_sequence(e1,e2) ->
    linkafter e1 Context.opzero falseemp;
    linkafter e2 Context.opid lp

| Eexp_while(e1,e2) ->
    linkafter e1 Context.opzero falseemp;
    linkafter e2 Context.opzero falseemp

| Eexp_for(ident,e1,e2,direction,body) ->
    linkbefore e1 Context.opzero false falseemp;
    linkbefore e2 Context.opzero false falseemp;
    linkafter body Context.opzero falseemp

| Eexp_when(cond,expr) ->
    linkafter expr Context.opid falseemp;
    linkafter cond Context.opzero falseemp

| Eexp_send (expr, m) -> (* TO DO : Be more precise for objects *)
    linkafter expr (Context.optop expr.e_type_info) falseemp
| Eexp_new _ -> ()
| Eexp_instvar(_, _) -> ()
| Eexp_setinstvar(_, _, expr) ->
    linkafter expr (Context.optop expr.e_type_info) falseemp
| Eexp_override(_, modifs) ->
    List.iter (fun (_, expr) ->
      linkafter expr (Context.optop expr.e_type_info) falseemp) modifs

let equations_for_expression = Profile.f6 "equations_for_expression" equations_for_expression

(* Main call *)

let eq_for_exp e env var_env result =
  let e = eq_for_exp_new_type env e true (fst result) in
  
  let sublist = ref [] in
  let sublist_stack = ref [] in
  equations_for_expression e env var_env sublist sublist_stack result;
  List.iter (update true e.e_type_info) !sublist;
  List.iter (update_stack true e.e_type_info) !sublist_stack;
  e

let eq_for_exp = Profile.f4 "eq_for_exp" eq_for_exp

open TEnv

let optimize_definition rec_flag bindings env =
    if !Control.detailed_chrono then
       Message.phasestart "generating semantic equations";
    KnuthSolve.reset();
    let boundinfo = ref [] in
    if rec_flag = Recursive then in_let_rec := true;
    let (newenv, lresexp) = List.fold_left
      (fun (newenv, lresexp) (pattern,expression) ->
          let parop = IdentInfo.according_exp expression in 
             (enrich_env_with_pattern boundinfo parop newenv pattern,
              parop::lresexp)) (env,[]) bindings in
    let bindings1 = List.map2  (fun (pattern,expression) resexp -> (pattern, 
        eq_for_exp expression (if rec_flag = Recursive then newenv else env)
                   (VarEnv.create()) resexp)) bindings (List.rev lresexp) in
    if !Control.detailed_chrono then
       Message.phaseend "generating semantic equations";
    if !Control.detailed_chrono then
       Message.phasestart "solving semantic equations";
    KnuthSolve.cfc();
    if !Control.detailed_chrono then
       Message.phaseend "solving semantic equations";
    end_scc boundinfo;
    if !Control.detailed_chrono then
       Message.phasestart "post_transform";
    let bindings = List.map (fun (pattern, e) -> (pattern, 
       Fstackoptimizer.post_transform_expression e)) bindings1 in
    if (rec_flag = Nonrecursive) && (!Control.inlining > 0) then
       List.iter (fun b -> Fstackoptimizer.inline_info := Fstackoptimizer.put_inline_info 
         (!Fstackoptimizer.inline_info) b) bindings1;
    if (!Control.inlining_sup > 0) then
       List.iter (fun b -> Simplify.inline_sup_info := Simplify.put_inline_info
         (!Simplify.inline_sup_info) rec_flag b) bindings;
    in_let_rec := false;
    if !Control.detailed_chrono then
       Message.phaseend "post_transform";
    (bindings,newenv)

let optimize_expression e env =
    let wildcard = {pat_desc=Tpat_any; pat_type=e.exp_type; pat_loc=Location.none} in
    let bindings = [(wildcard, e)] in
    let ([(_, e)], env) = optimize_definition Asttypes.Nonrecursive bindings env
    in e

(* Scan and optimize a class declaration *)

let optimize_class_field env = function
   Cf_inher(path, explist, l1,l2) ->
     Cf_inher(path, List.map (fun e -> optimize_expression e env) explist,
              l1, l2)
 | Cf_val(n, i, p, expopt) ->
     Cf_val(n, i, p, match expopt with
         None -> None | Some exp -> Some(optimize_expression exp env))
 | Cf_meth(m, exp) ->
     Cf_meth(m, optimize_expression exp env)

let optimize_class env (name, def) =
   (name, {cl_args = def.cl_args;
           cl_field = List.map (optimize_class_field env) def.cl_field;
           cl_loc = def.cl_loc})

(* Register types defined in signatures in the environment *)

let rec scan_module_type s env =
  match s with
    Tmty_ident name -> TranslType.copy_sig(Envir.lookup_module_type name env)
  | Tmty_signature sign -> Envir.get_hierarchy (scan_signature env sign)
  | Tmty_functor(name, sign, body) ->
    let typemap = !TranslType.typemap in
    let rand = scan_module_type sign (Envir.startmodule env) in
    let env = Envir.add_module name rand env in
       TranslType.typemap := typemap;
       Functor(rand, (scan_module_type body env))

and scan_signature env sign =
   let (ldecl, env) = List.fold_left scan_signature_item ([],env) sign
   in TranslType.decl env ldecl

and scan_signature_item (ldecl, env) = function
    Tsig_value(name,{val_type = ty}) -> 
       let env = TranslType.decl env ldecl in
       let ty = TranslType.expr env ty in
       if NewType.functional ty then
          let info = 
             {IdentInfo.type_expr = ty;
              IdentInfo.current_scc = false;
              IdentInfo.tupled = false;
              IdentInfo.unknown_tail_call = true;
              IdentInfo.escape_info = []}
          in ([], Envir.update name info env)
       else ([],env)
  | Tsig_type(name, decl) -> ((name,decl)::ldecl, env)
  | Tsig_exception(_,_) -> (ldecl, env)
  | Tsig_module(name, body) ->
       let env = TranslType.decl env ldecl in
       let typemap = !TranslType.typemap in
       let bodyenv = scan_module_type body (Envir.startmodule env) in
          TranslType.typemap := typemap;
          ([],Envir.add_module name bodyenv env)
  | Tsig_modtype(name, body) ->
       let env = TranslType.decl env ldecl in
       let typemap = !TranslType.typemap in
       let bodyenv = match body with
             Tmodtype_abstract -> Module TEnv.emptyhierarchy
           | Tmodtype_manifest modtype -> scan_module_type modtype (Envir.startmodule env) in
          TranslType.typemap := typemap;
          ([],Envir.add_module_type name bodyenv env)
  | Tsig_class(name,body) -> (ldecl, env) (* TO DO *)

(* Scan and optimize a structure *)

let rec optimize_structure (items: Typedtree.structure) (env: Envir.t) =
  let process_item (items, env) s = 
      let (s, env) = optimize_structure_item s env
      in (s::items, env)
  in
      let (items', env) = List.fold_left process_item ([],env) items
      in
          (List.rev items', env)

and optimize_structure_item (s: Typedtree.structure_item) (env: Envir.t) =
  match s with
    Tstr_eval(e) -> (Tstr_eval(optimize_expression e env), env)
  | Tstr_value(recursive, bindings) ->
        let (bindings', env) = optimize_definition recursive bindings env in
            (Tstr_value(recursive, bindings'), env)
  | Tstr_primitive(name,{val_type = valtype; val_kind = valkind}) -> 
        (s, match valkind with
              Val_reg | Val_ivar _ | Val_anc _ -> env
            | Val_prim {Primitive.prim_name=prim_name; 
                        Primitive.prim_native_name = prim_native_name} ->
           (** We should use prim_native_name when it is not empty **)
           let prim_name = if prim_native_name = "" then prim_name else prim_native_name in
           (* The user cannot redefine a library primitive *)
           let val_type = TranslType.expr env valtype in
           Envir.update name (IdentInfo.lookup_primitive prim_name val_type)
              env)
  | Tstr_type(decllist) -> (s, TranslType.decl env decllist)
  | Tstr_exception(_,_) -> (s, env)
  | Tstr_module(name, body) ->
    let typemap = !TranslType.typemap in
    let inlineinfo = !Fstackoptimizer.inline_info in
    let inlinesupinfo = !Simplify.inline_sup_info in
    let (body', bodyenv) = optimize_module_expr body (Envir.startmodule env) in
    let env = Envir.add_module name bodyenv env in
          TranslType.typemap := typemap;
          Fstackoptimizer.inline_info := inlineinfo;
          Simplify.inline_sup_info := inlinesupinfo;
          (Tstr_module(name, body'), env)
  | Tstr_modtype(name, body) -> 
    let typemap = !TranslType.typemap in
    let bodyenv = scan_module_type body (Envir.startmodule env) in
    let env = Envir.add_module_type name bodyenv env in
          TranslType.typemap := typemap;
          (s, env)
  | Tstr_open(_) -> (s, env)
  | Tstr_class(class_list) -> 
      (Tstr_class(List.map (optimize_class env) class_list),env)
      (* TO DO : Add the class to the environment *)

and optimize_module_expr {mod_desc=mod_desc; mod_loc=mod_loc;
   mod_type=mod_type; mod_env=mod_env} env =
    let (desc,env) = optimize_module_expr_desc mod_desc env in
    ({mod_desc=desc;
      mod_loc=mod_loc;
      mod_type=mod_type;
      mod_env=mod_env},
     env)

and optimize_module_expr_desc d (env: Envir.t) =
  match d with
    Tmod_ident(path) -> 
       (Tmod_ident(path),Envir.lookup_module path env)
  | Tmod_structure(str) ->
    let (str, env) = optimize_structure str env in
      (Tmod_structure(str),Envir.get_hierarchy env)
  | Tmod_functor(name, sign, body) ->
    let typemap = !TranslType.typemap in
    let rand = scan_module_type sign (Envir.startmodule env) in
    let env = Envir.add_module name rand env in
    TranslType.typemap := typemap;
    let (body, bodyenv) = optimize_module_expr body env in
      (Tmod_functor(name, sign, body),Functor(rand, bodyenv))
  | Tmod_apply(rator, rand, coercion) ->
    let (rator, ratorenv) = optimize_module_expr rator env in
    let (rand, randenv) = optimize_module_expr rand env in
    let (Functor(rand_var, body)) = ratorenv in
      (Tmod_apply(rator, rand, coercion),
       TranslType.instantiate rand_var randenv body)
  | Tmod_constraint(module_expr, sign, coercion) ->
    let (module_expr, env) = optimize_module_expr module_expr env in
      (Tmod_constraint(module_expr, sign, coercion),env)

open NewType

let decl_empty_group (Path.Pident name) env =
   Envir.add_type name
   {e_modname = "";
    e_name = name;
    e_type_params = [];
    e_color = 4;
    e_inverse_dep = [];
    e_type_kind = EType_group [];
    e_bad_type = false;
    e_nlp_decl = false;
    e_type_graph = []} env

let decl_exn (Path.Pident name) env =
   Envir.add_type name
   {e_modname = "";
    e_name = name;
    e_type_params = [];
    e_color = 4;
    e_inverse_dep = [];
    e_type_kind = EType_group [];
    e_bad_type = true;
    e_nlp_decl = true;
    e_type_graph = []} env

let decl_array (Path.Pident name) env =
  let v = ETvar 0 in
  let rec decl = {e_modname = "";
    e_name = name;
    e_type_params = [v];
    e_color = 4;
    e_inverse_dep = [];
    e_type_kind = EType_group [annotate_desc v];
    e_bad_type = false;
    e_nlp_decl = false;
    e_type_graph = [decl]} in
  Envir.add_type name decl env

let decl_format (Path.Pident name) env =
   Envir.add_type name
   {e_modname = "";
    e_name = name;
    e_type_params = [ETvar 0; ETvar 0; ETvar 0];
    e_color = 4;
    e_inverse_dep = [];
    e_type_kind = EType_abstract;
    e_bad_type = false;
    e_nlp_decl = false;
    e_type_graph = []} env

let initial_env() =
    decl_empty_group Predef.path_int (
    decl_empty_group Predef.path_char (
    decl_empty_group Predef.path_string (
    decl_empty_group Predef.path_float (
    decl_empty_group Predef.path_bool (
    decl_empty_group Predef.path_unit (
    decl_exn Predef.path_exn (
    decl_array Predef.path_array (
    decl_array Predef.path_list (
    decl_format Predef.path_format (
    Envir.create()))))))))))


let optimize(str: Typedtree.structure) =
    begin
	Compilenv.translator := fun modname sign ->
	    let oldmodname = !TranslType.modname in
	    TranslType.modname := modname;
	    let Module res = Envir.get_hierarchy (scan_signature (initial_env()) sign) in
	        TranslType.modname := oldmodname;
        	Allanal.fromF res
    end;
    let phase = "Escape analysis"
    and typemap = !TranslType.typemap 
    and inlineinfo = !Fstackoptimizer.inline_info
    and inlinesupinfo = !Simplify.inline_sup_info in
    Message.phasestart phase;
    let env = initial_env() in
    TranslType.modname := Compilenv.current_unit_name();
    let (str', env) = optimize_structure str env in
    Envir.set_escape_info env;
    (* Free useless data *)
    TranslType.typemap := typemap;
    Fstackoptimizer.inline_info := inlineinfo;
    Simplify.inline_sup_info := inlinesupinfo;
    Compilenv.empty_esc_info();
    Message.phaseend phase;
    if !Control.show_environment then Envir.print env;
    if !Control.dump_stackallocation then
       Printtypedtree.print str';
    Message.debug 1 (fun () -> print_int(!Control.promoted_lets_count);
			       print_string " letstack introduced.");
    str'
