(***********************************************************************)
(*                                                                     *)
(*                         Fast Escape Analysis                        *)
(*                                                                     *)
(*                  Bruno Blanchet, ENS Ulm - INRIA Rocquencourt       *)
(*                  Alain Deutsch, INRIA Rocquencourt                  *)
(*                                                                     *)
(*  Copyright 1995 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)


open Typedtree
open Format
open Asttypes
open Fsolver

let rec list_map3 f l1 l2 l3 =
  match (l1, l2, l3) with
    (a1::l1, a2::l2, a3::l3) -> let r = f a1 a2 a3
                                        in r :: list_map3 f l1 l2 l3
  | ([], [], []) -> []
  | (_,_,_) -> invalid_arg "list_map3"


let rec list_fold_max f l1 l2 =
   match (l1, l2) with
     ([], []) -> 0
   | (a1 :: l1, a2 :: l2) -> max (f a1 a2) (list_fold_max f l1 l2)
   | (_,_) -> invalid_arg "list_fold_max"

module NewType =
struct

  type t_type_declaration =
  { e_modname: string; (* To garantee unicity, we use both module name and
identifier (with its stamp, each identifier is unique in a file) *)
    e_name: Ident.t;
    mutable e_type_params: t_type_expr_desc list;
    mutable e_color: int;
    mutable e_inverse_dep: t_type_declaration list;
    mutable e_type_kind: t_type_kind;
    mutable e_bad_type: bool;
    mutable e_nlp_decl: bool;
    mutable e_type_graph: t_type_declaration list}

  and t_type_kind =
    EType_abstract
  | EType_group of t_type_expr list
  | EType_manifest of t_type_expr
      (* The names of constructors and fields have no importance *)

  and t_type_expr_desc =
    ETvar of int
  | ETarrow of t_type_expr * t_type_expr
  | ETtuple of t_type_expr list
  | ETconstr of t_type_declaration * t_type_expr list

  and t_type_expr =
  { mutable e_top : int;
    mutable e_nlp : bool;
    e_id : int;
    e_desc : t_type_expr_desc}

type t = t_type_expr

let rec max_list = 
  function [] -> 0
    | (a::l) -> max a.e_top (max_list l)

   let rec print_with_delimitors element separator list =
     match list with
       [] -> ()
     | [x] -> element x
     | (x::xs) -> element x;
                  print_string separator;
                  print_with_delimitors element separator xs
 
  let print_decl decl =
    if decl.e_modname <> "" then
      begin
	print_string decl.e_modname;
	print_string "."
      end;
    print_string (Ident.unique_name decl.e_name)

   let print (ty: t_type_expr) =
      let rec annotate prio ty =
        match (prio, ty.e_desc) with
          (_, ETconstr(decl, [])) ->
	    print_decl decl
        | (_, ETconstr(decl, [ty])) ->
            open_hovbox 0;
	    annotate 0 ty;
	    print_string " ";
	    print_decl decl;
            close_box()
        | (_, ETconstr(decl, types)) ->
            open_hovbox 0;
	    print_string "(";
	    print_with_delimitors (annotate 2 ) ", " types;
	    print_string ") ";
	    print_decl decl;
            close_box()
        | (_, ETvar n) -> print_string "'a"; print_int n
        | (0, _) ->
            open_hovbox 1;
	    print_string "(";
	    annotate 2 ty;
	    print_string ")";
            close_box()
        | (1, ETtuple(types)) ->
	    print_with_delimitors (annotate 0) " * " types
        | (2, ETarrow(ty1, ty2)) ->
            open_hovbox 0;
	    annotate 1 ty1;
	    print_string " -> ";
	    annotate 2 ty2;
            close_box()
        | (_, _) -> annotate (prio-1) ty
      in
      annotate 1 ty

let annotate_desc d = 
  { e_desc = d; 
    e_top = -1;
    e_id = 0;
    e_nlp = true}


let rec hashlist =
  function [] -> 0
    | (a::l) -> a.e_id + 997 * hashlist l

let hash ty =
  (match ty.e_desc with
    ETvar _ -> 1
  | ETarrow(ty1,ty2) -> 2 + 997 * ty2.e_id + 997 * 997 * ty1.e_id
  | ETtuple(tyl) -> hashlist tyl
  | ETconstr(decl,tyl) -> Hashtbl.hash decl.e_name + hashlist tyl)
    land 0x3FFFFFFF

let same_decl decl1 decl2 =
   (decl1 == decl2) or (Ident.same decl1.e_name decl2.e_name &&
decl1.e_modname = decl2.e_modname)
(* This test is reliable, because an identifier is unique in a file,
thanks to its stamp. I add the == test because it is the most frequent case
and is very quick to test. *)

let rec equal t1 t2 = match (t1.e_desc,t2.e_desc) with
      (ETvar _, ETvar _) -> true
    | (ETarrow(par1,ty1), ETarrow(par2,ty2)) -> 
	(par1 == par2) && (ty1 == ty2)
    | (ETtuple(types1), ETtuple(types2)) -> (try
          List.for_all2 (==) types1 types2
       with Invalid_argument("List.for_all2") -> false)
    | (ETconstr(decl1, types1), ETconstr(decl2, types2)) ->
         (same_decl decl1 decl2) && (List.for_all2 (==) types1 types2)
    | (_,_) -> false

let rec extract_types_not_tupled ty arity =
    if arity <= 0 then
        ([],ty)
    else
        match ty.e_desc with
          ETarrow(ty1, ty2) ->
             let (types, result) = extract_types_not_tupled ty2 (arity-1)
             in (ty1::types, result)
        | _ -> 
            Message.error "extract_types_of_parameters" (fun () ->
             print_string "arity too big for the number of parameters")

let extract_types_of_parameters tupled ty arity =
    if tupled then
       let ([typaram],tyres) = extract_types_not_tupled ty 1 in
       let rec extract_tupled typaram =
       match typaram.e_desc with
         ETtuple(typelist) ->
            if arity <>List.length typelist then
                Message.error "extract_types_of_parameters" (fun () ->
                  print_string "wrong arity for tupled function");
            (typelist, tyres)
       | _ -> Message.error "extract_types_of_parameters" (fun () ->
                         print_string "not a tupled function type")
       in extract_tupled typaram
    else
       extract_types_not_tupled ty arity

let functional ty = match ty.e_desc with
     ETarrow(_,_) -> true
   | _ -> false


end

module HashConsTbl = Hashtbl.Make(NewType)

module Subst =
struct

open NewType

let htbl = HashConsTbl.create 997 

let scc_accu = ref []

let count = ref 0

let annotate d top nlp =
  incr count;
  { e_desc = d; 
    e_top = top;
    e_id = !count;
    e_nlp = nlp}

let annotate_htbl d top nlp =
  let res = annotate d top nlp in
  try HashConsTbl.find htbl res
  with Not_found ->
    HashConsTbl.add htbl res res;
    Message.debug 4 (fun () -> 
      print_string "Top of ";
      print res;
      print_string " = ";
      print_int res.e_top);
    res

let annotate_accu d top nlp =
  let res = annotate d top nlp in
  try HashConsTbl.find htbl res
  with Not_found ->
    HashConsTbl.add htbl res res;
    scc_accu := res :: !scc_accu;
    res

let inst_subst = ref []

let rec calc_top decllist params =
  match decllist with
    [] -> 0
  | (decl::l) ->
      let top_decl = 
      	match decl.e_type_kind with
          EType_abstract | EType_manifest _ -> 
	    Message.error "Subst.calc_top" (fun () ->
	      print_string "No abbreviations or abstract types here")
      	| EType_group tyl ->
	    max_list (List.map (substitute2 decl.e_type_params params) tyl)
      in
      max (calc_top l params) top_decl

and trdecl2 decl params = 
  let res = annotate (ETconstr(decl,params)) 1
      (decl.e_nlp_decl || List.exists (fun ty -> ty.e_nlp) params) in
  try
    HashConsTbl.find htbl res
  with Not_found ->
    HashConsTbl.add htbl res res;
    begin
      if not decl.e_bad_type then
    	match decl.e_type_kind with
	  EType_abstract -> ()
    	| EType_manifest _ -> Message.error "Subst.trdecl2" (fun () ->
	    print_string "All abbreviations should have been expanded")
    	| EType_group _ ->
	    let anc_scc_accu = !scc_accu in
	    scc_accu := [];
	    let top = 1 + calc_top decl.e_type_graph params in
	    Message.debug 4 (fun () ->
	      print_string "--- scc : top = ";
	      print_int top);
	    List.iter (fun ty ->
	      ty.e_top <- top;
	      Message.debug 4 (fun () -> 
		print ty)) !scc_accu;
	    Message.debug 4 (fun () -> print res);
	    res.e_top <- top;
	    scc_accu := anc_scc_accu;
    end;
    res

and copy2 ty =
  if ty.e_top >= 0 then Message.error "copy2" (fun () ->
    print_string "top of types in declarations = -1 ou -2");
  match ty.e_desc with
    ETvar v as t ->
        (try
          List.assq t !inst_subst
        with Not_found ->
          Message.error "TranslType.copy" 
           (fun () -> print_string "Type variable not found"))
  | ETarrow(t1, t2) ->
      let dst = copy2 t2 in
      (if ty.e_top = -1 then annotate_accu else annotate_htbl)
 	(ETarrow(copy2 t1, dst)) dst.e_top true
  | ETtuple tl ->
      let tyl = List.map copy2 tl in
      (if ty.e_top = -1 then annotate_accu else annotate_htbl)
 	(ETtuple tyl) (1 + max_list tyl) 
	(List.exists (fun ty -> ty.e_nlp) tyl)
  | ETconstr(p, tl) ->
      let tyl = List.map copy2 tl in
      if ty.e_top = -1 then
	annotate_accu (ETconstr(p, tyl)) (1+max_list tyl) 
	  (p.e_nlp_decl || List.exists (fun ty -> ty.e_nlp) tyl)
      else
      	trdecl2 p tyl

and substitute2 params args body =
   let inst_subst_tmp = !inst_subst in
   inst_subst := List.combine params args;
   let res = copy2 body in
     inst_subst := inst_subst_tmp;
     res

let trdecl2 = Profile.f2 "Subst.trdecl2" trdecl2
let substitute2 = Profile.f3 "Subst.substitute2" substitute2
let annotate_htbl = Profile.f3 "Subst.annotate_htbl" annotate_htbl

end

module Lattice =
struct

  open NewType

  type t = int

  let topof ty = ty.e_top

  let not_level_preserving ty = ty.e_nlp

  let typeinfo ty = if ty.e_nlp then None else Some (ty.e_top) 

   let print_as_type_annotation (f1: t) (f2: t) (ty: t_type_expr) =
      let rec annotate prio just_print ty =
          let topof_ty = topof ty in
              if (not just_print) & (topof_ty <= f2)
                 & not (not_level_preserving ty) then
              begin
                      open_hovbox 0;
		      print_string "H(";
		      annotate 2 true ty;
		      print_string ")";
                      close_box()
              end
              else
                  match (prio, ty.e_desc) with
                      (_, ETconstr(decl, [])) -> NewType.print_decl decl
                    | (_, ETconstr(decl, [ty])) ->
                      open_hovbox 0;
		      annotate 0 just_print ty;
		      print_string " ";
		      NewType.print_decl decl;
                      close_box()
                    | (_, ETconstr(decl, types)) ->
                      open_hovbox 0;
		      print_string "(";
		      NewType.print_with_delimitors (annotate 2 just_print) ", " types;
		      print_string ") ";
		      NewType.print_decl decl;
                      close_box()
                    | (_, ETvar n) -> print_string "'a"; print_int n
                    | (0, _) ->
                      open_hovbox 1;
		      print_string "(";
		      annotate 2 just_print ty;
		      print_string ")";
                      close_box()
                    | (1, ETtuple(types)) ->
		       NewType.print_with_delimitors (annotate 0 just_print) " * " types
                    | (2, ETarrow(ty1, ty2)) ->
                       open_hovbox 0;
		       annotate 1 true ty1;
		       print_string " -> ";
		       annotate 2 just_print ty2;
                       close_box()
                    | (_, _) -> annotate (prio-1) just_print ty

      in
      if f1 = 1 then
      begin
         open_hovbox 0;
         print_string "H1(";
         annotate 1 true ty;
         print_string ")";
         close_box()
      end
      else
          annotate 1 false ty


   let is_ident ty1 ty2 =
      let rec scan_type ty1 ty2 = match (ty1.e_desc, ty2.e_desc) with
       (_, ETvar _) -> topof ty1 = 1
     | (ETarrow(_,dst1), ETarrow(_,dst2)) -> scan_type dst1 dst2
     | (ETtuple(types1), ETtuple(types2)) -> List.for_all2 scan_type types1 types2
     | (ETconstr(decl1, types1), ETconstr(decl2, types2)) when NewType.same_decl decl1 decl2 -> List.for_all2 scan_type types1 types2
     | (_, _) -> false
      in scan_type ty1 ty2

   let inst_gen ty1 ty2 n =
      let rec scan_type ty1 ty2 = let topof_ty1 = topof ty1 in
    if n >= topof_ty1 then topof ty2 + n - topof_ty1 else
    match (ty1.e_desc, ty2.e_desc) with
       (ETvar _, _) -> 0 (* n < topof_ty1=1 => n = 0 => result = 0 *)
     | (_, ETvar _) -> 1
     | (ETarrow(_,dst1), ETarrow(_,dst2)) -> scan_type dst1 dst2
     | (ETtuple(types1), ETtuple(types2)) -> list_fold_max scan_type types1 types2
     | (ETconstr(decl1, _), _) when decl1.e_type_kind = EType_abstract ->
         topof ty2
     | (_, ETconstr(decl2, _)) when decl2.e_type_kind = EType_abstract ->
         topof ty2
     | (ETconstr(decl1, types1), ETconstr(decl2, types2)) -> 
         if NewType.same_decl decl1 decl2 then
            list_fold_max scan_type types1 types2 
         else (* We must accept different declarations because when applying
                 functors we mix the declarations of the formal and real
                 parameters *)
         begin
            let rec intype seen ty1 ty2 =
                  if ty1.e_top = -2 (* Not in the top scc *) then
                      scan_type 
                      (Subst.substitute2 decl1.e_type_params types1 ty1)
                      (Subst.substitute2 decl2.e_type_params types2 ty2)
                  else
                      match (ty1.e_desc, ty2.e_desc) with
                        (* Type variables or abstract types should never be 
			   in the top scc *)
                        (ETtuple(ty1s), ETtuple(ty2s)) -> list_fold_max
                          (intype seen) ty1s ty2s
                      | (ETarrow(_,ty1), ETarrow(_,ty2)) -> intype seen ty1 ty2
                      | (ETconstr(decl1, ty1s), ETconstr(decl2, ty2s)) ->
                          if List.exists (NewType.same_decl decl1) seen
                          then 0 (* already seen decl *)
                          else if List.exists (fun ty -> 
                                    match ty.e_desc with
                                       ETvar _ -> false
                                     | _ -> true) ty1s
                                  (* not in the same scc *) 
                          then list_fold_max (intype seen) ty1s ty2s
                          else indecl seen decl1 decl2
                      | (_,_) -> failwith "(3) Declarations : "
  
               and indecl seen decl1 decl2 =
               match (decl1.e_type_kind, decl2.e_type_kind) with
                  (EType_abstract, _) | (_, EType_abstract) -> topof ty2
                | (EType_group fieldlist1, EType_group fieldlist2) ->
                     list_fold_max (intype (decl1 :: seen)) fieldlist1 fieldlist2
                | (_,_) -> failwith "(2) Declarations : "
              
              in indecl [] decl1 decl2
         end
     | (_, _) -> failwith ""
      in if n = 0 then 0 else 
         try
            scan_type ty1 ty2
         with Failure s ->
            Message.error "Lattice.inst_gen"
		 (fun () -> print_string s;
                            print_as_type_annotation 0 0 ty1;
                            print_string " and ";
                            print_as_type_annotation 0 0 ty2;
                            print_string " do not correspond.")
	 | Invalid_argument s ->
            Message.error "Lattice.inst_gen"
		 (fun () -> print_string ("Invalid_argument(" ^ s ^ ")");
                            print_as_type_annotation 0 0 ty1;
                            print_string " and ";
                            print_as_type_annotation 0 0 ty2;
                            print_string " do not correspond.")

   let instantiation tyinst tygen = 
       (inst_gen tygen tyinst, topof tyinst)

   let inst_nlp ty1 ty2 =
      let rec scan_type ty1 ty2 = 
    match (ty1.e_desc, ty2.e_desc) with
       (ETvar _, _) -> not_level_preserving ty2
     | (ETarrow(_,dst1), ETarrow(_,dst2)) -> scan_type dst1 dst2
     | (ETtuple(types1), ETtuple(types2)) -> List.exists2 scan_type types1 types2
     | (ETconstr(decl1, _), _) when decl1.e_type_kind = EType_abstract ->
         not_level_preserving ty2
     | (ETconstr(decl1, types1), ETconstr(decl2, types2)) -> 
         if NewType.same_decl decl1 decl2 then
            List.exists2 scan_type types1 types2 
         else (* We must accept different declarations because when applying
                 functors we mix the declarations of the formal and real
                 parameters *)
         begin
            let rec intype seen ty1 ty2 =
                  if ty1.e_top = -2 then
                      scan_type 
                      (Subst.substitute2 decl1.e_type_params types1 ty1)
                      (Subst.substitute2 decl2.e_type_params types2 ty2)
                  else
                      match (ty1.e_desc, ty2.e_desc) with
                        (* Type variables or abstract types should never be
			   in the graph *)
                        (ETtuple(ty1s), ETtuple(ty2s)) -> List.exists2
                          (intype seen) ty1s ty2s
                      | (ETarrow(_,ty1), ETarrow(_,ty2)) -> intype seen ty1 ty2
                      | (ETconstr(decl1, ty1s), ETconstr(decl2, ty2s)) ->
                          if List.exists (NewType.same_decl decl1) seen
                          then false (* Already seen *)
                          else if List.exists (fun ty -> 
                                    match ty.e_desc with
                                       ETvar _ -> false
                                     | _ -> true) ty1s
                                  (* not in the same scc *) 
                          then List.exists2 (intype seen) ty1s ty2s
                          else indecl seen decl1 decl2
                      | (_,_) -> failwith "(3) Declarations : "
  
               and indecl seen decl1 decl2 =
               match (decl1.e_type_kind, decl2.e_type_kind) with
                  (EType_abstract, _) -> not_level_preserving ty2
                | (EType_group fieldlist1, EType_group fieldlist2) ->
                     List.exists2 (intype (decl1 :: seen)) fieldlist1 fieldlist2
                | (_,_) -> failwith "(2) Declarations : "
              
              in indecl [] decl1 decl2
         end
     | (_, _) -> failwith ""
      in try
            scan_type ty1 ty2
         with Failure s ->
            Message.error "Lattice.inst_nlp"
		 (fun () -> print_string s;
                            print_as_type_annotation 0 0 ty1;
                            print_string " and ";
                            print_as_type_annotation 0 0 ty2;
                            print_string " do not correspond.")
	 | Invalid_argument s ->
            Message.error "Lattice.inst_nlp"
		 (fun () -> print_string ("Invalid_argument(" ^ s ^ ")");
                            print_as_type_annotation 0 0 ty1;
                            print_string " and ";
                            print_as_type_annotation 0 0 ty2;
                            print_string " do not correspond.")

end

module Context :
sig

open KnuthEquations
type op = (term * term * term) * (term * term * term) 
(* op1 = (func, imp, impr) means op = \lambda c. (c \sqcap func) \sqcup imp
                                 opr= \lambda c. (c \sqcap func) \sqcup impr
   opr describes what escapes in tail calls.
   The first element of the pair is for F1, the second for F2
 *)

type elemnode

type t = (elemnode * NewType.t_type_expr * NewType.t_type_expr) list ref
   (* Analysis for one free variable

      List of nodes in the tree where a variable is referenced
      We have two analyses F_1 and F_2. We do "addeq op_var (eval node)"
      when finishing the analysis of the binder of the variable.*)

val opid : op
val opzero : op
val optail : op
val optop : int option -> op
val optopinter : int -> op

val newnode : unit -> elemnode
val newop : string -> op
val genterm : op -> NewType.t_type_expr -> NewType.t_type_expr -> op
val addeq : op -> op -> unit
val eval : elemnode -> op
val link : elemnode -> elemnode -> op -> unit
val register_anal : op -> t -> unit
val register_anal0 : t -> op

end=
struct
  
open KnuthEquations

type op1 = term * term * term
(* op = (t1, t2, t2r) means op = \lambda c. (c \sqcap t1) \sqcup t2
                            opr= \lambda c. (c \sqcap t1) \sqcup t2r
 *)
type op = op1 * op1

type elemnode = {mutable pere : antiarbre; mutable oper : op}
 and antiarbre = Nil | Node of elemnode

type t = (elemnode * NewType.t_type_expr * NewType.t_type_expr) list ref
   (* Analysis for one free variable

      List of nodes in the tree where a variable is referenced
      We have two analyses F_1 and F_2. We do "addeq op_var (eval node)"
      when finishing the analysis of the binder of the variable.*)

let opid1 = (Term top, Term 0, Term 0)
let opzero1 = (Term 0,Term 0, Term 0)
let opcst1 x = (Term 0,Term x, Term x)
let opinter1 x = (Term x, Term 0, Term 0)

let opid = (opid1, opid1)

let opzero = (opid1, opzero1)

let optail = ((Term top, Term 0, Term 1),
                 (Term top, Term 0, Term 0))

let optop typaram = match typaram with
     None -> (opcst1 1, opzero1)
   | Some top -> (opid1, opcst1 top)

let optopinter top = (opid1, opinter1 top)

let newnode () = {pere = Nil; oper = opid}

let newop1 name = (Nonterm (KnuthSolve.newnonterm (name ^ "f")), 
                  Nonterm (KnuthSolve.newnonterm (name ^ "i")),
                  Nonterm (KnuthSolve.newnonterm (name ^ "r")))
let newop name = (newop1(name ^ "1"), newop1(name ^ "2"))

(* compose op1 op2 = op1 o op2 (composition of functions) *)
let compose1 (func,imp,impr) (func2,imp2,impr2) = 
   (KnuthSolve.termmin func func2, KnuthSolve.termmax (KnuthSolve.termmin func imp2) imp, KnuthSolve.termmax (KnuthSolve.termmin func impr2) impr)
let compose (f1,f2) (f1',f2') =
   (compose1 f1 f1', compose1 f2 f2')

let addeq1 (Nonterm tfunc,Nonterm timp, Nonterm timpr) (tfunc',timp',timpr') = 
      KnuthSolve.addeq tfunc tfunc';
      KnuthSolve.addeq timp timp';
      KnuthSolve.addeq timpr timpr'

let addeq (f1,f2) (f1',f2') =
      addeq1 f1 f1';
      addeq1 f2 f2'

let genterm (f1, ((f2func,f2imp,f2impr) as f2)) tyinst tygen = 
      (f1,
      if Lattice.topof tygen<>Lattice.topof tyinst then
            let genfun = Lattice.inst_gen tyinst tygen in
               (KnuthSolve.knuthfun f2func genfun,
                KnuthSolve.knuthfun f2imp genfun,
                KnuthSolve.knuthfun f2impr genfun)
      else f2)

let rec compress = 
   function {pere=Nil} -> failwith "Compress needs parent(v) not Nil"
         |  {pere = Node {pere = Nil}} -> ()
         |  {pere = Node pv} as v ->
                compress pv;
                v.oper <- compose v.oper pv.oper;
                v.pere <- pv.pere

(* Modification from Tarjan's paper : the root of the tree has no operation *)

let eval =
   function {pere = Nil} -> opid                  (* root *)
          | v -> compress v; v.oper             (* not the root *)

let eval = Profile.f1 "Context.eval" eval

let link v w ((_, f1imp, f1impr), f2) = 
   w.pere <- Node v; 
   w.oper <- ((Term top, f1imp, f1impr), f2)
(* Analysis F_1 must be greater than identity for all free variables, so a link
   in the tree must be greater than identity  -  14/11/96  *) 

let register_anal op anal =
      List.iter (fun (node,tyinst,tygener) -> 
          addeq op (genterm (eval node) tyinst tygener)) !anal

let register_anal0 anal = 
      match !anal with
          | [] -> (opzero1, opzero1)
          | [(node,tyinst,tygener)] ->
                genterm (eval node) tyinst tygener
          | _ -> let op = newop "analinterm" in
             List.iter (fun (node,tyinst,tygener) -> 
                addeq op (genterm (eval node) tyinst tygener)) !anal;
             op

end

module CTree :
sig

type op = (int * int) * (int * int)

type letone = Asttypes.rec_flag * Asttypes.stack_flag * int * int *
(Typedtree.pattern * Typedtree.expression) list
type blocks = (op * lettree) list
and  lettree = LetGroup of blocks
             | LetOne of letone

val optop : op
val opzero : op
val opid : op

val getvalue : op -> int option -> int -> int -> bool
(* tells if a block can be stack-allocated, according to the associated
   op to the control expression, the type of the control expression,
   maxc1 and maxc2 *)

val eval : blocks -> op * letone * blocks

val clearone : blocks -> blocks
(* clearone should come after eval *)

val compose : op -> op -> op

end =
struct
  
type op1 = int * int
(* op = (t1, t2) means op = \lambda c. (c \sqcap t1) \sqcup t2 *)
type op = op1 * op1

type letone = Asttypes.rec_flag * Asttypes.stack_flag * int * int *
(Typedtree.pattern * Typedtree.expression) list

type blocks = (op * lettree) list
and  lettree = LetGroup of blocks
             | LetOne of letone

let optop = ((0,1),(0,0))
let opzero = ((0,0),(0,0))
let opid = ((1,0),(top,0))

let getvalue1 (func,imp) context = max (min func context) imp
let getvalue (f1,f2) ty_englob maxc1 maxc2 = 
    match ty_englob with 
       None ->  getvalue1 f1 1 <= maxc1
     | Some topenglob -> (getvalue1 f1 0 <= maxc1) &&
        ((getvalue1 f2 topenglob) <= maxc2)

(* compose op op' = op o op' (composition of functions) *)
let compose1 (func,imp) (func',imp') = 
   (min func func', max (min func imp') imp)
let compose (f1,f2) (f1',f2') =
   (compose1 f1 f1', compose1 f2 f2')

let rec compress op accu = 
   function LetOne t -> (op, LetOne t) :: accu
          | LetGroup [] -> accu
          | LetGroup((op', a)::l) -> compress (compose op' op) 
                (if l = [] then accu else (op, LetGroup l) :: accu) a

let rec eval =
   function [] -> raise Not_found
          | ((op,LetOne t)::l) as tree -> (op, t, tree)
          | ((op,tree)::l) -> eval (compress op l tree)

let clearone = function (a::l) -> l
   (* clearone should come after eval *)

end

module Store :
sig

open KnuthEquations

type op = (term * term) * (term * term)

val newop : string -> op
val to_norm : bool -> op -> int option -> Context.op

val opid : op
val opzero : op
val opcst : int -> op
val opinter : int -> op
val opnlp : op

val get_norm : Context.op -> op
val get_tail : bool -> Context.op -> op
val to_const : op -> CTree.op

val getvalue : (term * term) -> int -> int
val instterm : op -> NewType.t_type_expr -> NewType.t_type_expr -> op
val register_anal : op -> Context.t -> unit

val compose : op -> op -> op (* To use only when equations are solved,
                                because it takes the explicit values *)
end =
struct

open KnuthEquations
(* Analysis for one function parameter *)

type op = (term * term) * (term * term)

let newop name = ((Nonterm (KnuthSolve.newnonterm (name ^ "1f")),
                   Nonterm (KnuthSolve.newnonterm (name ^ "1i"))),
                  (Nonterm (KnuthSolve.newnonterm (name ^ "2f")),
                   Nonterm (KnuthSolve.newnonterm (name ^ "2i"))))

let to_norm tail ((_,f1imp),(f2func,f2imp)) ty = if tail then
      match ty with
         None -> ((Term top,f1imp,Term 1),(f2func,f2imp,Term 0))
       | Some topty -> ((Term top,f1imp,Term 0),(f2func,f2imp,Term topty))
   else ((Term top,f1imp,f1imp),(f2func,f2imp,f2imp))

let opid = ((Term top, Term 0),(Term top,Term 0))
let opzero = ((Term top, Term 0),(Term 0,Term 0))
let opcst x = ((Term top, Term 0),(Term 0,Term x))
let opinter x = ((Term top, Term 0),(Term x,Term 0))
let opnlp = ((Term 0, Term 1),(Term 0, Term 0))

let get_norm ((f1func,f1imp,_),(f2func,f2imp,_)) =
    ((f1func,f1imp),(f2func,f2imp))

let get_tail tail ((f1func,_,f1impr),(f2func,_,f2impr)) =
       if tail then ((f1func,f1impr),(f2func,f2impr)) else opnlp

let to_const ((f1func,f1imp),(f2func,f2imp)) =
      ((KnuthSolve.compute f1func, KnuthSolve.compute f1imp),
       (KnuthSolve.compute f2func, KnuthSolve.compute f2imp))

let getvalue (t1,t2) context = max (min (KnuthSolve.compute t1) context) (KnuthSolve.compute t2)

let instterm (((f1func,f1imp) as f1),((f2func,f2imp) as f2)) tyinst tygen =
   ((if Lattice.inst_nlp tygen tyinst then
        (f1func, KnuthSolve.termmax f1imp (KnuthSolve.termmin (Term 1) f2imp))
     else f1),
    (if not (Lattice.is_ident tyinst tygen) then 
        let (instfun,cstapprox) = Lattice.instantiation tyinst tygen
        in (KnuthSolve.nonknuthfun f2func instfun cstapprox,
            KnuthSolve.nonknuthfun f2imp instfun cstapprox)
      else
          f2))

let addeq1 (Nonterm func,Nonterm imp) (func',imp',_) =
      KnuthSolve.addeq func func';
      KnuthSolve.addeq imp imp'

let addeq (f1,f2) (f1',f2') =
      addeq1 f1 f1';
      addeq1 f2 f2'

let register_anal op anal =
      List.iter (fun (node,tyinst,tygener) -> 
          addeq op (Context.genterm (Context.eval node) tyinst tygener)) !anal

let compose op1 op2 = let ((f1func, f1imp), (f2func, f2imp)) = 
    CTree.compose (to_const op1) (to_const op2)
  in ((Term f1func, Term f1imp), (Term f2func, Term f2imp))

end

module VarEnv =
struct

  (* Analysis for all the free variables *)
  type t = Context.t Ident.tbl

  let create() = Ident.empty

  let update ident value (env : t) = Ident.add ident value env

  let lookup x (env : t) = try Some(Ident.find_same x env) with Not_found -> None

end

type expression =
  { e_desc: expres_desc;
    e_loc: Location.t;
    e_type: Types.type_expr;
    e_env: Env.t;
    e_tail: bool;
    e_tailcalls: bool;
    e_unknown_tailcalls: bool;
    e_node: Context.elemnode;
    e_tye: NewType.t_type_expr;
    e_type_info: int option;
    mutable e_immop: Store.op;
    mutable e_stackop: Store.op;
    mutable e_stackty: int option;
    mutable e_stacktailop: Store.op;
    mutable e_stacktailty: int option }

and expres_desc =
    Eexp_ident of Path.t * Types.value_description
  | Eexp_constant of constant
  | Eexp_let of rec_flag * stack_flag * (pattern * expression) list * expression
  | Eexp_function of (pattern * expression) list
  | Eexp_apply of expression * expression list
  | Eexp_tuple_apply of expression * expression list
  | Eexp_match of expression * (pattern * expression) list
  | Eexp_try of expression * (pattern * expression) list
  | Eexp_tuple of expression list
  | Eexp_construct of Types.constructor_description * expression list
  | Eexp_record of (Types.label_description * expression) list
  | Eexp_field of expression * Types.label_description
  | Eexp_setfield of expression * Types.label_description * expression
  | Eexp_array of expression list
  | Eexp_ifthenelse of expression * expression * expression option
  | Eexp_sequence of expression * expression
  | Eexp_while of expression * expression
  | Eexp_for of
      Ident.t * expression * expression * direction_flag * expression
  | Eexp_when of expression * expression
  | Eexp_send of expression * string
  | Eexp_new of Path.t
  | Eexp_instvar of Path.t * Path.t
  | Eexp_setinstvar of Path.t * Path.t * expression
  | Eexp_override of Path.t * (Path.t * expression) list

module IdentInfo =
struct

  (* Analysis for one function (all its parameters) *)
  type t = {type_expr : NewType.t_type_expr;
            tupled : bool;
            mutable current_scc : bool;
            mutable unknown_tail_call : bool;
            escape_info : Store.op list }

(* A function will be "Tupled" if
in all the clauses of the definition of the function its first parameter
is matched against a tuple, and we are not in the case of a function with
one clause, and a body which is still a function (in the last case, the
function is "Curried"). See bytecomp/translcore.ml.
   When a function is tupled, we compile it as if the members of the tuple
were passed one after the other, which avoids allocating the tuple. In this
case, we must not "letstack" the tuple (such a letstack would forbid the
tuple optimization).
*)

  let escape_info(x: t) = x.escape_info
  let type_expr(x: t) = x.type_expr
  let arity(x: t) = List.length (x.escape_info)

  let empty ty in_let_rec =
         {type_expr = ty;
          tupled = false;
          current_scc = in_let_rec;
          unknown_tail_call = true;
          escape_info = []}

  let according_exp exp =
     let rec lpar_according_exp exp = match exp.exp_desc with
            Texp_function([(_,body)]) -> 
                  (Store.newop "n") :: (lpar_according_exp body)
         |  Texp_function(_) -> [Store.newop "n"]
         |  _ ->  []
     in match exp.exp_desc with
          Texp_function([_, {exp_desc = Texp_function pl}]) -> 
             (false, lpar_according_exp exp)
        | Texp_function((({pat_desc = Tpat_tuple pl}, _) :: _) as clauses) 
             when !Clflags.tupleopt -> 
              if List.for_all (function
		  ({pat_desc = Tpat_any | Tpat_tuple _},_) -> true
                | _ -> false) clauses then
                 (true, List.map (fun _ -> Store.newop "ntuple") pl)
              else
                 (false, lpar_according_exp exp)
        | _ -> (false, lpar_according_exp exp)

  let according_Eexp exp =
     let rec lpar_according_exp exp = match exp.e_desc with
            Eexp_function([(_,body)]) -> 
                  (Store.newop "n") :: (lpar_according_exp body)
         |  Eexp_function(_) -> [Store.newop "n"]
         |  _ ->  []
     in match exp.e_desc with
          Eexp_function([_, {e_desc = Eexp_function pl}]) -> 
             (false, lpar_according_exp exp)
        | Eexp_function((({pat_desc = Tpat_tuple pl}, _) :: _) as clauses) 
             when !Clflags.tupleopt -> 
              if List.for_all (function
		  ({pat_desc = Tpat_any | Tpat_tuple _},_) -> true
                | _ -> false) clauses then
                 (true, List.map (fun _ -> Store.newop "ntuple") pl)
              else
                 (false, lpar_according_exp exp)
        | _ -> (false, lpar_according_exp exp)

  let build ty (tupled, esc_info) in_let_rec =
                  {type_expr = ty;
                   tupled = tupled;
                   current_scc = in_let_rec;
                   unknown_tail_call = false;
                   escape_info = esc_info}

  let print {type_expr = ty; 
             tupled = tupled;
             escape_info = escape_info} =
      let arity = List.length escape_info in
      let (param_types, result_type) = NewType.extract_types_of_parameters tupled ty arity in
      let arrow = " -> " in
      let typesep = if tupled then " * " else arrow in
      let topres = Lattice.topof result_type in
      print_newline();

      let functional_escape_info = List.map2 (fun ((_,x2),op2) ty -> (KnuthSolve.compute x2, Store.getvalue op2 topres, ty)) escape_info param_types
      and imperative_escape_info = List.map2 (fun ((_,x2),(_,x2'))  ty -> (KnuthSolve.compute x2, KnuthSolve.compute x2', ty)) escape_info param_types in

      let print_escape_info escape_info (f1res, f2res) =
	  let g(f1,f2,ty) = Lattice.print_as_type_annotation f1 f2 ty
	  in
              open_hovbox 0;
              NewType.print_with_delimitors g typesep escape_info;
              print_string arrow;
	      Lattice.print_as_type_annotation f1res f2res result_type;
              close_box();
              print_newline()
      in
	  print_escape_info functional_escape_info (0,topres);
	  if not (List.for_all (fun (f1,f2,_) -> f1=0 & f2=0) imperative_escape_info) then
              print_escape_info imperative_escape_info (0,0)

  let polymorphic_instantiation tail
      ({tupled = tupled; escape_info = escape_info; 
        type_expr = generic_type} as id_info) instance_type =
      let arity = List.length escape_info in
      let (instance_type_for_args, _) = 
        NewType.extract_types_of_parameters tupled instance_type arity in
      let instance_type_infos = List.map Lattice.typeinfo instance_type_for_args in
      List.map2 (Store.to_norm tail)
        (if instance_type == generic_type then escape_info else
              let (generic_type_for_args, _) = 
                NewType.extract_types_of_parameters tupled generic_type arity in
              let escape_info' = list_map3 Store.instterm escape_info instance_type_for_args generic_type_for_args in
	      if !Control.debug_polymorphic_instanciation then
                      begin
                          open_vbox 0;
                          print_string "generic_type =";
                          print id_info;
                          print_string "instance_type =";
                          print {type_expr = instance_type;
                                 current_scc = false;
                                 tupled = tupled;
                                 unknown_tail_call = false;
                                 escape_info = escape_info'};
                          close_box()
                      end;
       escape_info') instance_type_infos

  let z type_of_x = Store.opzero (* Never escapes *)

  let r type_of_x =              (* Escapes when the result escapes *)
   (* If the top of the result is smaller than the top of the parameter,
      only the part of the parameter which has smaller top than
      the result escapes *)
     Store.opinter (Lattice.topof type_of_x)

  let a type_of_x =              (* Always escapes *)
     match Lattice.typeinfo type_of_x with
       None -> Store.opnlp
     | Some top -> Store.opcst top

  let lookup_primitive name ty =
      let escape_info = match name with
          "%raise" -> [a]
        | "%equal" | "%notequal" | "%lessthan" | "%greaterthan"
            | "%lessequal" | "%greaterequal" | "compare"
            | "%eq" | "%noteq" | "%sequand" | "%sequor"
            | "%addint" | "%subint" | "%mulint" | "%divint" 
            | "%modint" | "%andint" | "%orint" | "%xorint" 
            | "%lslint" | "%lsrint" | "%asrint"
            | "%addfloat" | "%subfloat" | "%mulfloat" | "%divfloat"
            | "pow" | "fmod" | "atan2" | "ldexp_float"
	    | "md5_chan" | "format_int" | "format_float"
	    | "obj_block" | "input_value_from_string"
 	    | "output_int" | "output_char" | "output_value" 
	    | "seek_out" | "seek_in" | "sys_rename" -> [z;z]
        | "%boolnot" | "%negint" | "%succint" | "%predint"
            | "exp" | "log" | "sqrt" | "sin" | "cos" | "tan" | "asin" | "acos" 
            | "atan" | "sinh" | "cosh" | "%absfloat" | "log10" | "ceil"
            | "floor" | "frexp_float" | "tanh" | "modf" | "%negfloat"
            | "%floatofint" | "%intoffloat" | "ml_string_length"
            | "create_string" | "%string_length"
	    | "gc_stat" | "gc_get" | "gc_set" | "gc_minor" | "gc_major"
	    | "gc_full_major" | "output_value_to_string"
	    | "obj_is_block" | "obj_tag" | "%obj_size" | "open_descriptor"
	    | "input_int" | "input_char" | "input_value"
            | "input_scan_line" | "is_printable"
            | "int_of_string" | "float_of_string" | "flush" | "pos_out" 
            | "pos_in" | "channel_size" | "close_channel" 
            | "sys_exit" | "sys_get_config" | "sys_close" 
            | "%incr" | "%decr" | "%array_length"
 	    | "sys_get_argv" | "sys_file_exists" | "sys_remove" | "sys_getenv" 
	    | "sys_system_command" | "sys_chdir" | "sys_getcwd" -> [z]
        | "blit_string" -> [z;z;z;z;z]
        | "%field0" | "%field1" -> [r]
	| "%identity" -> [a] (* We use this for unsafe operations
					in module obj *)
	| "md5_string" -> [z;z;z]
	| "hash_univ_param" -> [z;z;z]
	| "%obj_field" -> [r;z]
	| "%obj_set_field" -> [z;z;a]
        | "sys_open" -> [a;z;z]
        | "output" | "input" -> [z;z;z;z]
        | "%makemutable" -> [r]
        | "%setfield0" -> [z;a]
        | "%array_safe_get" | "%array_unsafe_get" -> [r;z]
        | "%array_safe_set" | "%array_unsafe_set" -> [z;z;a]
        | "make_vect" -> [z;r]
	| "%string_safe_get" | "%string_unsafe_get" -> [z;z]
	| "%string_safe_set" | "%string_unsafe_set" -> [z;z;z]
	| "fill_string" -> [z;z;z;z]
	| "install_signal_handler" -> [z;a]
        | _ -> Message.debug 0 
                    (fun () -> print_string ("unknown primitive: " ^ name));
               []
       in let arity = List.length(escape_info)
       in let (types_of_parameters, result_type) = 
             NewType.extract_types_of_parameters false ty arity
       in let info =
              {type_expr = ty;
               current_scc = false;
               tupled = false;
               unknown_tail_call = false;
               escape_info = List.map2 (fun x type_of_x -> x type_of_x)
                  escape_info types_of_parameters}
       in if !Control.debug_primitives then
              begin
                  print_string (name ^ ": ");
                  print info
              end;
          info

  let box_float name = match name with
      "%negfloat" | "%addfloat" | "%subfloat" | "%mulfloat" | "%divfloat" |
      "%intoffloat" | "%absfloat" | "pow" | "frexp" | "floor" | "ceil" |
      "exp" | "log" | "sqrt" | "sin" | "cos" | "tan" | "asin" | "acos" |
      "atan" | "atan2" | "sinh" | "cosh" | "fmod" | "tanh" -> true
    | _ -> false

  let is_an_allocation name = match name with
      (* "float" primitives give back boxed floats, so they allocate *)
      "%negfloat" | "%addfloat" | "%subfloat" | "%mulfloat" | "%divfloat" |
      "%floatofint" | "%absfloat" | "%makemutable"| "make_vect" | "pow" |
      "exp" | "log" | "sqrt" | "sin" | "cos" | "tan" | "asin" | "acos" |
      "atan" | "atan2" | "sinh" | "cosh" | "tanh" | "floor" | "ceil" |
      "fmod" -> true
    | _ -> false

end

module TEnv =
struct

  type t_env = IdentInfo.t Ident.tbl

  type t_typedecl = NewType.t_type_declaration Ident.tbl

  type t_hierarchy = {env : t_env;
                      types : t_typedecl;
                      submod : t_submodules;
                      submod_types: t_submodules}

  and t_hierarchy_mod = Module of t_hierarchy
                  | Functor of t_hierarchy_mod * t_hierarchy_mod
                        (* Pointer to the parameter environment * body
        The parameter environment will be changed when applying the functor *)

  and t_submodules = t_hierarchy_mod Ident.tbl

  let emptyhierarchy = {env = Ident.empty;
                        types = Ident.empty;
                        submod = Ident.empty;
                        submod_types = Ident.empty}

end

