(***********************************************************************)
(*                                                                     *)
(*                         Fast Escape Analysis                        *)
(*                                                                     *)
(*                  Bruno Blanchet, ENS Ulm - INRIA Rocquencourt       *)
(*                  Alain Deutsch, INRIA Rocquencourt                  *)
(*                                                                     *)
(*  Copyright 1995 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

open Types
open Typedtree
open Format
open Asttypes
open L1solver
open L1env

exception Too_big
exception No_allocator

type inline_info_t = (int * L1env.expression) Ident.tbl

let inline_info = ref (Ident.empty : inline_info_t)

let indesc e desc = 
  {e_desc = desc;
   e_loc = e.e_loc;
   e_type = e.e_type;
   e_env = e.e_env;
   e_tail = e.e_tail;
   e_tailcalls = e.e_tailcalls;
   e_unknown_tailcalls = e.e_unknown_tailcalls;
   e_node = e.e_node;
   e_type_info = e.e_type_info;
   e_tye = e.e_tye;
   e_immop = e.e_immop;
   e_stackop = e.e_stackop;
   e_stackty = e.e_stackty;
   e_stacktailop = e.e_stacktailop;
   e_stacktailty = e.e_stacktailty
   }

(* Computes the size of an expression *)

let rec taille taillemax e = if taillemax = 0 then
   raise Too_big
else match e.e_desc with
   Eexp_ident(_,_) -> taillemax-1
 | Eexp_constant(_) -> taillemax-1
 | Eexp_let(_,_, bindings, body) -> taille (taillebind (taillemax-1) bindings) body
 | Eexp_function(clauses) -> taillebind (taillemax-1) clauses
 | Eexp_apply(rator, rands) -> taille (List.fold_left taille (taillemax-1) rands) rator
 | Eexp_tuple_apply(rator, rands) -> taille (List.fold_left taille (taillemax-1) rands) rator
 | Eexp_match(body, clauses) -> taille (taillebind (taillemax-1) clauses) body
 | Eexp_try(body, clauses) -> taille (taillebind (taillemax-1) clauses) body
 | Eexp_tuple(rands) -> List.fold_left taille (taillemax-1) rands
 | Eexp_construct(_, rands) -> List.fold_left taille (taillemax-1) rands
 | Eexp_record(fielddesc) -> let (_, rands) = List.split fielddesc in
      List.fold_left taille (taillemax-1) rands
 | Eexp_field(body, _) -> taille (taillemax-1) body
 | Eexp_setfield(lhs, _, rhs) -> taille (taille (taillemax-1) lhs) rhs
 | Eexp_array(rands) -> List.fold_left taille (taillemax-1) rands
 | Eexp_ifthenelse(e1, e2, e3) -> let t = taille (taille (taillemax-1) e1) e2 in
      (match e3 with
          None -> t
        | Some e -> taille t e)
 | Eexp_sequence(e1, e2) -> taille (taille (taillemax-1) e1) e2
 | Eexp_while(cond, body) -> taille (taille (taillemax-1) cond) body
 | Eexp_for(_, first, last, _, body) -> taille (taille (taille (taillemax-1) first) last) body
 | Eexp_when(cond, body) -> taille (taille (taillemax-1) cond) body 
 | Eexp_send(exp, _) -> taille (taillemax-1) exp
 | Eexp_new _ -> taillemax-1
 | Eexp_instvar _ -> taillemax-1
 | Eexp_setinstvar(_,_,exp) -> taille (taillemax-1) exp
 | Eexp_override(_, bindlist) -> let (_, explist) = List.split bindlist in
     List.fold_left taille (taillemax-1) explist

and taillepat taillemax pat = if taillemax = 0 then
   raise Too_big
else match pat.pat_desc with
   Tpat_any -> taillemax-1
 | Tpat_var(_) -> taillemax-1
 | Tpat_alias(pat, _) -> taillepat (taillemax-1) pat
 | Tpat_constant(_) -> taillemax-1
 | Tpat_tuple(patlist) ->  List.fold_left taillepat (taillemax-1) patlist
 | Tpat_construct(_, patlist) -> List.fold_left taillepat (taillemax-1) patlist
 | Tpat_record(fieldpat) -> let (_, patlist) = List.split fieldpat in
       List.fold_left taillepat (taillemax-1) patlist
 | Tpat_or(pat1, pat2) -> taillepat (taillepat (taillemax-1) pat1) pat2
   
and taillebind taillemax bindings = if taillemax = 0 then
   raise Too_big
else match bindings with
   [] -> taillemax
 | ((pat, body)::bindings') -> taillebind (taillepat (taille 
         (taillemax-1) body) pat) bindings'


(* Renames an expression *)

(* Careful ! After inlining, types are not correct any more :
   we should instanciate them to get the real types in the inlined
   expression. But it works anyway, because types are not much used
   in the rest of the compiler *)

let rec renameidents e stackcomp stacktailcomp table (exp : L1env.expression) =
(* e : the new containing expression becomes the one of e, so update the escape information
   stackcomp : does the containing expression for exp go beyond the expression it was,
       then it becomes the containing expression of e
   stacktailcomp idem for letstack' 
   table : new names of idents, to avoid having two different idents for the same name
   exp : the exp to rename *)
    let indesc desc = 
       {e_desc = desc; 
        e_loc = exp.e_loc; 
        e_type = exp.e_type;
        e_env = exp.e_env;
        e_tail = e.e_tail & ((not stackcomp) || exp.e_tail);
        e_tailcalls = exp.e_tailcalls;
	e_unknown_tailcalls = exp.e_unknown_tailcalls;
        e_node = exp.e_node; (* Not used by stackoptimizer. It was useful only
                                when building equations *)
        e_type_info = exp.e_type_info;
        e_tye = exp.e_tye;
        e_immop = exp.e_immop;
        e_stackop = if not stackcomp 
                    then exp.e_stackop 
                    else Store.compose exp.e_stackop e.e_stackop;
        e_stackty = if not stackcomp 
                    then exp.e_stackty else e.e_stackty;
        e_stacktailop = if not stacktailcomp 
                        then exp.e_stacktailop 
                        else Store.compose exp.e_stacktailop e.e_stacktailop;
        e_stacktailty = if not stackcomp 
                        then exp.e_stacktailty else e.e_stacktailty
   } in
         match exp.e_desc with
             Eexp_ident(Path.Pident i, v) ->
                (try
                   indesc (Eexp_ident(Path.Pident(Ident.find_same i table), v))
                with
                   Not_found -> exp)
           | Eexp_ident(_,_) -> indesc exp.e_desc
           | Eexp_constant(_) -> indesc exp.e_desc
           | Eexp_let(rec_flag, stack_flag, bindings, body) -> 
                if rec_flag = Recursive then
                   let (table', bindings') = createtablebindings table bindings
                   in indesc(Eexp_let(rec_flag, stack_flag, 
                      renamebindings e (stackcomp & body.e_tailcalls) stacktailcomp 
                         table' bindings', renameidents e stackcomp stacktailcomp table' body))
                else
                   let (table', bindings') = createtablebindings table bindings
                   in indesc(Eexp_let(rec_flag, stack_flag, 
                      renamebindings e (stackcomp & body.e_tailcalls) stacktailcomp 
                         table bindings', renameidents e stackcomp stacktailcomp table' body))
           | Eexp_function(clauses) -> indesc (Eexp_function(List.map (fun (p, e) ->
                 let (table', p') = renamepat table p in
                     (p', renameidents e false false table' e)) clauses))
           | Eexp_apply({e_desc = Eexp_ident(i,({val_kind=Val_prim
                 {Primitive.prim_name = prim_name}} as v))} as rator, rands)
                  when (prim_name = "%sequand" or prim_name = "%sequor") -> 
                  indesc(Eexp_apply(rator,
                 List.map (renameidents e false false table) rands))
           | Eexp_apply(rator, rands) -> 
                 indesc(Eexp_apply(renameidents e (stackcomp & exp.e_tailcalls) stacktailcomp table rator,
                 List.map (renameidents e (stackcomp & exp.e_tailcalls) stacktailcomp table) rands))
           | Eexp_tuple_apply(rator, rands) -> 
                 indesc(Eexp_tuple_apply(renameidents e (stackcomp & exp.e_tailcalls) stacktailcomp table rator,
                 List.map (renameidents e (stackcomp & exp.e_tailcalls) stacktailcomp table) rands))
           | Eexp_match(body, clauses) -> indesc(Eexp_match(renameidents e false false table body,
                 List.map (fun (p, e') ->
                 let (table', p') = renamepat table p in
                     (p', renameidents e false false table' e')) clauses))
           | Eexp_try(body,clauses) -> indesc(Eexp_try(renameidents e false false table body,
                 List.map (fun (p, e') ->
                 let (table', p') = renamepat table p in
                     (p', renameidents e false false table' e')) clauses))
           | Eexp_tuple(rands) -> 
                indesc(Eexp_tuple(List.map (renameidents e stackcomp stacktailcomp table) rands))
           | Eexp_construct(c, rands) -> 
                indesc(Eexp_construct(c, List.map 
                   (renameidents e stackcomp stacktailcomp table) rands))
           | Eexp_record(fieldlist) -> let (fl, el) = List.split fieldlist in
                indesc(Eexp_record(List.combine fl 
                   (List.map (renameidents e stackcomp stacktailcomp table) el)))
           | Eexp_field(f, lbl) -> indesc(Eexp_field(renameidents e stackcomp stacktailcomp table f, lbl))
           | Eexp_setfield(e1, lbl, e2) -> 
                indesc(Eexp_setfield(renameidents e stackcomp stacktailcomp table e1, lbl, 
                                     renameidents e false false table e2))
           | Eexp_array(rands) ->  
                indesc(Eexp_array(List.map (renameidents e stackcomp stacktailcomp table) rands))
           | Eexp_ifthenelse(cond, e1,e2) -> 
                indesc(Eexp_ifthenelse(
                  renameidents e false false table cond, 
                  renameidents e false false table e1, 
                  match e2 with
                  None -> None
                | Some e3 -> Some (renameidents e false false table e3)))
           | Eexp_sequence(e1, e2) -> indesc(Eexp_sequence(renameidents e false false table e1, 
                   renameidents e false false table e2))
           | Eexp_while(cond, body) -> indesc(Eexp_while(renameidents e false false table cond, 
                   renameidents e false false table body))
           | Eexp_for(i, first, last, dir, body) -> 
                  let x = Ident.create (Ident.name i) in
                  indesc(Eexp_for(x, renameidents e stackcomp stacktailcomp table first, 
                                     renameidents e stackcomp stacktailcomp table last, dir, 
                                     renameidents e false false (Ident.add i x table) body))
           | Eexp_when(cond, body) ->
                  indesc(Eexp_when(renameidents e false false table cond, 
                                   renameidents e false false table body))
	   | Eexp_send(exp, m) ->
                  indesc(Eexp_send(renameidents e false false table exp,m))
           | Eexp_new t -> exp
           | Eexp_instvar(n_self,n_field) -> exp
           | Eexp_setinstvar(n_self, n_field, exp) ->
                  indesc(Eexp_setinstvar(n_self, n_field, renameidents e 
                     false false table exp))
           | Eexp_override(n_self, modifs) ->
                  indesc(Eexp_override(n_self, List.map (fun (n_field, exp) 
                  -> (n_field, renameidents e false false table exp)) modifs))
          
 
and createtablebindings table = function [] -> (table, [])
    | ((p,e)::l) -> let (table', p') = renamepat table p in
             let (table'', bind') = createtablebindings table' l in
                 (table'', (p',e)::bind')

and renamebindings e stackcomp stacktailcomp table = List.map 
   (fun (p,ebind) -> (p, renameidents e stackcomp stacktailcomp table ebind))

and renamepat table pat =
    let indescpat p = {pat_desc = p; pat_loc = pat.pat_loc; pat_type = pat.pat_type} in
    match pat.pat_desc with
       Tpat_any -> (table, pat)
     | Tpat_var(i) -> let x = Ident.create (Ident.name i) in
          (Ident.add i x table, indescpat (Tpat_var x))
     | Tpat_alias(p, i) -> let (table', p') = renamepat table p in
          let x = Ident.create (Ident.name i) in
          (Ident.add i x table', indescpat (Tpat_alias(p',x)))
     | Tpat_constant(_) -> (table, pat)
     | Tpat_tuple(patlist) -> let (table', pl') = renamepatlist table patlist
          in (table', indescpat(Tpat_tuple(pl')))
     | Tpat_construct(c, patlist) -> let (table', pl') = renamepatlist table patlist in (table', indescpat(Tpat_construct(c, pl')))
     | Tpat_record(fieldlist) -> let (fl, pl) = List.split fieldlist in
         let (table', pl') = renamepatlist table pl in
         (table', indescpat(Tpat_record(List.combine fl pl')))
     | Tpat_or(_,_) -> (table, pat) (* An or pattern cannot declare variables *)

and renamepatlist table = function
   [] -> (table, [])
 | (p::l) -> let (table', p') = renamepat table p in
             let (table'', l') = renamepatlist table' l in
               (table'', p'::l')

let renamemain e patlist body =
    let (table, patlist') = renamepatlist Ident.empty patlist in
       (patlist', renameidents e true true table body)


(* For debug messages : to indicate if we inline a function in the program itself
   or in the inline information of a function *)

let current_function_name = ref (None : Ident.t option)

(* - parse_function :
   Checks if after application to n parameters the expression exp is an allocation.
   Inlining of functions of at most taillemax size is allowed.
   - expression_is_an_allocation :
   The same in the particular case n=0 *)

let rec parse_function taillemax n exp = 
  if n=0 then 
    ([], expression_is_an_allocation taillemax exp) 
          (* Notice that if this fails, the complete expression is left
             unchanged, although inlining would be interesting in this case :
             it would avoid closure creation. But it would not improve
             stack allocation. This could perhaps be changed *)
  else
    match exp.e_desc with
      Eexp_ident(Path.Pident i,v) -> 
        if !Control.inlining <= 0 then raise No_allocator;
        (try
          let (taille, expr) = Ident.find_same i (!inline_info) in
          if taillemax > taille then
            begin
              Message.debug 3 (fun () -> 
                print_string "Inlining function ";
                Printtyp.ident i;
                match !current_function_name with
                  None -> ()
                | Some name ->
                    print_string " in inline information for function ";
                    Printtyp.ident name);
              let (pat', body') = parse_function (taillemax-taille) n expr in
              if taillemax = !Control.inlining then
                renamemain exp pat' body'
              else
                (pat', body')
            end
          else
            begin
              Message.debug 4 (fun () ->
                print_string "Cannot inline ";
                Printtyp.ident i;
                (match !current_function_name with
                  None -> ()
                | Some name ->
                    print_string " in inline information for function ";
                    Printtyp.ident name);
                print_string " because too big");
              raise No_allocator
            end
        with Not_found -> raise No_allocator)
    |  Eexp_function([pattern,body]) ->
        let (patterns, body') = parse_function taillemax (n-1) body in
        (pattern :: patterns, body')
       (* If n=1 and the expression is fun pat1 -> body1 | pat2 -> body2 | ...
          it could be interesting to beta-reduce but this will not improve
          stack allocation, since no allocator can be displayed *)
    |  Eexp_let(rec_flag, stack_flag, bindings, body) ->
        let (patterns, body') = parse_function taillemax n body in
        (patterns, indesc body' (Eexp_let(rec_flag, stack_flag, bindings, body')))
    |  _ -> raise No_allocator

and expression_is_an_allocation taillemax e =
  if !Control.closure_only then
  begin
  match e.e_desc with
    Eexp_function(_) -> e
  | Eexp_apply({e_desc=Eexp_let(rec_flag, stack_flag, bindings, body)} as 
		 rator, rands) ->
      indesc e (Eexp_let(rec_flag, stack_flag, bindings, 
	    expression_is_an_allocation taillemax
	    (indesc e (Eexp_apply(body, rands)))))
  | Eexp_tuple_apply({e_desc=Eexp_let(rec_flag, stack_flag, bindings, body)} as
		       rator, rands) ->
     indesc e (Eexp_let(rec_flag, stack_flag, bindings, 
	   expression_is_an_allocation taillemax
           (indesc e (Eexp_tuple_apply(body, rands)))))
  | Eexp_apply(rator, rands) ->
            (try
              let (pat', body') = parse_function taillemax (List.length rands) rator in
              indesc e (Eexp_let(Nonrecursive, Anystack,
                                 List.rev (List.combine pat' rands), body'))
            with
              No_allocator ->
                match (Ctype.repr e.e_type).desc with
                  Tarrow(_,_) -> e
                | _ -> raise No_allocator)
                       (* When an application's result is a function,
                          the expression may allocate a closure 
                          (partial application). It may also not allocate
                          anything (if it gets an already existing closure) *) 
  | Eexp_tuple_apply(rator,rands) ->
      (try
        let ([pat'], body') = parse_function taillemax 1 rator
        in match pat'.pat_desc with
          Tpat_tuple(pat_list) ->
            indesc e (Eexp_let(Nonrecursive, Anystack,
                               List.rev (List.combine pat_list rands), body'))
      with
        No_allocator | Match_failure _ ->
          match (Ctype.repr e.e_type).desc with
            Tarrow(_,_) -> e
          | _ -> raise No_allocator)
  | Eexp_let(recflag, stackflag, bindings, body) -> 
      indesc e (Eexp_let(recflag, stackflag, bindings, 
			 expression_is_an_allocation taillemax body))
  | _ -> raise No_allocator
  end
  else    
  match e.e_desc with
    Eexp_function(_) -> e
  | Eexp_tuple(_) -> e
  | Eexp_construct(_,_) -> e
  | Eexp_record(_) -> e
  | Eexp_array(_) -> e
  | Eexp_apply({e_desc=Eexp_let(rec_flag, stack_flag, bindings, body)} as 
		 rator, rands) ->
      indesc e (Eexp_let(rec_flag, stack_flag, bindings, 
	    expression_is_an_allocation taillemax
	    (indesc e (Eexp_apply(body, rands)))))
  | Eexp_tuple_apply({e_desc=Eexp_let(rec_flag, stack_flag, bindings, body)} as
		       rator, rands) ->
     indesc e (Eexp_let(rec_flag, stack_flag, bindings, 
	   expression_is_an_allocation taillemax
           (indesc e (Eexp_tuple_apply(body, rands)))))
  | Eexp_apply(rator, rands) ->
      begin
        match rator.e_desc with
          Eexp_ident(_, {val_kind = Val_prim({Primitive.prim_name=
	     prim_name})}) when IdentInfo.is_an_allocation prim_name -> e
        | _ -> 
            (try
              let (pat', body') = parse_function taillemax (List.length rands) rator in
              indesc e (Eexp_let(Nonrecursive, Anystack,
                                 List.rev (List.combine pat' rands), body'))
            with
              No_allocator ->
		if !Clflags.stack_inline_threshold > 0 then e else
                match (Ctype.repr e.e_type).desc with
                  Tarrow(_,_) -> e
                | _ -> raise No_allocator)
                       (* When an application's result is a function,
                          the expression may allocate a closure 
                          (partial application). It may also not allocate
                          anything (if it gets an already existing closure) *) 
      end
  | Eexp_tuple_apply(rator,rands) ->
      (try
        let ([pat'], body') = parse_function taillemax 1 rator
        in match pat'.pat_desc with
          Tpat_tuple(pat_list) ->
            indesc e (Eexp_let(Nonrecursive, Anystack,
                               List.rev (List.combine pat_list rands), body'))
      with
        No_allocator | Match_failure _ ->
	  if !Clflags.stack_inline_threshold > 0 then e else
          match (Ctype.repr e.e_type).desc with
            Tarrow(_,_) -> e
          | _ -> raise No_allocator)
  | Eexp_let(recflag, stackflag, bindings, body) -> 
      indesc e (Eexp_let(recflag, stackflag, bindings, 
			 expression_is_an_allocation taillemax body))
  | _ -> raise No_allocator

let rec allocafterfun e =
   match e.e_desc with
      Eexp_function([pat,body]) -> indesc e (Eexp_function([pat, allocafterfun body]))
    | _ -> expression_is_an_allocation (!Control.inlining) e

let put_inline_info info (pat, e) =
   match (pat.pat_desc, e.e_desc) with
      (Tpat_var(x), Eexp_function([_, _])) -> 
            (try
              current_function_name := Some x;
              let e = allocafterfun e in
              current_function_name := None;
              let taille1 = !Control.inlining - (taille (!Control.inlining) e) in
              Message.debug 3 (fun () -> 
                   print_string "Function ";
                   Printtyp.ident x;
                   print_string " can be inlined (size : ";
                   print_int taille1;
                   print_string ")");
              Ident.add x  (taille1, e) info
            with
              Too_big -> Message.debug 4 (fun () -> 
                            print_string "Function ";
                            Printtyp.ident x;
                            print_string " cannot be inlined because too big (size > ";
                            print_int (!Control.inlining);
                            print_string ")");
                         info
            | No_allocator -> info)
    | _ -> info

let expression_value_maybe_heap_allocated e =
   match (Ctype.repr e.e_type).desc with
      Tvar -> true
    | Tarrow(_, _) -> true
    | Ttuple(types) -> true
    | Tconstr(path, [], _) when path = Predef.path_int -> false
    | Tconstr(path, [], _) when path = Predef.path_bool -> false
    | Tconstr(path, [], _) when path = Predef.path_char -> false
    | Tconstr(path, [], _) when path = Predef.path_unit -> false
    | Tconstr(path, [], _) when path = Predef.path_float -> !Control.stack_allocate_floats
    | Tconstr(_, _, _) -> true
    | Tfield _ | Tnil -> Message.error "expr_val_maybe_heap_alloc"
         (fun () -> print_string "Object field cannot be in place of types")
    | Tobject _ -> true
    | Tlink _ -> Message.error "expr_val_maybe_heap_alloc"
         (fun () -> print_string "Ctype.repr should have removed Tlink")

let expression_is_non_trivial e =
   match e.e_desc with
      Eexp_ident(_,_) -> false
    | Eexp_constant(_) -> false
    | _ -> true

let expression_cannot_escape e =
  let rec inner_most_closure = function
      {e_desc = Eexp_function [_,({e_desc = Eexp_function _} as e')]} -> 
	inner_most_closure e'
    | e -> e
  in
  let e' = inner_most_closure e in
    (  (* Is letstack' possible ? *)
    (Store.getvalue e.e_stacktailop e.e_stacktailty < e'.e_type_info),

       (* Is letstack possible ? *)
    (Store.getvalue e.e_stackop e.e_stackty < e'.e_type_info))

let expression_can_be_stack_allocated e =
   if expression_value_maybe_heap_allocated e &
      !Control.optimize_calls then
     let (may_stack', may_stack) = expression_cannot_escape(e) in
        if (may_stack' & (!Control.put_letstack_tail)) or may_stack then
           try
              let e' = expression_is_an_allocation (!Control.inlining) e in
                 (may_stack', may_stack, e')
           with
              No_allocator -> (false, false, e)
        else (false, false, e)
   else (false, false, e)

open CTree

let rec minstack = 
   function [] -> (true, true, top)
     | ((pattern, e)::l) -> 
          let (maystack, maystack', maxc2) = minstack l in
          let (maystack1,maystack1')= if expression_value_maybe_heap_allocated e &
               !Control.optimize_lets then
                  expression_cannot_escape e
             else
                 (false, false)
          and op = Store.to_const e.e_immop
          and tye = e.e_type_info
          in (maystack & maystack1, maystack' & maystack1',
              if op.c_infimp >= tye then
                 -1 (* No stack allocation *)
              else 
                 min (if op.c_supimp >= tye then - op.c_seuil else maxc2)
		     (* Perhaps the escape level must be <= seuil *)
                     (if op.c_func >= tye then tye-1 else maxc2)
		     (* Perhaps the escape level must be <=tye-1 *))

let set_stack may_stack may_stack' = 
       let stackflag = if may_stack' & (!Control.put_letstack_tail) 
          then StackTail else if may_stack then Stack else Nostack in
       if stackflag<>Nostack then
       begin
          incr(Control.promoted_lets_count);
          if (!Control.min_cntr <= !Control.promoted_lets_count)
           & (!Control.max_cntr >= !Control.promoted_lets_count)
           & not (List.mem !Control.promoted_lets_count !Control.exclude_cntr) then
              stackflag
          else
             Nostack
       end
       else Nostack

let rec make_blocks blocks body body_type body_env tyb =
        try
           let (op, (rec_flag, stack_flag, maxc, bindings), blocks) = 
             CTree.eval blocks in
             if bindings = [] then
                make_blocks (CTree.clearone blocks) body body_type body_env tyb
             else if (stack_flag = Nostack)
                  or (CTree.getvalue op tyb maxc)
             then
                make_blocks (CTree.clearone blocks) (
                Texp_let(rec_flag, (if stack_flag <> Nostack then Stack else
                         Nostack), bindings, 
                         {exp_desc= body;
                          exp_loc= Location.none;
                          exp_type= body_type;
                          exp_env= body_env})) body_type body_env tyb
             else (blocks, body)
       with Not_found -> ([], body)

let rec make_blocks_top blocks body body_type body_env =
        try
           let (_, (rec_flag, stack_flag, _, bindings), blocks) = 
             CTree.eval blocks in
             if bindings = [] then
                make_blocks_top (CTree.clearone blocks) body body_type body_env
             else
                make_blocks_top (CTree.clearone blocks) (
                Texp_let(rec_flag, stack_flag, bindings, 
                         {exp_desc= body;
                          exp_loc= Location.none;
                          exp_type= body_type;
                          exp_env= body_env})) body_type body_env
       with Not_found -> body

let to_trivial = Simplify.to_trivial

    let rec post_transform_expression e =
        let (blocks,d) = post_transform_desc e.e_desc in 
        let d' = make_blocks_top blocks d e.e_type e.e_env in
            {exp_desc = d';
             exp_loc  = e.e_loc;
             exp_type = e.e_type;
             exp_env = e.e_env}

    and post_transform e decompose =
	    let (blocks,d) = post_transform_desc e.e_desc in
            let (blocks,d') = 
                if decompose or not (!Control.move_lets) 
                then try
                        let (_, _, blocks) = CTree.eval blocks in
                                      (blocks,d)
                     with Not_found -> ([], d)
             (* Do path compression so that the test "snd block = LetGroup []"
                   really tells if the block is empty *)
                else make_blocks blocks d e.e_type e.e_env e.e_type_info
	    in
		(blocks,
		 {exp_desc = d';
		  exp_loc  = e.e_loc;
		  exp_type = e.e_type;
                  exp_env = e.e_env})

    and post_transform1 e =
        let (blocks, e') = post_transform e false in
            ((Store.to_const e.e_immop, CTree.LetGroup blocks), e')

    and post_transform_rands rands =
        let f (rands, blocks, decompose) rand =
            if expression_is_non_trivial(rand) then
                let (may_stack', may_stack, rand) = expression_can_be_stack_allocated rand in
		let immop = Store.to_const rand.e_immop in
                let tye = rand.e_type_info in
                let stackflag = set_stack may_stack may_stack' in
                if decompose or (stackflag <> Nostack) then
                begin
                   let (blocks',rand) = post_transform rand (stackflag <>
Nostack) in
                   let (binding, rand') = to_trivial rand in
                   let block1 = (immop, CTree.LetGroup blocks') in
                   let block2 = (immop, CTree.LetOne(Nonrecursive, stackflag,
                    tye-1, [binding])) in
                   (rand'::rands, block1::block2::blocks, true)
                end
                else
                      let (block, rand) = post_transform1 rand
                      in (rand :: rands, block :: blocks, snd block<>CTree.LetGroup [])
            else
               ((post_transform_expression rand) :: rands, blocks, decompose)
        in
            let (rands, blocks, decompose) = List.fold_left f ([], [], false) rands
            in
                (List.rev rands, blocks, decompose)

    and decompose_rands rands =
        let f (rands, blocks, decompose) rand =
            if expression_is_non_trivial(rand) then
                if decompose then
                begin
                   let immop = Store.to_const rand.e_immop in
                   let (blocks',rand') = post_transform rand false in
                   let (binding, rand') = to_trivial rand' in
                   let block1 = (immop, CTree.LetGroup blocks') in
                   let block2 = (immop, CTree.LetOne(Nonrecursive, Nostack,
                    rand.e_type_info-1, [binding])) in
                   (rand'::rands, block1::block2::blocks, true)
                end
                else
                      let (block, rand) = post_transform1 rand
                      in (rand :: rands, block :: blocks, snd block<>CTree.LetGroup [])
            else
               ((post_transform_expression rand) :: rands, blocks, decompose)
        in
            let (rands, blocks, decompose) = List.fold_left f ([], [], false) rands
            in
                (List.rev rands, blocks, decompose)

    and post_transform_desc d =
        match d with
            Eexp_ident(a, b) -> ([], Texp_ident(a,b))

          | Eexp_constant(c) -> ([], Texp_constant(c))

          | Eexp_let(rec_flag, stack_flag, bindings, body) ->
                let inlineinfo = !inline_info in
                     if (rec_flag = Nonrecursive) && (!Control.inlining > 0) then
                        List.iter (fun b -> inline_info := put_inline_info (!inline_info) b) bindings;
        	     let (blocks, body_desc') = post_transform_desc body.e_desc in
                     inline_info := inlineinfo;
                (* We don't put lets in the body of a let because it is 
                   equivalent to put them outside the let and if the containing
                   expression is the control expression, we will be able to
                   put a letstack', which would not be the case in the
                   body of the let *)
                let block_body = (CTree.opid, CTree.LetGroup blocks) in
                if rec_flag = Nonrecursive then
                begin
                let blocks = List.fold_left (fun blocks (p, e) ->
                   let (may_stack', may_stack, e) = 
                      expression_can_be_stack_allocated e in
                   let immop = Store.to_const e.e_immop in
                   let tye = e.e_type_info in
                   let stack_flag = if stack_flag <> Anystack then stack_flag
                                    else set_stack may_stack may_stack' in
                   let (blocks', e) = post_transform e (stack_flag <> Nostack) in
                   let block1 = (immop, CTree.LetGroup blocks') in
                   let block2 = (immop, CTree.LetOne (rec_flag, stack_flag,
                     tye-1, [p,e])) in
                    block2 :: block1 :: blocks
                   ) [] bindings in
                   (block_body :: blocks, body_desc')
                end
                else
                begin
                let (may_stack', may_stack, maxc) = minstack bindings in
                let stack_flag = if stack_flag <> Anystack then stack_flag 
                                 else set_stack may_stack may_stack' in
                let bindings = List.map
                    (fun (p,e) -> (p, post_transform_expression e)) bindings 
                in let block = (CTree.opid, CTree.LetOne (rec_flag, stack_flag, maxc, bindings)) in
		    ([block_body; block], body_desc')
                end

          | Eexp_function(bindings) -> 
                ([],
                 Texp_function(List.map (fun (p,e) -> (p, post_transform_expression e)) bindings))

          | Eexp_apply({e_desc = Eexp_ident(i,({val_kind=Val_prim{Primitive.prim_name = prim_name}} as v))} as rator, [arg1; arg2])
              when (prim_name = "%sequand" or prim_name = "%sequor") ->
            let arg1 = post_transform_expression arg1 in
            let arg2 = post_transform_expression arg2 in
                ([], Texp_apply({exp_desc = Texp_ident(i,v); 
                                 exp_loc = rator.e_loc;
                                 exp_type = rator.e_type;
                                 exp_env = rator.e_env}, [arg1; arg2]))

          | Eexp_apply({e_desc = Eexp_ident(i,({val_kind=Val_prim
               {Primitive.prim_name = prim_name;
                Primitive.prim_native_name = native_name}} as v))} as rator, rands)
              when (IdentInfo.box_float (if native_name<>"" then
                native_name else prim_name)) ->
            let (rands, blocks, _) = decompose_rands rands in
                (List.rev blocks, Texp_apply(
                   {exp_desc = Texp_ident(i,v); 
                    exp_loc = rator.e_loc;
                    exp_type = rator.e_type;
                    exp_env = rator.e_env}, rands))
	  
          | Eexp_apply(rator, rands) ->
            let (rands, blocks, decompose) = post_transform_rands rands in
            if expression_is_non_trivial(rator) then
                let (may_stack', may_stack, rator) = expression_can_be_stack_allocated rator in
		let immop = Store.to_const rator.e_immop in
		let tye = rator.e_type_info in
                let stackflag = set_stack may_stack may_stack' in
                if decompose or (stackflag <> Nostack) then
                begin
                   let (blocks',rator)  = post_transform rator (stackflag <> Nostack) in
                   let (binding, rator') = to_trivial rator in
                   let block1 = (immop, CTree.LetGroup blocks') in
                   let block2 = (immop, CTree.LetOne(Nonrecursive, stackflag,
                     tye-1,[binding])) in
                   (List.rev (block1::block2::blocks), Texp_apply(rator',
rands))
                end
                else
                   let (block, rator) = post_transform1 rator
                   in (List.rev (block :: blocks), Texp_apply(rator, rands))
            else
               (List.rev blocks, Texp_apply(post_transform_expression rator, rands))

          | Eexp_tuple_apply(rator,rands) -> 
            let (rands, blocks, decompose) = post_transform_rands rands in
            if expression_is_non_trivial(rator) then
                let (may_stack', may_stack, rator) = expression_can_be_stack_allocated rator in
		let immop = Store.to_const rator.e_immop in
		let tye = rator.e_type_info in
                let stackflag = set_stack may_stack may_stack' in
                if decompose or (stackflag <> Nostack) then
                begin
                   let (blocks',rator)  = post_transform rator (stackflag <> Nostack) in
                   let (binding, rator') = to_trivial rator in
                   let block1 = (immop, CTree.LetGroup blocks') in
                   let block2 = (immop, CTree.LetOne(Nonrecursive, stackflag,
                     tye-1,[binding])) in
                   (List.rev (block1::block2::blocks), Texp_apply(rator',
rands))
                end
                else
                   let (block, rator) = post_transform1 rator
                   in (List.rev (block :: blocks), Texp_apply(rator, rands))
            else
               (List.rev blocks, Texp_apply(post_transform_expression rator, 
                 [{exp_desc = Texp_tuple(rands);
                   exp_env = rator.e_env;
                   exp_loc = Location.none;
                   exp_type = {desc = Ttuple(List.map (fun r -> r.exp_type) rands);
                               level = -1}}]))

          | Eexp_match(dispatch_expr, clauses) ->
                let dispatch_expr = post_transform_expression dispatch_expr
                in
                    ([],
                     Texp_match(dispatch_expr,
                                List.map (fun (p,e) -> (p, post_transform_expression e)) clauses))

          | Eexp_try(dispatch_expr, clauses) ->
                let dispatch_expr = post_transform_expression dispatch_expr 
                in
                    ([],
                     Texp_try(dispatch_expr,
                                List.map (fun (p,e) -> (p, post_transform_expression e)) clauses))

          | Eexp_tuple(rands) ->
                let (rands, blocks, _) = post_transform_rands rands
                in
                    (List.rev blocks, Texp_tuple(rands)) 

          | Eexp_construct(x, rands) ->
                let (rands, blocks, _) = post_transform_rands rands
                in
                    (List.rev blocks, Texp_construct(x, rands)) 

          | Eexp_record(fields_inits) ->
                let (fields, inits) = List.split(fields_inits) in
                let (inits, blocks, _) = 
                   match fields with
                     ({Types.lbl_repres = Record_float}::_) ->
                         decompose_rands inits (* Unboxed floats *)
                   | _ -> post_transform_rands inits
                in
                    (List.rev blocks, Texp_record(List.combine fields inits)) 

          | Eexp_field(expr, x) ->
                let (block, expr)  = post_transform1 expr
                in
                    ([block], Texp_field(expr, x)) 

          | Eexp_setfield(expr1, x, expr2) ->
                let expr2  = post_transform_expression expr2 in
                let (block, expr1)  = post_transform1 expr1 in
                    ([block], Texp_setfield(expr1, x, expr2)) 

          | Eexp_array(rands) ->
                let (rands, blocks, _) = post_transform_rands rands
                in
                    (List.rev blocks, Texp_array(rands)) 

          | Eexp_ifthenelse(pred, consequent, alternative) ->
                let pred  = post_transform_expression pred in
                let consequent  = post_transform_expression consequent in
                let alternative  =
                	match alternative with
                	    None -> None
                	  | Some(e) -> Some(post_transform_expression e)
                in
                    ([], Texp_ifthenelse(pred, consequent, alternative))

          | Eexp_sequence(expr1, expr2) ->
                let expr1  = post_transform_expression expr1 in
                let expr2  = post_transform_expression expr2 
                in
                    ([], Texp_sequence(expr1, expr2))

          | Eexp_while(expr1, expr2) ->
                let expr1  = post_transform_expression expr1 in
                let expr2  = post_transform_expression expr2 in
                    ([], Texp_while(expr1, expr2))

	  | Eexp_for(ident, expr1, expr2, dir, body) ->
		let body = post_transform_expression body in
                let (block0, expr2)  = post_transform1 expr2 in
                if (expression_is_non_trivial(expr1)) &
                   (snd block0 <> CTree.LetGroup []) then
                begin
                   let immop = Store.to_const expr1.e_immop in
                   let (blocks',expr1)  = post_transform expr1 false in
                   let (binding, expr1') = to_trivial expr1 in
                   let block1 = (immop, CTree.LetGroup blocks') in
                   let block2 = (immop, CTree.LetOne(Nonrecursive, Nostack,0,
                      [binding])) in
                   ([block0; block2; block1], Texp_for(ident, expr1', expr2,
dir, body))
                end
                else
                   let (block, expr1) = post_transform1 expr1
                   in ([block0; block], Texp_for(ident, expr1, expr2, dir, body))

          | Eexp_when(expr1, expr2) ->
                let expr1  = post_transform_expression expr1 in
                let expr2  = post_transform_expression expr2
		in
                    ([], Texp_when(expr1, expr2))

          | Eexp_send(expr, m) ->
                ([], Texp_send(post_transform_expression expr, m))
          | Eexp_new(t) ->
                ([], Texp_new t)
          | Eexp_instvar(n_self, n_field) ->
                ([], Texp_instvar(n_self, n_field))
          | Eexp_setinstvar(n_self, n_field, expr) ->
                ([], Texp_setinstvar(n_self, n_field, 
                     post_transform_expression expr))
          | Eexp_override(n_self, modifs) ->
                ([], Texp_override(n_self, List.map (fun (n_field, expr) ->
                   (n_field, post_transform_expression expr)) modifs))

   let post_transform_expression = Profile.f1 "StackOptimizer.post_transform_expression" post_transform_expression

