(***********************************************************************)
(*                                                                     *)
(*                         Fast Escape Analysis                        *)
(*                                                                     *)
(*                  Bruno Blanchet, ENS Ulm - INRIA Rocquencourt       *)
(*                                                                     *)
(*  Copyright 1995 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

(* $Id: tanalysis.ml,v 1.26 2001/02/21 10:14:24 bblanche Exp $ *)

open Types
open Typedtree
open Format
open Asttypes
open Tsolver
open Ttype
open Tenv
open Tenvir
open Ttrtype

(* Analysis of patterns *)

let in_let_rec = ref false

(* TO DO : Avoir un seul environnement au lieu de env et var_env ? *)

let enrich_env_with_pattern boundinfo id_info arity tupled env pattern =
(* var_list -> NewType.t
            -> env -> Typedtree.pattern -> env *) 
   let rec enrich env id_info pattern = match pattern.pat_desc with
                 Tpat_any -> env
              |  Tpat_var(ident) -> 
                    if NewType.functional id_info then
                       let info = IdentInfo.build id_info tupled (!in_let_rec) arity in
                       boundinfo := info :: !boundinfo;
                       Envir.update ident info env
                    else env
              |  Tpat_alias(pat,ident) ->
                    if NewType.functional id_info then
                       let info = IdentInfo.build id_info tupled (!in_let_rec) arity in
                       boundinfo := info :: !boundinfo;
                       Envir.update ident info (enrich env id_info pat)
                    else enrich env id_info pat
              |  Tpat_constant _ -> env
              |  Tpat_tuple patlist -> 
                    List.fold_left2 enrich env 
                    (NewType.extract_list id_info patlist) patlist
              |  Tpat_construct(constr,patlist) -> 
                   List.fold_left2 enrich env 
                   (NewType.extract_constr_list id_info constr) patlist
              |  Tpat_record recpatlist -> 
	           let list_types = NewType.extract_list_n id_info in
                   List.fold_left
                   (fun env (field_desc,pat) -> enrich env 
                      (list_types field_desc.lbl_pos) pat) 
                   env recpatlist
              |  Tpat_or(_,_) -> env (* An or pattern cannot declare any variable *)
   in enrich env id_info pattern

let enrich_env_useless ty env pattern =
    NewType.iter_params Graph.mark ty;
    enrich_env_with_pattern (ref []) ty 0 false env pattern

let dum_enrich_env_with_pattern boundinfo arity tupled env pattern =
(* var_list -> env -> Typedtree.pattern -> env *) 
   let rec enrich env pattern = match pattern.pat_desc with
                 Tpat_any -> env
              |  Tpat_var(ident) -> 
                  let info = IdentInfo.build NewType.all tupled (!in_let_rec) arity in
                  boundinfo := info :: !boundinfo;
                  Envir.update ident info env
              |  Tpat_alias(pat,ident) ->
                  let info = IdentInfo.build NewType.all tupled (!in_let_rec) arity in
                  boundinfo := info :: !boundinfo;
                  Envir.update ident info (enrich env pat)
              |  Tpat_constant _ -> env
              |  Tpat_tuple patlist -> 
                    List.fold_left enrich env patlist
              |  Tpat_construct(constr,patlist) -> 
                   List.fold_left enrich env patlist
              |  Tpat_record recpatlist -> 
                   List.fold_left
                   (fun env (field_desc,pat) -> enrich env pat) 
                   env recpatlist
              |  Tpat_or(_,_) -> env (* An or pattern cannot declare any variable *)
   in enrich env pattern

let dum_enrich_env_useless env pattern =
    dum_enrich_env_with_pattern (ref []) 0 false env pattern

let rec enrich_var_env_with_pattern env analpar pattern =
(* var_env -> NewType.t -> Typedtree.pattern -> var_env *) 
           match pattern.pat_desc with
               Tpat_any -> env
            |  Tpat_var(ident) -> VarEnv.update ident analpar env
            |  Tpat_alias(pat,ident) -> VarEnv.update ident analpar
                 (enrich_var_env_with_pattern env analpar pat)
            |  Tpat_constant _ -> env
            |  Tpat_tuple patlist -> 
                 List.fold_left2 enrich_var_env_with_pattern
                 env (NewType.extract_list analpar patlist) patlist
            |  Tpat_construct(constr,patlist) -> 
                 List.fold_left2 enrich_var_env_with_pattern
                 env (NewType.extract_constr_list analpar constr)
                 patlist
            |  Tpat_record recpatlist -> 
      	         let list_types = NewType.extract_list_n analpar in
      	       	 List.fold_left
                  (fun env (field_desc,pat) -> enrich_var_env_with_pattern env
                     (list_types field_desc.lbl_pos) pat)
                  env recpatlist
            |  Tpat_or (_,_) -> env (* An or pattern cannot declare any variable *)

let rec set_unknown_tail_call boundinfo =
   List.iter (fun info -> info.unknown_tail_call <- true) !boundinfo

let end_scc boundinfo = 
   List.iter (fun info -> info.current_scc <- false) !boundinfo

let end_scc2 boundinfo = 
   List.iter (fun info -> info.current_scc2 <- false) !boundinfo

let noregister = fun _ -> ()

let update tail typ e = 
      e.e_stackop <- (snd typ);
      e.e_stacktailop <- if not tail then NewType.nostack else (snd typ)

let update_stack tail typ e =
      e.e_stacktailop <- if not tail then NewType.nostack else (snd typ)

let rec ty_arity ty = match ty.desc with
  Tarrow(_, r) -> 1 + ty_arity r
| _ -> 0

let rec tail_for_expression e env tail  =

(* gives back e' : expression
   e : Typedtree.expression
   env : Envir.t = Var --> IdentInfo.t
   tail : bool
 *)

let e = Simplify.simplify e in

let annotate_desc tailcalls unk_tailcalls d =
  {e_desc = d;
   e_loc = e.exp_loc;
   e_type = e.exp_type;
   e_env = e.exp_env;
   e_tail = tail;
   e_tailcalls = tailcalls;
   e_unknown_tailcalls = unk_tailcalls;
   e_type_info = NewType.all;
   e_stackop = NewType.nostack;
   e_stacktailop = NewType.nostack
   }

in

let equations_for_clauses clauses tail =
  List.map (fun (pattern,body) -> 
    let body = tail_for_expression body (dum_enrich_env_useless env pattern)
        tail in
    (pattern, body)) clauses

in

let get_func_info rator =(* find information about the function call *)
  match rator.exp_desc with
    Texp_ident(_,{val_type=val_type; val_kind=Val_prim _}) -> 
      (false, !Control.preserve_tail_calls, !Control.preserve_tail_calls, ty_arity val_type)
  | Texp_ident(path,_) ->
      (match Envir.lookup_ident path env with
        None -> (false, !Control.preserve_tail_calls 
		   or !Control.preserve_tail_recursion,
		 !Control.preserve_tail_calls 
		   or !Control.preserve_tail_recursion, 0)
      | Some(rator_info) -> 
	  (rator_info.tupled, !Control.preserve_tail_calls or 
	   ((rator_info.current_scc or
	     rator_info.unknown_tail_call) &
	    !Control.preserve_tail_recursion), 
	  !Control.preserve_tail_calls or 
	   (rator_info.unknown_tail_call &
	    !Control.preserve_tail_recursion), 
	  rator_info.arity))
  | _ -> (false, !Control.preserve_tail_calls
	    or !Control.preserve_tail_recursion,
	  !Control.preserve_tail_calls or !Control.preserve_tail_recursion, 0)

in

match e.exp_desc with
  Texp_ident(Path.Pident(name) as i , v) ->
     annotate_desc false false (Eexp_ident(i,v))

| Texp_ident(i,v) -> annotate_desc false false (Eexp_ident(i,v))

| Texp_constant(c) -> annotate_desc false false (Eexp_constant(c))

| Texp_let(rec_flag, stack_flag, bindings, body) ->
     let first_rec = (not (!in_let_rec)) & (rec_flag = Recursive) in
     if first_rec then in_let_rec := true;
     let boundinfo = ref [] in
     let newenv = List.fold_left
       (fun newenv (pattern,expression) ->
          dum_enrich_env_with_pattern boundinfo
	   (IdentInfo.arity_according_exp expression)
           (IdentInfo.tupled expression) newenv pattern)
       env bindings in
     let bindings' = 
     List.map (fun (pattern,expression) -> 
        let expression' =  tail_for_expression expression (if rec_flag = Recursive then newenv else env) false in
        if expression'.e_unknown_tailcalls then set_unknown_tail_call boundinfo;
        (pattern, expression')) bindings in
     if first_rec then 
       begin
	 in_let_rec := false;
	 end_scc boundinfo
       end;
     let inlinesupinfo = !Simplify.inline_sup_info in
     if (!Control.inlining_sup > 0) then
        List.iter (fun b -> Simplify.inline_sup_info :=
          Simplify.put_inline_info
          (!Simplify.inline_sup_info) rec_flag b) bindings;
     let body' = tail_for_expression body newenv tail in
     Simplify.inline_sup_info := inlinesupinfo;
     annotate_desc body'.e_tailcalls body'.e_unknown_tailcalls
       (Eexp_let(rec_flag, stack_flag, bindings', body'))

| Texp_function(clauses) ->
    let clauses = equations_for_clauses clauses true in
    annotate_desc
      (List.exists (fun (_,e) -> e.e_tailcalls) clauses)
      (List.exists (fun (_,e) -> e.e_unknown_tailcalls) clauses)
      (Eexp_function(clauses)) 

| Texp_apply({exp_desc = Texp_ident(i,({val_kind=Val_prim 
     {Primitive.prim_name = prim_name}} as v))} as rator, [arg1; arg2])
   when (prim_name = "%sequand" or prim_name = "%sequor") -> 
     let arg1 = tail_for_expression arg1 env false
     and arg2 = tail_for_expression arg2 env tail in
     (* arg1 and arg2 are of type bool, like the whole expression e *)
      annotate_desc arg2.e_tailcalls arg2.e_unknown_tailcalls (Eexp_apply(
  {e_desc = Eexp_ident(i,v);
   e_loc = rator.exp_loc;
   e_type = rator.exp_type;
   e_env = rator.exp_env;
   e_tail = false;
   e_tailcalls = false;
    e_unknown_tailcalls = false;
   e_type_info = NewType.all;
   e_stackop = NewType.nostack;
   e_stacktailop = NewType.nostack
   }, [arg1; arg2]))

| Texp_apply(rator,rands) ->
    let (tupled,rec_tail_call, unk_tail_call, arity) = get_func_info rator in
    let tail_call = tail & (rec_tail_call or (arity < List.length rands & !Control.preserve_tail_recursion)) in
    let unk_tail_call = tail & (unk_tail_call or (arity < List.length rands & !Control.preserve_tail_recursion)) in
    let rands' = List.map (fun param  ->
       tail_for_expression param env false) rands
    in 
    let rator' = tail_for_expression rator env false in
    annotate_desc tail_call unk_tail_call (
          if tupled then 
             match rands' with
                [{e_desc = Eexp_tuple(exp_list)}] ->
                   Eexp_tuple_apply(rator', exp_list)
              | _ -> Eexp_apply(rator', rands')
          else
             Eexp_apply(rator', rands'))

| Texp_match(expression, clauses) -> 
   let expression' = tail_for_expression expression env false
   and clauses' = equations_for_clauses clauses tail
   in annotate_desc
     (List.exists (fun (_,e) -> e.e_tailcalls) clauses')
     (List.exists (fun (_,e) -> e.e_unknown_tailcalls) clauses')
     (Eexp_match(expression', clauses'))

| Texp_try(expression, clauses) -> 
   let expression' = tail_for_expression expression env false
   and clauses' = equations_for_clauses clauses tail
   in annotate_desc 
     (List.exists (fun (_,e) -> e.e_tailcalls) clauses') 
     (List.exists (fun (_,e) -> e.e_unknown_tailcalls) clauses')
     (Eexp_try(expression', clauses'))

| Texp_tuple(args) -> 
   let args = List.map (fun arg -> tail_for_expression arg env false) args
   in annotate_desc false false (Eexp_tuple(args))

| Texp_construct(constructor_desc, args) -> 
   let args = List.map (fun arg -> tail_for_expression arg env false) args 
   in annotate_desc false false (Eexp_construct(constructor_desc, args))

| Texp_record(field_descriptions) -> 
   let field_descriptions = List.map 
          (fun (f, init) -> (f, tail_for_expression init env false)) 
          field_descriptions
   in annotate_desc false false (Eexp_record(field_descriptions))

| Texp_field(expr,desc) -> 
     let expr = tail_for_expression expr env false 
     in annotate_desc false false (Eexp_field(expr,desc))

| Texp_setfield(lhs,desc,rhs) ->
     let lhs = tail_for_expression lhs env false in
     let rhs = tail_for_expression rhs env false
     in annotate_desc false false (Eexp_setfield(lhs,desc,rhs))

| Texp_array(args) -> 
     let args = List.map (fun arg -> tail_for_expression arg env false) args
     in annotate_desc false false (Eexp_array(args))

| Texp_ifthenelse(predicate, consequent, alternative) ->
     let predicate = tail_for_expression predicate env false
     and consequent = tail_for_expression  consequent env tail
     and alternative = match alternative with None -> None | Some(e) -> 
        Some (tail_for_expression e env tail)
     in annotate_desc 
       (match alternative with
         None -> consequent.e_tailcalls
       | Some e -> consequent.e_tailcalls or e.e_tailcalls) 
       (match alternative with
         None -> consequent.e_unknown_tailcalls
       | Some e -> consequent.e_unknown_tailcalls or e.e_unknown_tailcalls) 
       (Eexp_ifthenelse(predicate, consequent, alternative))

| Texp_sequence(e1,e2) ->
     let e1 = tail_for_expression e1 env false
     and e2 = tail_for_expression e2 env tail
     in annotate_desc e2.e_tailcalls e2.e_unknown_tailcalls (Eexp_sequence(e1,e2))

| Texp_while(e1,e2) ->
     let e1 = tail_for_expression e1 env false
     and e2 = tail_for_expression e2 env false
     in annotate_desc false false (Eexp_while(e1,e2))

| Texp_for(ident,e1,e2,direction,body) ->
     let e1 = tail_for_expression e1 env false
     and e2 = tail_for_expression e2 env false
     and body = tail_for_expression body env false
     in annotate_desc false false (Eexp_for(ident,e1,e2,direction,body))

| Texp_when(cond,expr) ->
     let expr = tail_for_expression expr env false
     and cond = tail_for_expression cond env tail
     in annotate_desc cond.e_tailcalls cond.e_unknown_tailcalls (Eexp_when(cond,expr))

| Texp_send (expr, m) -> (* TO DO : Be more precise for objects *)
     let expr = tail_for_expression expr env false
     in annotate_desc false false (Eexp_send(expr,m))
| Texp_new t ->
     annotate_desc false false (Eexp_new t)
| Texp_instvar(n_self, n_field) ->
     annotate_desc false false (Eexp_instvar(n_self, n_field))
| Texp_setinstvar(n_self, n_field, expr) ->
     let expr = tail_for_expression expr env false
     in annotate_desc false false (Eexp_setinstvar(n_self, n_field, expr))
| Texp_override(n_self, modifs) ->
     annotate_desc false false (Eexp_override(n_self, 
        List.map (fun (n_field, expr) ->
         (n_field, tail_for_expression expr env false)) modifs))

(* Emit equations for escape analysis *)

let rec equations_for_expression e (tye : NewType.t) env var_env tail sublist sublist_stack =

(* 
   e : Tenv.expression
   tye : NewType.t_type_expr;
   env : Envir.t = Var --> IdentInfo.t
   var_env : VarEnv.t = Var --> NewType.t
   sublist : expression list ref
   sublist_stack : expression list ref
 *)

(*Message.debug 0 (fun () ->
	print_string "Mark ";
	print_int (Graph.get_mark (snd tye));
	print_string " ";
	Pretty.pr (stdout, Printtypedtree.print_expression e, 80));*)
Graph.incr_mark();

let annotate_desc() = e.e_type_info <- tye;

in

let fv gnode e =
let rec freevariables e = match e.e_desc with
  Eexp_ident(Path.Pident(name) as i ,v) ->
     (match VarEnv.lookup name var_env with
       None -> ()
     | Some tyv -> Graph.linksource (snd tyv) gnode)
| Eexp_ident(i,v) -> ()
| Eexp_constant(c) -> ()
| Eexp_let(rec_flag, stack_flag, bindings, body) ->
    List.iter (fun (pat, exp) -> freevariables exp) bindings;
    freevariables body
| Eexp_function(clauses) -> (* Could we do faster by not recomputing this ? *)
    List.iter (fun (pat,exp) -> freevariables exp) clauses
| Eexp_apply(rator,rands) ->
    List.iter freevariables rands;
    freevariables rator
| Eexp_tuple_apply(rator,rands) ->
    List.iter freevariables rands;
    freevariables rator
| Eexp_match(expression, clauses) -> 
    freevariables expression;
    List.iter (fun (pat,exp) -> freevariables exp) clauses
| Eexp_try(expression, clauses) -> 
    freevariables expression;
    List.iter (fun (pat,exp) -> freevariables exp) clauses
| Eexp_tuple(args) -> 
    List.iter freevariables args
| Eexp_construct(constructor_desc, args) -> 
    List.iter freevariables args
| Eexp_record(field_descriptions) -> 
    List.iter (fun (f,init) -> freevariables init) field_descriptions
| Eexp_field(expr,desc) -> 
    freevariables expr
| Eexp_setfield(lhs,desc,rhs) ->
    freevariables lhs;
    freevariables rhs
| Eexp_array(args) -> 
    List.iter freevariables args
| Eexp_ifthenelse(predicate, consequent, alternative) ->
    freevariables predicate;
    freevariables consequent;
    (match alternative with
      None -> ()
    | Some csq -> freevariables csq)
| Eexp_sequence(e1,e2) ->
    freevariables e1;
    freevariables e2
| Eexp_while(e1,e2) ->
    freevariables e1;
    freevariables e2
| Eexp_for(ident,e1,e2,direction,body) ->
    freevariables e1;
    freevariables e2;
    freevariables body
| Eexp_when(cond,expr) ->
    freevariables cond;
    freevariables expr
| Eexp_send(expr,_) -> freevariables expr
| Eexp_new _ -> ()
| Eexp_instvar _ -> ()
| Eexp_setinstvar(_,_,expr) -> freevariables expr
| Eexp_override(_,modifs) -> 
    List.iter (fun (_,expr) -> freevariables expr) modifs
in freevariables e

in

let linkafter e tye tail =
    let sublist = ref [] in
    let sublist_stack = ref [] in
    equations_for_expression e tye env var_env tail sublist sublist_stack;
    List.iter (update tail tye) !sublist;
    List.iter (update_stack tail tye) !sublist_stack

in

let equations_for_clauses clauses tye typat tail =
   List.iter (fun (pattern,body) -> 
      let sublist = ref [] in
      let sublist_stack = ref [] in
      equations_for_expression body tye 
         (enrich_env_useless typat env pattern)
         (enrich_var_env_with_pattern var_env typat pattern)
         tail sublist sublist_stack;
             (* I don't know what the parameter p1 does to his parameters
                assume the worst => they escape *)
      List.iter (update tail tye) !sublist;
      List.iter (update_stack tail tye) !sublist_stack
      ) clauses

in

let linkbefore e tye body_tailcalls =
        if body_tailcalls then
        begin
           let sublist' = ref [] in
           equations_for_expression e tye env var_env false sublist' sublist_stack;
           List.iter (fun subexpr -> 
              if !Control.move_lets then
                 subexpr.e_stackop <- snd tye;
              (* If we cannot move lets, letstack is simply forbidden
                 because of the tail call *)
              sublist_stack:= subexpr :: !sublist_stack) !sublist';
           (* e.e_stackop will be left NewType.nostack => no letstack *)
           sublist_stack:= e::!sublist_stack
        end
        else
        begin
           equations_for_expression e tye env var_env false sublist sublist_stack;
           sublist:= e::!sublist
        end

in

let inst_func_info (i,v) tyrator =(* find information about the function *)
  match (fst tyrator).b_desc with
    BTarrow _ ->
  let escapeinfo =
    match v with
         (* Normally, a primitive is declared in the .ml file, then its
            escape information is stored in the environment at the time
            of the declaration, but if the primitive is only declared in
            the .mli file, the escape information must be retrieved
            here *)
      { val_type=val_type;
        val_kind=Val_prim { Primitive.prim_name=prim_name;
			    Primitive.prim_native_name = prim_native_name}} -> 
           (** We should use prim_native_name when it is not empty **)
	   let prim_name = if prim_native_name = "" then prim_name else prim_native_name in
           (* The user cannot redefine a library primitive *)
           let val_type = TranslType.expr env val_type in
           IdentInfo.lookup_primitive prim_name val_type
    | _ -> 
        (match Envir.lookup_ident i env with
          None -> IdentInfo.empty tyrator (!in_let_rec)
        | Some(rator_info) -> rator_info)
  in IdentInfo.polymorphic_instantiation escapeinfo tyrator
  | _ -> ()
in

match e.e_desc with
  Eexp_ident(Path.Pident(name) as i , v) ->
     (match VarEnv.lookup name var_env with
       None -> ()
     | Some varinfo -> NewType.generalize tye varinfo);
     inst_func_info (i,v) tye;
     annotate_desc()

| Eexp_ident(i,v) -> 
    inst_func_info (i,v) tye;
    annotate_desc()

| Eexp_constant(c) -> annotate_desc()

| Eexp_let(rec_flag, stack_flag, bindings, body) ->
     let boundinfo = ref [] in
     let (newenv, newvar_env, typelist) = List.fold_left
       (fun (newenv, newvar_env, typelist) (pattern,expression) ->
          let typat = TranslType.expr env pattern.pat_type in
          (enrich_env_with_pattern boundinfo typat 
	     0 (IdentInfo.tupledE expression) newenv pattern,
           enrich_var_env_with_pattern newvar_env typat pattern,
           typat::typelist))
       (env,var_env,[]) bindings in
     equations_for_expression body tye newenv 
        newvar_env tail sublist sublist_stack;
     if rec_flag = Recursive then
     List.iter2 (fun (pattern,expression) tye -> 
        let sublist' = ref []
        and sublist_stack' = ref [] in
        equations_for_expression expression tye newenv 
           newvar_env false sublist' sublist_stack';
        List.iter (update false tye) !sublist';
        List.iter (update_stack false tye) !sublist_stack';
              (* sublist_stack' should be empty *)
        if body.e_tailcalls then
          sublist_stack := expression::!sublist_stack
              (* No letstack if there is an unknown tail call *)
        else
          sublist := expression::!sublist
				     ) bindings (List.rev typelist)
     else
       List.iter2 (fun (pattern, e') tye -> linkbefore e' tye body.e_tailcalls)
         bindings (List.rev typelist);
     annotate_desc ()

| Eexp_function(clauses) ->
    let (typar, tyres) = NewType.extract_fun tye in
    equations_for_clauses clauses tyres typar true;
    annotate_desc ();
    (* Mention that free variables escape *)
    (match clauses with
      [_, {e_desc = Eexp_function cl}] ->
	  (* The parameter escapes in the inner most closure *)
	let rec inner_most_closure ty =
	  function [_, {e_desc = Eexp_function cl}] ->
	      let (_, tyres) = NewType.extract_fun ty in
	      inner_most_closure tyres cl
	    | _ -> snd ty
	in
	Graph.linksource (snd typar) (inner_most_closure tyres cl)
	  (* It's enough to trace free variables in the inner most closure
	     when there is immediately imbricated closures *)
    | _ -> fv (snd tye) e)

| Eexp_apply({e_desc = Eexp_ident(i,({val_kind=Val_prim 
     {Primitive.prim_name = prim_name}} as v))} as rator, [arg1; arg2])
   when (prim_name = "%sequand" or prim_name = "%sequor") -> 
     linkafter arg1 tye false;
     linkafter arg2 tye tail;
     (* arg1 and arg2 are of type bool, like the whole expression e *)
     annotate_desc ()

| Eexp_apply(rator,rands) ->
    let tyrands = List.map (fun rand -> TranslType.expr env rand.e_type) rands in
    let tyrator = List.fold_right (fun (ty1, nt1) (ty2, nt2) ->
       (BuildType.annotate2 1 (BTarrow(ref 1, ref 0, ty1, ty2)),
	Graph.newnode (Expanded [| nt2 |]) (Some nt1))) tyrands tye in
    let tail_call = e.e_tailcalls in
    List.iter2 (fun param typaram ->
       linkbefore param typaram tail_call) rands tyrands;
    if tail then Graph.mark_tail (snd tyrator);
    if tail then NewType.iter_params Graph.mark_tail tyrator;
                    (* rator tail-escapes *)
    linkbefore rator tyrator tail_call;
    annotate_desc ()

| Eexp_tuple_apply(rator,rands) ->
    let rec getrand t =
      match t.desc with
	Tarrow(p, _) -> TranslType.expr env p
      |	Tlink t -> getrand t
    in
    let tyrand = getrand rator.e_type in
    let tyrands = NewType.extract_list tyrand rands in
(*    
    let tyrands = List.map (fun rand -> TranslType.expr env rand.e_type) rands in
    let n = ref 0 in
    let long = List.length tyrands in
    let tysons = List.map (fun ty -> 
	      incr n; 
	      (ref (!n), ty)) tyrands in
    let tytuplerand = BuildType.annotate2 long 
	(BTgroup(NewType.tuple_name, tysons)) in
    let ntsons = .... in
    let nttuplerand = Graph.newnode ntsons None in
*)
    let (typar, ntpar) = tyrand in
    let (tyres, ntres) = tye in
    let tyrator = (BuildType.annotate2 1
	(BTarrow(ref 1, ref 0, typar, tyres)),
	Graph.newnode (Expanded [| ntres |]) (Some ntpar)) in
    let tail_call = e.e_tailcalls in
    List.iter2 (fun param typaram ->
       linkbefore param typaram tail_call) rands tyrands;
    if tail then Graph.mark_tail (snd tyrator);
    if tail then NewType.iter_params Graph.mark_tail tyrator;
                    (* rator tail-escapes *)
    linkbefore rator tyrator tail_call;
    annotate_desc()

| Eexp_match(expression, clauses) -> 
   let tyexp = TranslType.expr env expression.e_type in
   linkafter expression tyexp false;
   equations_for_clauses clauses tye tyexp tail;
   annotate_desc()

| Eexp_try(expression, clauses) -> 
   linkafter expression tye false;
   equations_for_clauses clauses tye 
      (BuildType.annotate2 0 BTbadType, Graph.newnode (Expanded [| |]) None)
      (*Exceptions*) tail;
   annotate_desc()

| Eexp_tuple(args) -> 
   List.iter2
       (fun arg tyarg -> linkbefore arg tyarg false) 
       args (NewType.extract_list tye args);
   annotate_desc ()


| Eexp_construct(constructor_desc, args) -> 
   List.iter2 
     (fun arg tyarg -> linkbefore arg tyarg false) args 
     (NewType.extract_constr_list tye constructor_desc);
   annotate_desc ()

| Eexp_record(field_descriptions) -> 
   List.iter 
      (fun (f, init) -> linkbefore init 
	  (NewType.extract_list_n tye f.lbl_pos) false) 
      field_descriptions;
   annotate_desc ()

| Eexp_field(expr,desc) -> 
     let tyexpr = NewType.set_constr (TranslType.expr env expr.e_type)
       	       	   desc tye in
       (* tye = desc field / other fields do not escape *)
     linkbefore expr tyexpr false ;
     annotate_desc ()

| Eexp_setfield(lhs,desc,rhs) ->
     let tylhs = TranslType.expr env lhs.e_type in
     linkbefore lhs tylhs false;
     linkafter rhs NewType.all false;
     annotate_desc ()

| Eexp_array(args) -> 
     let tyarg = NewType.extract_list_n tye 0 in
     List.iter (fun arg -> linkbefore arg tyarg false) args;
     annotate_desc()

| Eexp_ifthenelse(predicate, consequent, alternative) ->
     let typred = TranslType.expr env predicate.e_type in
     linkafter predicate typred false;
     linkafter consequent tye tail;
     begin
       match alternative with None -> () | Some(e) -> 
        linkafter e tye tail
     end;
     annotate_desc()

| Eexp_sequence(e1,e2) ->
     let tye1 = TranslType.expr env e1.e_type in
     linkafter e1 tye1 false;
     linkafter e2 tye tail;
     annotate_desc ()

| Eexp_while(e1,e2) ->
     let tye1 = TranslType.expr env e1.e_type in
     let tye2 = TranslType.expr env e2.e_type in
     linkafter e1 tye1 false;
     linkafter e2 tye2 false;
     annotate_desc()

| Eexp_for(ident,e1,e2,direction,body) ->
     let tye1 = TranslType.expr env e1.e_type in
     let tybody = TranslType.expr env body.e_type in
     linkbefore e1 tye1 false;
     linkbefore e2 tye1 false;
     linkafter body tybody false;
     annotate_desc ()

| Eexp_when(cond,expr) ->
     let tycond = TranslType.expr env cond.e_type in
     linkafter expr tye false;
     linkafter cond tycond tail;
     annotate_desc()

| Eexp_send (expr, m) -> (* TO DO : Be more precise for objects *)
     linkafter expr NewType.all false;
     annotate_desc ()
| Eexp_new t ->
     annotate_desc()
| Eexp_instvar(n_self, n_field) ->
     annotate_desc()
| Eexp_setinstvar(n_self, n_field, expr) ->
     linkafter expr NewType.all false;
     annotate_desc()
| Eexp_override(n_self, modifs) ->
    List.iter (fun (n_field, expr) ->
      linkafter expr NewType.all false) modifs;
    annotate_desc()

let eq_for_exp e env var_env tye =
  let sublist = ref [] in
  let sublist_stack = ref [] in
  let e = tail_for_expression e env true in
  equations_for_expression e tye env var_env true sublist sublist_stack;
  List.iter (update true tye) !sublist;
  List.iter (update_stack true tye) !sublist_stack;
  e

let eq_for_exp = Profile.f4 "eq_for_exp" eq_for_exp

let optimize_definition rec_flag bindings env =
    if !Control.detailed_chrono then
       Message.phasestart "generating escape graph";
    let boundinfo = ref [] in
    if rec_flag = Recursive then in_let_rec := true;
    let (newenv, tylist) = List.fold_left
      (fun (newenv, tylist) (pattern,expression) ->
          let tye = TranslType.expr env expression.exp_type in
          (enrich_env_with_pattern boundinfo tye 
	     (IdentInfo.arity_according_exp expression)
             (IdentInfo.tupled expression) newenv pattern, 
             tye :: tylist)) (env, []) bindings in
    let bindings1 = List.map2  (fun (pattern,expression) tye -> (pattern, 
        eq_for_exp expression (if rec_flag = Recursive then newenv else env)
                   (VarEnv.create()) tye)) bindings (List.rev tylist) in
    if !Control.detailed_chrono then
       Message.phaseend "generating escape graph";
    end_scc boundinfo;
    end_scc2 boundinfo;
    if !Control.detailed_chrono then
       Message.phasestart "post_transform";
    let bindings = List.map (fun (pattern, e) -> (pattern, 
       Tstackoptimizer.post_transform_expression e)) bindings1 in
    if (rec_flag = Nonrecursive) && (!Control.inlining > 0) then
       List.iter (fun b -> Tstackoptimizer.inline_info := Tstackoptimizer.put_inline_info 
         (!Tstackoptimizer.inline_info) b) bindings1;
    if (!Control.inlining_sup > 0) then
       List.iter (fun b -> Simplify.inline_sup_info := Simplify.put_inline_info
         (!Simplify.inline_sup_info) rec_flag b) bindings;
    in_let_rec := false;
    if !Control.detailed_chrono then
       Message.phaseend "post_transform";
    if !Control.detailed_chrono then
       Message.phasestart "reducing escape graph";
    List.iter (fun id_info -> id_info.type_expr <- 
       Treduce.reduce id_info.type_expr) !boundinfo;
    if !Control.detailed_chrono then
       Message.phaseend "reducing escape graph";
    (bindings,newenv)

let optimize_expression e env =
    let wildcard = {pat_desc=Tpat_any; pat_type=e.exp_type; pat_loc=Location.none} in
    let bindings = [(wildcard, e)] in
    let ([(_, e)], env) = optimize_definition Asttypes.Nonrecursive bindings env
    in e

(* Scan and optimize a class declaration *)

let optimize_class_field env = function
   Cf_inher(path, explist, l1,l2) ->
     Cf_inher(path, List.map (fun e -> optimize_expression e env) explist,
              l1, l2)
 | Cf_val(n, i, p, expopt) ->
     Cf_val(n, i, p, match expopt with
         None -> None | Some exp -> Some(optimize_expression exp env))
 | Cf_meth(m, exp) ->
     Cf_meth(m, optimize_expression exp env)

let optimize_class env (name, def) =
   (name, {cl_args = def.cl_args;
           cl_field = List.map (optimize_class_field env) def.cl_field;
           cl_loc = def.cl_loc})

(* Register types defined in signatures in the environment *)

let rec scan_module_type s env =
  match s with
    Tmty_ident name -> 
	(* TO DO: Should probably make a copy of the signature with 
	   TranslType.copy_sig to avoid sharing and unexpected changes 
	   when applying functors. But I have found no program which 
	   exhibits this bug !!! *) 
	Envir.lookup_module_type name env
  | Tmty_signature sign -> Envir.get_hierarchy (scan_signature env sign)
  | Tmty_functor(name, sign, body) ->
    let typemap = !TranslType.typemap in
    let rand = scan_module_type sign (Envir.startmodule env) in
    let env = Envir.add_module name rand env in
       TranslType.typemap := typemap;
       Functor(rand, (scan_module_type body env))

and scan_signature env sign =
   let (ldecl, env) = List.fold_left scan_signature_item ([],env) sign
   in TranslType.decl env ldecl

and scan_signature_item (ldecl, env) = function
    Tsig_value(name,{val_type = ty}) -> 
       let env = TranslType.decl env ldecl in
       let ty = TranslType.expr env ty in
       NewType.iter_params Graph.mark ty;
       if NewType.functional ty then
          let info = 
             {type_expr = ty;
              current_scc = false;
	       current_scc2 = false;
              tupled = false;
              unknown_tail_call = true;
	      arity = NewType.arity ty}
          in ([], Envir.update name info env)
       else ([],env)
  | Tsig_type(name, decl) -> ((name,decl)::ldecl, env)
  | Tsig_exception(_,_) -> (ldecl, env)
  | Tsig_module(name, body) ->
       let env = TranslType.decl env ldecl in
       let typemap = !TranslType.typemap in
       let bodyenv = scan_module_type body (Envir.startmodule env) in
          TranslType.typemap := typemap;
          ([],Envir.add_module name bodyenv env)
  | Tsig_modtype(name, body) ->
       let env = TranslType.decl env ldecl in
       let typemap = !TranslType.typemap in
        let bodyenv = match body with
             Tmodtype_abstract -> Module Ttype.emptyhierarchy
           | Tmodtype_manifest modtype -> scan_module_type modtype (Envir.startmodule env) in
          TranslType.typemap := typemap;
          ([],Envir.add_module_type name bodyenv env)
  | Tsig_class(name,body) -> (ldecl, env) (* TO DO *)

(* Scan and optimize a structure *)

let rec optimize_structure (items: Typedtree.structure) (env: Envir.t) =
  let process_item (items, env) s = 
      let (s, env) = optimize_structure_item s env
      in (s::items, env)
  in
      let (items', env) = List.fold_left process_item ([],env) items
      in
          (List.rev items', env)

and optimize_structure_item (s: Typedtree.structure_item) (env: Envir.t) =
  match s with
    Tstr_eval(e) -> (Tstr_eval(optimize_expression e env), env)
  | Tstr_value(recursive, bindings) ->
        let (bindings', env) = optimize_definition recursive bindings env in
            (Tstr_value(recursive, bindings'), env)
  | Tstr_primitive(name,{val_type = valtype; val_kind = valkind}) -> 
        (s, match valkind with
              Val_reg | Val_ivar _ | Val_anc _ -> env
            | Val_prim {Primitive.prim_name=prim_name; 
                        Primitive.prim_native_name = prim_native_name} ->
           (** We should use prim_native_name when it is not empty **)
           let prim_name = if prim_native_name = "" then prim_name else prim_native_name in
           (* The user cannot redefine a library primitive *)
           let val_type = TranslType.expr env valtype in
           Envir.update name (IdentInfo.lookup_primitive prim_name val_type)
              env)
  | Tstr_type(decllist) -> (s, TranslType.decl env decllist)
  | Tstr_exception(_,_) -> (s, env)
  | Tstr_module(name, body) ->
    let typemap = !TranslType.typemap in
    let inlineinfo = !Tstackoptimizer.inline_info in
    let inlinesupinfo = !Simplify.inline_sup_info in
    let (body', bodyenv) = optimize_module_expr body (Envir.startmodule env) in
    let env = Envir.add_module name bodyenv env in
          TranslType.typemap := typemap;
          Tstackoptimizer.inline_info := inlineinfo;
          Simplify.inline_sup_info := inlinesupinfo;
          (Tstr_module(name, body'), env)
  | Tstr_modtype(name, body) -> 
    let typemap = !TranslType.typemap in
    let bodyenv = scan_module_type body (Envir.startmodule env) in
    let env = Envir.add_module_type name bodyenv env in
          TranslType.typemap := typemap;
          (s, env)
  | Tstr_open(_) -> (s, env)
  | Tstr_class(class_list) -> 
      (Tstr_class(List.map (optimize_class env) class_list),env)
      (* TO DO : Add the class to the environment *)

and optimize_module_expr {mod_desc=mod_desc; mod_loc=mod_loc;
   mod_type=mod_type; mod_env=mod_env} env =
    let (desc,env) = optimize_module_expr_desc mod_desc env in
    ({mod_desc=desc;
      mod_loc=mod_loc;
      mod_type=mod_type;
      mod_env=mod_env},
     env)

and optimize_module_expr_desc d (env: Envir.t) =
  match d with
    Tmod_ident(path) -> 
       (Tmod_ident(path),Envir.lookup_module path env)
  | Tmod_structure(str) ->
    let (str, env) = optimize_structure str env in
      (Tmod_structure(str),Envir.get_hierarchy env)
  | Tmod_functor(name, sign, body) ->
    let typemap = !TranslType.typemap in
    let rand = scan_module_type sign (Envir.startmodule env) in
    let env = Envir.add_module name rand env in
    TranslType.typemap := typemap;
    let (body, bodyenv) = optimize_module_expr body env in
      (Tmod_functor(name, sign, body),Functor(rand, bodyenv))
  | Tmod_apply(rator, rand, coercion) ->
    let (rator, ratorenv) = optimize_module_expr rator env in
    let (rand, randenv) = optimize_module_expr rand env in
    let (Functor(rand_var, body)) = ratorenv in
      (Tmod_apply(rator, rand, coercion),
       TranslType.instantiate rand_var randenv body)
  | Tmod_constraint(module_expr, sign, coercion) ->
    let (module_expr, env) = optimize_module_expr module_expr env in
      (Tmod_constraint(module_expr, sign, coercion),env)

open BuildType

let decl_empty_group (Path.Pident name) env =
   let name2 = {d_modname = ""; d_name = name} in
   Envir.add_type name
   {e_name = name2;
    e_type_params = [];
    e_color = 4;
    e_orig_decl = NoDecl;
    e_type_kind = annotate2 0 (BTgroup (name2, []))} env

let decl_exn (Path.Pident name) env =
   Envir.add_type name
   {e_name = {d_modname = ""; d_name = name};
    e_type_params = [];
    e_color = 4;
    e_orig_decl = NoDecl;
    e_type_kind = annotate2 0 BTbadType} env

let decl_array (Path.Pident name) env =
   let name2 = {d_modname = ""; d_name = name} in
   Envir.add_type name
   (let v = annotate2 0 BTvar in
   {e_name = name2;
    e_type_params = [v];
    e_color = 4;
    e_orig_decl = NoDecl;
    e_type_kind = annotate2 1 (BTgroup (name2, [ref 0,v]))}) env

let decl_list (Path.Pident name) env =
   let name2 = {d_modname = ""; d_name = name} in
   let v = annotate2 0 BTvar in
   let type_list = annotate2 1 BTbadType in
   type_list.b_desc <- BTgroup (name2, 
       [ref (-1), annotate2 0 (BTgroup(NewType.tuple_name, [ref 0,v; ref (-1),type_list]))]);
   Envir.add_type name
   {e_name = name2;
    e_type_params = [v];
    e_color = 4;
    e_orig_decl = NoDecl;
    e_type_kind = type_list} env

let decl_format (Path.Pident name) env =
   let name2 = {d_modname = ""; d_name = name} in
   let decl = {e_name = name2;
    e_type_params = [annotate2 0 BTvar; annotate2 0 BTvar; annotate2 0 BTvar];
    e_color = 4;
    e_orig_decl = NoDecl;
    e_type_kind = annotate2 0 BTbadType} in
   decl.e_type_kind.b_desc <- BTabstract (name2, decl, decl.e_type_params);
   Envir.add_type name decl env

let initial_env() =
    decl_empty_group Predef.path_int (
    decl_empty_group Predef.path_char (
    decl_empty_group Predef.path_string (
    decl_empty_group Predef.path_float (
  (* Constant constructors are omitted -> 
     we omit "true" and "false" for bool, "()" for unit, "[]" for lists *)
    decl_empty_group Predef.path_bool (
    decl_empty_group Predef.path_unit (
    decl_exn Predef.path_exn (
    decl_array Predef.path_array (
    decl_list Predef.path_list (
    decl_format Predef.path_format (
    Envir.create()))))))))))


let optimize(str: Typedtree.structure) =
    begin
	Compilenv.translator := fun modname sign ->
	    let oldmodname = !TranslType.modname in
	    TranslType.modname := modname;
	    let Module res = Envir.get_hierarchy (scan_signature (initial_env()) sign) in
        	TranslType.modname := oldmodname;
	        Allanal.fromT res
    end;
    Profile.start();
    let phase = "Escape analysis"
    and typemap = !TranslType.typemap 
    and inlineinfo = !Tstackoptimizer.inline_info
    and inlinesupinfo = !Simplify.inline_sup_info in
    Message.phasestart phase;
    let env = initial_env() in
    TranslType.modname := Compilenv.current_unit_name();
    let (str', env) = optimize_structure str env in
    Envir.set_escape_info env;
    (* Free useless data *)
    TranslType.typemap := typemap;
    Tstackoptimizer.inline_info := inlineinfo;
    Simplify.inline_sup_info := inlinesupinfo;
    Compilenv.empty_esc_info();
    Message.phaseend phase;
    Profile.stop();
    if !Control.show_environment then Envir.show_graphs env;
    if !Control.dump_stackallocation then
       Printtypedtree.print str';
    Message.debug 1 (fun () -> print_int(!Control.promoted_lets_count);
			       print_string " letstack introduced.\n";
      print_int !Graph.expand_count;
      print_string " expansions");
    str'
