(***********************************************************************)
(*                                                                     *)
(*                         Fast Escape Analysis                        *)
(*                                                                     *)
(*                  Bruno Blanchet, ENS Ulm - INRIA Rocquencourt       *)
(*                                                                     *)
(*  Copyright 1995 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

(* $Id: tenv.ml,v 1.62 2001/02/21 10:14:24 bblanche Exp $ *)

open Typedtree
open Format
open Asttypes
open Tsolver
open Ttype

module NewType : 
sig

  type t_type_name = Ttype.t_type_name

  val tuple_name : t_type_name

  type t = Ttype.newtype_t

val same_name : t_type_name -> t_type_name -> bool
val functional : t -> bool
val arity : t -> int

val extract_fun : t -> (t * t)
val extract_constr_list : t -> Types.constructor_description -> t list
val set_constr : t -> Types.label_description -> t -> t
val extract_list : t -> 'a list -> t list
val extract_list_n : t -> int -> t

val iter_params : (Graph.t -> unit) -> t -> unit
val result_mark : t -> int

val generalize : t -> t -> unit

val all : t (* Everything escapes *)
val nostack : graph_t (* Particular value to forbid stack allocation *)
val copy : int -> t -> t

end
=
struct

  type t_type_name = Ttype.t_type_name
  type t = Ttype.newtype_t

let tuple_name = Tprinttyp.tuple_name
let iter_params = Tprinttyp.iter_params
let get = Graph.get

let same_name decl1 decl2 =
   (decl1 == decl2) or (Ident.same decl1.d_name decl2.d_name &&
decl1.d_modname = decl2.d_modname)
(* This test is reliable, because an identifier is unique in a file,
thanks to its stamp. I add the == test because it is the most frequent case
and is very quick to test. *)


let nostack = Graph.nostack


let all_nt = Graph.newnode (Expanded [| |]) None
let _ = Graph.mark all_nt
let all = 
  ({b_desc = BTAll; b_mark = max_int; n_sons = 0; b_link = NoLink; b_link3 = []},
   all_nt)

let extract_fun ((ty, nodetree) as ty2) =
   match ty.b_desc with
      BTarrow(idxpar, idxdst, par,dst) ->
	((match nodetree.e_paramson with
	  None -> all
	| Some param -> (par, param)), (dst, get (!idxdst) nodetree))
    | BTAll -> (all,all)
    | BTbadType -> (ty2,ty2)
    | _ -> Misc.fatal_error "extract_fun"  (* Not a function type *)

(* Cf typing/datarepr.ml constructor_descrs *)

let extract_constr_list ((ty, nodetree) as ty2) constr = 
  match ty.b_desc with
    BTgroup(_, tylist) ->
      begin
         match constr.Types.cstr_tag with
           Types.Cstr_constant n -> []
         | Types.Cstr_block n -> 
             let (idx, variant) = List.nth tylist n in
	     let son = get (!idx) nodetree in
	     if constr.Types.cstr_arity = 1 then
               [variant, son] (*One parameter constructor*)
	     else
	     begin
             match variant.b_desc with
                BTgroup(_,lty) ->
  	          if List.length lty <> constr.Types.cstr_arity then
	            Misc.fatal_error "extract_constr_list 1" 
		      (* internal error - bad length *);
		  List.map (fun (idx, ty) -> (ty, get (!idx) son)) lty
              | BTAll -> Misc.replicate_list all constr.Types.cstr_arity
	      | BTbadType -> Misc.replicate_list ty2 constr.Types.cstr_arity
	      | _ -> Misc.fatal_error "extract_constr_list 2"
                       (* Not a variant type *)
	     end
	 | Types.Cstr_exception _ -> Misc.fatal_error "extract_constr_list 3"
	      (* Exceptions are handled by BadTypes *)
      end
  | BTAll -> Misc.replicate_list all constr.Types.cstr_arity
  | BTbadType -> Misc.replicate_list ty2 constr.Types.cstr_arity
  | _ -> Misc.fatal_error "extract_constr_list 4"
          (* Not a variant type *)

let extract_list ((ty, nodetree) as ty2) l = match ty.b_desc with
   BTgroup(_, tylist) -> 
     if List.length tylist <> List.length l then
       Misc.fatal_error "extract_list 1" 
	 (* internal error - bad length *);
     List.map (fun (idx, tyson) -> (tyson, get (!idx) nodetree)) tylist
 | BTAll -> List.map (fun _ -> all) l
 | BTbadType -> List.map (fun _ -> ty2) l
 | _ -> Misc.fatal_error "extract_list 2" 
      (* Not a group type *) 
 
let extract_list_n ((ty, nodetree) as ty2) n = 
  match ty.b_desc with
   BTgroup(_, tylist) -> 
     let (idx, elem) = List.nth tylist n in
     (elem, get (!idx) nodetree)
 | BTAll -> all
 | BTbadType -> ty2
 | _ -> Misc.fatal_error "extract_list_n" 
      (* Not a group type *)

let functional (ty, _) = match ty.b_desc with
     BTarrow _ | BTAll -> true
   | _ -> false

let arity (ty, _) = 
  let rec arity_rec ty = match ty.b_desc with
     BTarrow (_,_,t1, t2) -> 1 + arity_rec t2
   | _ -> 0
  in arity_rec ty

let result_mark (_, nodetree) =
  let rec result_markrec nodetree =
    match nodetree.e_paramson with
      None -> Graph.get_mark nodetree
    | Some param -> result_markrec (get 0 nodetree)
  in
  result_markrec nodetree

(**** Generalization ****)

open Tlinksons

let generalize (inst_ty, inst_nt) (gen_ty, gen_nt) =
  linksons gen_ty gen_nt inst_ty inst_nt
let generalize = Profile.f2 "NewType.generalize" generalize

(* cf typing/datarepr.ml label_descrs *)
let set_constr ((ty, nodetree) as ty2) lbl (elem_ty, elem_nt) =
  match ty.b_desc with
    BTgroup(name, tylist) -> 
      let (idx, elem) = List.nth tylist lbl.Types.lbl_pos in
      linksons elem (get (!idx) nodetree) elem_ty elem_nt;
      ty2
  | BTAll -> ty2
  | BTbadType -> ty2
  | _ -> Misc.fatal_error "set_constr" (* Not a record type *)

(***** Copy (useful for inlining) *****)

let copy incr (ty, nt) =
  (ty, Graph.copy incr nt)

end

module CTree :
sig

type letone = Asttypes.rec_flag * Asttypes.stack_flag * int *
(Typedtree.pattern * Typedtree.expression) list
type blocks = lettree list
and  lettree = LetGroup of blocks
             | LetOne of letone

val eval : blocks -> letone * blocks

val clearone : blocks -> blocks
(* clearone should come after eval *)

end =
struct
  
type letone = Asttypes.rec_flag * Asttypes.stack_flag * int *
(Typedtree.pattern * Typedtree.expression) list

type blocks = lettree list
and  lettree = LetGroup of blocks
             | LetOne of letone

let rec compress accu = 
   function LetOne t -> (LetOne t) :: accu
          | LetGroup [] -> accu
          | LetGroup(a::l) -> compress 
                (if l = [] then accu else (LetGroup l) :: accu) a

let rec eval =
   function [] -> raise Not_found
          | ((LetOne t)::l) as tree -> (t, tree)
          | (tree::l) -> eval (compress l tree)

let clearone = function (a::l) -> l
   (* clearone should come after eval *)

end

module VarEnv =
struct

  (* Analysis for all the free variables *)
  type t = NewType.t Ident.tbl

  let create() = Ident.empty

  let update ident value (env: t) = Ident.add ident value env

  let lookup x env = try Some(Ident.find_same x env) with Not_found -> None

end

type expression =
  { e_desc: expres_desc;
    e_loc: Location.t;
    e_type: Types.type_expr;
    e_env: Env.t;
    e_tail: bool;
    e_tailcalls: bool;
    e_unknown_tailcalls: bool;
    mutable e_type_info: NewType.t;
    mutable e_stackop: Graph.t;
    mutable e_stacktailop: Graph.t}

and expres_desc =
    Eexp_ident of Path.t * Types.value_description
  | Eexp_constant of constant
  | Eexp_let of rec_flag * stack_flag * (pattern * expression) list * expression
  | Eexp_function of (pattern * expression) list
  | Eexp_apply of expression * expression list
  | Eexp_tuple_apply of expression * expression list
  | Eexp_match of expression * (pattern * expression) list
  | Eexp_try of expression * (pattern * expression) list
  | Eexp_tuple of expression list
  | Eexp_construct of Types.constructor_description * expression list
  | Eexp_record of (Types.label_description * expression) list
  | Eexp_field of expression * Types.label_description
  | Eexp_setfield of expression * Types.label_description * expression
  | Eexp_array of expression list
  | Eexp_ifthenelse of expression * expression * expression option
  | Eexp_sequence of expression * expression
  | Eexp_while of expression * expression
  | Eexp_for of
      Ident.t * expression * expression * direction_flag * expression
  | Eexp_when of expression * expression
  | Eexp_send of expression * string
  | Eexp_new of Path.t
  | Eexp_instvar of Path.t * Path.t
  | Eexp_setinstvar of Path.t * Path.t * expression
  | Eexp_override of Path.t * (Path.t * expression) list

module IdentInfo =
struct

  (* Analysis for one function (all its parameters) *)
  type t = Ttype.identinfo_t

(* A function will be "Tupled" if
in all the clauses of the definition of the function its first parameter
is matched against a tuple, and we are not in the case of a function with
one clause, and a body which is still a function (in the last case, the
function is "Curried"). See bytecomp/translcore.ml.
   When a function is tupled, we compile it as if the members of the tuple
were passed one after the other, which avoids allocating the tuple. In this
case, we must not "letstack" the tuple (such a letstack would forbid the
tuple optimization).
*)

  let type_expr(x: t) = x.type_expr

  let empty ty in_let_rec =
         NewType.iter_params Graph.mark ty;
         { type_expr = ty;
           tupled = false;
           current_scc = in_let_rec;
	   current_scc2 = true;
           unknown_tail_call = true;
	   arity = 0}

  let arity_according_exp exp =
     let rec lpar_according_exp exp = match exp.exp_desc with
            Texp_function([(_,body)]) -> 
                  1 + (lpar_according_exp body)
         |  Texp_function(_) -> 1
         |  _ -> 0
     in match exp.exp_desc with
          Texp_function([_, {exp_desc = Texp_function pl}]) -> 
             lpar_according_exp exp
        | Texp_function((({pat_desc = Tpat_tuple pl}, _) :: _) as clauses) 
             when !Clflags.tupleopt -> 
              if List.for_all (function
		  ({pat_desc = Tpat_any | Tpat_tuple _},_) -> true
                | _ -> false) clauses then
                 1
              else
                 lpar_according_exp exp
        | _ -> lpar_according_exp exp

  let tupled exp =
     match exp.exp_desc with
          Texp_function([_, {exp_desc = Texp_function pl}]) -> 
             false
        | Texp_function((({pat_desc = Tpat_tuple pl}, _) :: _) as clauses) 
             when !Clflags.tupleopt -> 
              List.for_all (function
		  ({pat_desc = Tpat_any | Tpat_tuple _},_) -> true
                | _ -> false) clauses
        | _ -> false

  let tupledE exp =
     match exp.e_desc with
          Eexp_function([_, {e_desc = Eexp_function pl}]) -> 
             false
        | Eexp_function((({pat_desc = Tpat_tuple pl}, _) :: _) as clauses) 
             when !Clflags.tupleopt -> 
              List.for_all (function
		  ({pat_desc = Tpat_any | Tpat_tuple _},_) -> true
                | _ -> false) clauses
        | _ -> false

  let build ty tupled in_let_rec arity =
    { type_expr = ty;
      tupled = tupled;
      current_scc = in_let_rec;
      current_scc2 = true;
      unknown_tail_call = false;
      arity = arity}

  let print_info {type_expr = ty} =
      print_newline();
      Tprinttyp.print (NewType.result_mark ty) ty;
      print_newline();
      Tprinttyp.print 0 ty;
      print_newline()

  let show_graph id {type_expr = ty} =
      Tprinttyp.graphprint (Ident.unique_name id) ty

  let polymorphic_instantiation
        ({current_scc2 = cscc;
	  tupled = tupled;
          type_expr = generic_type} as id_info) instance_type =
    if instance_type == generic_type then () else
    begin
      if !Control.debug_polymorphic_instanciation then
	begin
          open_vbox 0;
          print_string "generic_type =";
          print_info id_info;
	  close_box()
	end;	  
      if cscc then
        Tcall.call instance_type generic_type
      else
        Tinstanc.instantiate instance_type generic_type;
      if !Control.debug_polymorphic_instanciation then
        begin
          open_vbox 0;
          print_string "instance_type =";
          print_info { type_expr = instance_type;
                       current_scc = false;
		       current_scc2 = false;
                       tupled = tupled;
                       unknown_tail_call = false;
		       arity = 0};
          close_box()
        end
    end

  let rec a n nt =              (* Parameter n Always escapes *)
     match nt.e_paramson with
       None -> ()
     | Some typar -> 
	 if n = 1 then Graph.mark typar
         else a (n-1) (Graph.get 0 nt)

  let rec param_closure = function
       {e_paramson = Some par} as nt ->
	 let res = Graph.get 0 nt in
	 if res.e_paramson != None then
	 begin
           (* The parameter escapes when the closure escapes *)
       	   Graph.linksource par res;
	   param_closure res
	 end
     | _ -> ()

  let lookup_primitive name ((_, nt) as ty) =
      begin
      match name with
          "%raise" -> a 1 nt
        | "%equal" | "%notequal" | "%lessthan" | "%greaterthan"
            | "%lessequal" | "%greaterequal" | "compare"
            | "%eq" | "%noteq" | "%sequand" | "%sequor"
            | "%addint" | "%subint" | "%mulint" | "%divint" 
            | "%modint" | "%andint" | "%orint" | "%xorint" 
            | "%lslint" | "%lsrint" | "%asrint"
            | "%addfloat" | "%subfloat" | "%mulfloat" | "%divfloat"
            | "pow" | "fmod" | "atan2" | "ldexp_float"
	    | "md5_chan" | "format_int" | "format_float"
	    | "obj_block" | "input_value_from_string"
 	    | "output_int" | "output_char" | "output_value" 
	    | "seek_out" | "seek_in" | "sys_rename"
            | "%boolnot" | "%negint" | "%succint" | "%predint"
            | "exp" | "log" | "sqrt" | "sin" | "cos" | "tan" | "asin" | "acos" 
            | "atan" | "sinh" | "cosh" | "%absfloat" | "log10" | "ceil"
            | "floor" | "frexp_float" | "tanh" | "modf" | "%negfloat"
            | "%floatofint" | "%intoffloat" | "ml_string_length"
            | "create_string" | "%string_length"
	    | "gc_stat" | "gc_get" | "gc_set" | "gc_minor" | "gc_major"
	    | "gc_full_major" | "output_value_to_string"
	    | "obj_is_block" | "obj_tag" | "%obj_size" | "open_descriptor"
	    | "input_int" | "input_char" | "input_value"
            | "input_scan_line" | "is_printable"
            | "int_of_string" | "float_of_string" | "flush" | "pos_out" 
            | "pos_in" | "channel_size" | "close_channel" 
            | "sys_exit" | "sys_get_config" | "sys_close" 
            | "%incr" | "%decr" | "%array_length"
 	    | "sys_get_argv" | "sys_file_exists" | "sys_remove" | "sys_getenv" 
	    | "sys_system_command" | "sys_chdir" | "sys_getcwd"
            | "blit_string"
	    | "%string_safe_get" | "%string_unsafe_get"
	    | "%string_safe_set" | "%string_unsafe_set"
	    | "fill_string" | "md5_string" | "hash_univ_param" -> ()
        | "%field0" -> (* [r] *)
	    begin
	      match nt with
		{e_paramson = Some param} -> 
		    (* fst : 'a * 'b -> 'a = "%field0" *)
		    (* (!): 'a ref -> 'a = "%field0" *)
                    Graph.link (Graph.get 0 param) (Graph.get 0 nt)
	      |	_ -> Message.error "lookup_primitive" (fun () ->
		  print_string  ("incorrect type for " ^ name))
	    end
	| "%field1" -> (* [r] *)
	    begin
	      match nt with
		(* snd : 'a * 'b -> 'b = "%field1" *)
		{e_paramson = Some param} ->
		  Graph.link (Graph.get 1 param) (Graph.get 0 nt)
	      |	_ -> Message.error "lookup_primitive" (fun () ->
		  print_string  ("incorrect type for " ^ name))
	    end	      
	| "%identity" -> a 1 nt
            (* We use this for unsafe operations in module obj *)
	| "%obj_field" -> (* [r;z] *)
	    begin
	      match nt with
		{e_paramson = Some param} -> 
		  Graph.link param (Graph.get 0 (Graph.get 0 nt))
	      |	_ -> Message.error "lookup_primitive" (fun () ->
		  print_string  ("incorrect type for " ^ name))
	    end
	| "%obj_set_field" -> a 3 nt
        | "sys_open" -> a 1 nt
        | "output" | "input" -> ()
        | "%makemutable" -> (* [r] *)
	    begin
	      match nt with
		{e_paramson = Some param} -> 
		  Graph.link param (Graph.get 0 (Graph.get 0 nt))
	      |	_ -> Message.error "lookup_primitive" (fun () ->
		  print_string  ("incorrect type for " ^ name))
	    end
        | "%setfield0" -> a 2 nt
        | "%array_safe_get" | "%array_unsafe_get" -> (*[r;z]*)
	    begin
	      match nt with
		{e_paramson = Some param} -> 
		  Graph.link (Graph.get 0 param) (Graph.get 0 (Graph.get 0 nt))
	      |	_ -> Message.error "lookup_primitive" (fun () ->
		  print_string  ("incorrect type for " ^ name))
	    end
        | "%array_safe_set" | "%array_unsafe_set" -> a 3 nt
        | "make_vect" -> (*[z;r]*)
	    begin
	      let res = Graph.get 0 nt in
	      match res.e_paramson with
		Some param ->
		  Graph.link param (Graph.get 0 (Graph.get 0 res))
	      |	_ -> Message.error "lookup_primitive" (fun () ->
		  print_string  ("incorrect type for " ^ name))
	    end	    
	| "install_signal_handler" -> a 2 nt
        | _ -> Message.debug 0 
                    (fun () -> print_string ("unknown primitive: " ^ name));
               NewType.iter_params Graph.mark ty
       end;
       param_closure nt;
       let info =
              {type_expr = ty;
               current_scc = false;
		current_scc2 = false;
               tupled = false;
               unknown_tail_call = false;
	       arity = NewType.arity ty}
       in if !Control.debug_primitives then
              begin
                  print_string (name ^ ": ");
                  print_info info
              end;
          info

  let box_float name = match name with
      "%negfloat" | "%addfloat" | "%subfloat" | "%mulfloat" | "%divfloat" |
      "%intoffloat" | "%absfloat" | "pow" | "frexp" | "floor" | "ceil" |
      "exp" | "log" | "sqrt" | "sin" | "cos" | "tan" | "asin" | "acos" |
      "atan" | "atan2" | "sinh" | "cosh" | "fmod" | "tanh" -> true
    | _ -> false

  let is_an_allocation name = match name with
      (* "float" primitives give back boxed floats, so they allocate *)
      "%negfloat" | "%addfloat" | "%subfloat" | "%mulfloat" | "%divfloat" |
      "%floatofint" | "%absfloat" | "%makemutable"| "make_vect" | "pow" |
      "exp" | "log" | "sqrt" | "sin" | "cos" | "tan" | "asin" | "acos" |
      "atan" | "atan2" | "sinh" | "cosh" | "tanh" | "floor" | "ceil" |
      "fmod" -> true
    | _ -> false

end

module BuildType =
struct

type t = Ttype.buildtype_t

let print (ty: t) =
  let rec annotate prio ty =
    if ty.b_mark = max_int then
      begin
	ty.b_mark <- 0;
	begin
          match (prio, ty.b_desc) with
            (_, BTgroup(decl,[])) -> 
	      Printtyp.ident decl.d_name
          | (_, BTgroup(decl,[idx,ty])) ->
              open_hovbox 0;
	      print_string "[";
	      print_int (!idx);
	      print_string "]";
	      annotate 0 ty;
	      print_string " ";
	      Printtyp.ident(decl.d_name);
              close_box()
          | (_, BTgroup(decl,types)) ->
              open_hovbox 0;
	      print_string "(";
	      Tprinttyp.print_with_delimitors (fun (idx, ty) -> 
	      	print_string "[";
	      	print_int (!idx);
	      	print_string "]";
		annotate 2 ty) ", " types;
	      print_string ") ";
	      Printtyp.ident(decl.d_name);
              close_box()
          | (_, BTvar) -> print_string "'a"
	  | (_, BTAll) -> print_string "All"
          | (_, BTabstract (decl, _, _)) ->
	      Printtyp.ident(decl.d_name);
          | (_, BTbadType) ->
              print_string "bad"
          | (0, _) ->
              (* This is not really a recursive call *)
              ty.b_mark <- max_int;
              open_hovbox 1;
	      print_string "(";
	      annotate 2 ty;
	      print_string ")";
              close_box()
          | (2, BTarrow(_, _, ty1, ty2)) ->
              open_hovbox 0;
	      annotate 1 ty1;
	      print_string " -> ";
	      annotate 2 ty2;
              close_box()
          | (_, _) -> ty.b_mark <- max_int;
              annotate (prio-1) ty
              (* This is not really a recursive call *)
        end;
        ty.b_mark <- max_int
      end
    else print_string "rec"
  in
  annotate 1 ty

let annotate desc =
  {b_desc = desc; b_mark = -1; n_sons = -1; b_link = NoLink; b_link3 = []}

let annotate2 n desc =
  {b_desc = desc; b_mark = max_int; n_sons = n; b_link = NoLink; b_link3 = []}

let rec copy_type1 ty =
    match ty.b_link with
       BTlink ty' -> ty' (* Forward  or back edge *)
     | _ -> let ty' = annotate BTbadType in
                  ty.b_link <- BTlink ty';
                  begin
                  match ty.b_desc with
                     BTvar | BTbadType | BTAll -> 
                        ty'.b_desc <- ty.b_desc
		   | BTabstract(d, decl, params) ->
		        ty'.b_desc <- BTabstract(d, decl, List.map copy_type1 params)
                   | BTarrow(paridx,dstidx,par,dst) -> 
                        ty'.b_desc <- BTarrow(ref (-2), ref(-2), copy_type1 par, copy_type1 dst)
                   | BTgroup(d_name, d_content) ->
                        ty'.b_desc <- BTgroup(d_name, List.map (fun (idx,ty) -> (ref(-2), copy_type1 ty))  d_content)
                  end;
                  ty'

let rec copy_type ty =
    match ty.b_link with
       BTlink ty' -> ty' (* Forward  or back edge *)
     | _ -> let ty' = {b_desc = BTbadType; b_mark = ty.b_mark; n_sons = 
		    ty.n_sons; b_link = NoLink; b_link3 = []} in
                  ty.b_link <- BTlink ty';
                  begin
                  match ty.b_desc with
                     BTvar | BTbadType | BTAll -> 
                        ty'.b_desc <- ty.b_desc
		   | BTabstract(d, decl, params) ->
		        ty'.b_desc <- BTabstract(d, decl, List.map copy_type params)
                   | BTarrow(paridx,dstidx,par,dst) -> 
                        ty'.b_desc <- BTarrow(paridx, dstidx, copy_type par, copy_type dst)
                   | BTgroup(d_name, d_content) ->
                        ty'.b_desc <- BTgroup(d_name, List.map (fun (idx,ty) -> (idx, copy_type ty))  d_content)
                  end;
                  ty'

let rec cleanup ty =
   match ty.b_link with
      NoLink -> () (* Forward or back edge *)
    | _ -> ty.b_link <- NoLink;
           match ty.b_desc with
                  BTvar | BTbadType | BTAll -> ()
		| BTabstract(_, _, params) -> List.iter cleanup params
                | BTarrow(_,_,par,dst) -> cleanup par; cleanup dst
                | BTgroup(_, d_content) -> List.iter (fun (_, ty) -> cleanup ty) d_content

let declsubstitute1 params args substtype =
   (* = substtype[params <- args] *)
     List.iter2 (fun p a -> p.b_link <- BTlink a) params args;
     let thetype = copy_type1 substtype in
        cleanup substtype;
	thetype

let declsubstitute params args substtype =
   (* = substtype[params <- args] *)
     List.iter2 (fun p a -> p.b_link <- BTlink a) params args;
     let thetype = copy_type substtype in
        cleanup substtype;
	thetype

end

