(***********************************************************************)
(*                                                                     *)
(*                         Fast Escape Analysis                        *)
(*                                                                     *)
(*                  Bruno Blanchet, ENS Ulm - INRIA Rocquencourt       *)
(*                                                                     *)
(*  Copyright 1995 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

open Format
open Ttype
open Tsolver
open Tprinttyp

(* Put to 0 the field b_mark of all types of the scc of ty *)

let get = Graph.get

let rec putmark ty =
  if ty.b_mark = max_int then
    begin
      ty.b_mark <- 0;
      match ty.b_desc with
	BTvar | BTabstract _ | BTbadType | BTAll -> ()
      |	BTarrow(_,idxdst1,_,dst1) ->
	  if !idxdst1 = -1 then putmark dst1
      |	BTgroup(_,tyl) ->
	  List.iter (fun (idx,ty) -> 
	    if !idx = -1 then putmark ty) tyl
    end

(* Put to 0 the field b_mark of the subtypes of generic_type if they
are differently unfolded than the corresponding subtypes of instance_type *)

let rec same_type generic_type instance_type =
  (instance_type == generic_type) ||
  match (generic_type.b_link3, instance_type.b_link3) with
    ([ty1'],_) when ty1' == instance_type -> true (* Back edge *)
  | (_, [ty2']) when ty2' == generic_type -> true (* Back edge *)
  | (_::_::_, _::_::_) -> Message.error "NewType.same_type" 
        (fun () ->
	  print_string "Wrong assumptions : ";
          print_sum instance_type;
          print_string " and ";
          print_sum generic_type;
          print_string "types do not correspond.")
  | _ -> 
      generic_type.b_link3 <- instance_type :: generic_type.b_link3;
      instance_type.b_link3 <- generic_type :: instance_type.b_link3;
      let res = 
       	match (generic_type.b_desc, instance_type.b_desc) with
	  ((BTvar | BTabstract _ | BTbadType | BTAll), (BTvar | BTabstract _ | BTbadType | BTAll)) -> true
        | (BTvar, _) | (BTabstract _, _) | (BTbadType, _) 
        | (_, BTbadType) | (BTAll, _) | (_, BTAll) -> true
    (*  | (_, BTvar) | (_, BTabstract _) -> Erreur ?? *)
        | (BTarrow(_,idxdst1,_,dst1), BTarrow(_,idxdst2,_,dst2)) -> 
	    let res = same_type dst1 dst2 in
	    if (!idxdst1 != -1) && not res then
	      putmark dst1;
            (!idxdst1 = !idxdst2) && res
        | (BTgroup(_, ty1), BTgroup(_, ty2)) ->
	    same_type_list ty1 ty2
        | (_, _) -> Message.error "NewType.same_type"
              (fun () -> 
            	print_sum instance_type;
            	print_string " and ";
            	print_sum generic_type;
		print_string "types do not correspond.")
      in
      generic_type.b_link3 <- List.tl generic_type.b_link3;
      instance_type.b_link3 <- List.tl instance_type.b_link3;
      res

and same_type_list gen_tyl inst_tyl = match (gen_tyl, inst_tyl) with
  ([],[]) -> true
| ((gen_idx,gen_ty)::gen_tyl, (inst_idx,inst_ty)::inst_tyl) ->
    let resl = same_type_list gen_tyl inst_tyl in
    let res = same_type gen_ty inst_ty in
    if (!gen_idx != -1) && not res then
      putmark gen_ty;
    (!gen_idx = !inst_idx) && res && resl

(* Cleanup the b_mark field *)

let rec cleanup ty =
  if ty.b_mark = 0 then
    begin
      ty.b_mark <- max_int;
      match ty.b_desc with
	BTvar | BTabstract _ | BTbadType | BTAll -> ()
      |	BTarrow(_, _, _, dst) -> cleanup dst
      |	BTgroup(_,tyl) -> List.iter (fun (_, ty) -> cleanup ty) tyl
    end

(* Links the corresponding sons of generic_type and instance_type *)

let rec linksons_rec generic_type gen_nt instance_type inst_nt  =
  if (instance_type == generic_type) || (generic_type.b_mark = max_int) then 
    Graph.linksons gen_nt inst_nt
    (* to avoid adding the type twice in b_link3 *)
  else
    match (generic_type.b_link3, instance_type.b_link3) with
      ([ty1'],_) when ty1' == instance_type -> () (* Back edge *)
    | (_, [ty2']) when ty2' == generic_type -> () (* Back edge *)
    | (_::_::_, _::_::_) -> Message.error "NewType.linksons" 
          (fun () -> 
	    print_string "Wrong assumptions : ";
            print_sum instance_type;
            print_string " and ";
            print_sum generic_type;
            print_string " do not correspond.")
    | _ -> 
       	generic_type.b_link3 <- instance_type :: generic_type.b_link3;
       	instance_type.b_link3 <- generic_type :: instance_type.b_link3;
       	begin
	  match (generic_type.b_desc, instance_type.b_desc) with
            (BTvar, _) | (BTabstract _, _) | (BTbadType, _) 
          | (_, BTbadType) -> Graph.linksons gen_nt inst_nt
          | (BTAll, _) -> ()
          | (_, BTAll) -> Graph.mark gen_nt
      (*  | (_, BTvar) | (_, BTabstract _) -> Erreur ?? *)
          | (BTarrow(_,idxdst1,_,dst1), BTarrow(_,idxdst2,_,dst2)) -> 
              Graph.link gen_nt inst_nt;
              linksons_rec dst1 (get (!idxdst1) gen_nt) dst2 (get (!idxdst2) inst_nt)
          | (BTgroup(_, ty1), BTgroup(_, ty2)) ->
              Graph.link gen_nt inst_nt;
              List.iter2 (fun (idx_gen, ty_gen) (idx_inst, ty_inst) ->
	       	linksons_rec ty_gen (get (!idx_gen) gen_nt) ty_inst
		  (get (!idx_inst) inst_nt) 
		  ) ty1 ty2
          | (_, _) -> Message.error "NewType.linksons"
                (fun () -> 
		  print_sum instance_type;
                  print_string " and ";
                  print_sum generic_type;
                  print_string " do not correspond.")
       	end;
       	generic_type.b_link3 <- List.tl generic_type.b_link3;
       	instance_type.b_link3 <- List.tl instance_type.b_link3

(* Careful ! Before calling this function, gen_ty and inst_ty must have
   empty b_link3 field, and gen_ty must have its b_mark == max_int *)

let linksons gen_ty gen_nt inst_ty inst_nt = 
  if not (same_type gen_ty inst_ty) then
    putmark gen_ty;
  linksons_rec gen_ty gen_nt inst_ty inst_nt;
  cleanup gen_ty

