(***********************************************************************)
(*                                                                     *)
(*                    Fast Escape Analysis - Control                   *)
(*                                                                     *)
(*                  Bruno Blanchet, ENS Ulm - INRIA Rocquencourt       *)
(*                  Alain Deutsch, INRIA Rocquencourt                  *)
(*                                                                     *)
(*  Copyright 1995 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

    type option_kind =
	Boolean of bool ref
      | Numeric of int ref

    let flags = ref [
      "stack", Numeric(Clflags.optimize_stackallocation);
      "tupleopt", Boolean(Clflags.tupleopt);
      "debug_level", Numeric(Message.debug_level);
      "intermod_inlining", Numeric(Clflags.stack_inline_threshold)]

    let exclude_cntr = ref ([] : int list)

    let make_flag name init =
        let cell = ref init in
        flags := (name, Boolean(cell)) :: !flags;
        cell

    let make_counter name init =
        let cell = ref init in
        flags := (name, Numeric(cell)) :: !flags;
        cell

    let parse_option (x: string) =

        let prefix p x tk fk =
            let lp = String.length p
            and lx = String.length x in
            if lx >= lp then
               if (String.sub x 0 lp) = p then
                  tk(String.sub x lp (lx - lp))
               else
                  fk(x)
            else
               fk(x) in

        let ff(x) = (x, false) in
        let tt(x) = (x, true) in
        let id(x) = (x, true) in

        let (x, truth) =
            prefix "-no" x ff (fun x -> prefix "-" x tt id) in

	let parse_numeric flag_name argument =
            let lf = String.length flag_name
            and la = String.length argument in
	    if la >= lf then
		let prefix_of_argument = String.sub argument 0 lf
		in
		    if (prefix_of_argument = flag_name) then
                    try
			Some(int_of_string(String.sub argument lf (la - lf)))
                    with Failure("int_of_string") -> None
		    else
			None
	    else
		None in

        let rec assign = function
            [] -> Message.debug 0 (fun () -> 
                 print_string "Unknown stack optimizer option ";
                 print_string x;
                 print_string "\n")
          | (a, Boolean(cell))::l ->
             if (a = x) then cell := truth else assign l
          | (a, Numeric(cell))::l ->
	    (match parse_numeric a x with
		None -> assign l
	      | Some(value) -> cell := value)
        in
            match parse_numeric "exclude_cntr" x with
               None -> assign !flags
             | Some n -> exclude_cntr := n :: !exclude_cntr

    let dump_stackallocation = make_flag "display" false
    let debug_primitives = make_flag "debug_primitives" false
    let debug_polymorphic_instanciation = make_flag "debug_polymorphic_instanciation" false
    let show_valuation = make_flag "show_valuation" false
    let show_equations = make_flag "show_equations" false
    let show_environment = make_flag "show_environment" false
    let stack_allocate_floats = make_flag "stack_allocate_floats" true
    let optimize_lets = make_flag "optimize_lets" true
    let optimize_calls = make_flag "optimize_calls" true
    let preserve_tail_recursion = make_flag "preserve_tail_recursion" true
    let preserve_tail_calls = make_flag "preserve_tail_calls" false
    let put_letstack_tail = make_flag "put_letstack_tail" true
    let move_lets = make_flag "move_lets" true
    let max_stack_size = make_flag "max_stack_size" false
    let detailed_chrono = make_flag "detailed_chrono" false
    let chrono_phases = make_flag "chrono_phases" false
    let closure_only = make_flag "closure_only" false
    let profile = make_flag "profile" false
    let let_only = make_flag "let_only" false
    let inlining = make_counter "inlining" 0
    let inlining_sup = make_counter "inlining_sup" 0

    let min_cntr = make_counter "min_cntr" (-1)
    let max_cntr = make_counter "max_cntr" 32767

    let promoted_lets_count = ref 0
    let stackalloc_count = ref 0

let split_words s = 
    let long = String.length s in
    let rec split cur finmot l = 
       (* cur = position courante
          finmot = position de l'espace suivant *)
       if cur = 0 then 
       begin
          if s.[cur] = ' ' then
            (* Espace en debut de commande *)
            if cur+1 = finmot then
               (* Plusieurs espaces en debut de commande *)
               l
            else
               (String.sub s 1 (finmot-1)) :: l
          else
            (String.sub s 0 finmot)::l
       end
       else 
       begin
          if s.[cur] = ' ' then
             if cur+1=finmot then
                (* Cas de deux espaces consecutifs *)
                split (cur-1) cur l
             else 
                split (cur-1) cur ((String.sub s (cur+1) (finmot-cur-1))::l) 
          else 
             split (cur-1) finmot l
       end
    in if long=0 then [] else split (long-1) long []

let _ = 
   try
      List.iter parse_option (split_words(Sys.getenv "CAML_OPTF"))
   with Not_found -> ()
