(***********************************************************************)
(*                                                                     *)
(*                         Fast Escape Analysis                        *)
(*                                                                     *)
(*                  Alain Deutsch, INRIA Rocquencourt                  *)
(*                                                                     *)
(*  Copyright 1995 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

(* $Id: printtypedtree.ml,v 1.1 1998/02/15 18:16:08 blanchet Exp $ *)

(* Abstract interpretation aimed at replacing unlimited extend allocations
    with lifetime-explicit, stack allocations *)

open Types
open Typedtree
open Format
open Asttypes

let rec print_with_delimitors element separator list =
    match list with
      	[] -> []
      | [x] -> [element(x)]
      | (x::xs) -> (element(x) :: separator) @ (print_with_delimitors element separator xs)

let rec equals ty1 ty2 =
    match ((Ctype.repr ty1).desc, (Ctype.repr ty2).desc) with
	(Tvar, Tvar) -> true
      | (Tarrow(t1, t1'), Tarrow(t2, t2')) ->
	   (equals t1 t2) & (equals t1' t2')
      | ((Ttuple tl1), Ttuple tl2) ->
	   List.for_all2 equals tl1 tl2
      | (Tconstr(p1, tl1, _), Tconstr(p2, tl2, _)) ->
	   (p1 = p2) & List.for_all2 equals tl1 tl2
      | (Tnil, Tnil) -> true
      | (Tobject(t1, _), Tobject(t2, _)) ->
	   equals t1 t2 (* NOT SAFE *)
      | (Tfield(s1, t1, t1'), Tfield(s2, t2, t2')) ->
	   (s1 = s2) & (equals t1 t2) & (equals t1' t2') 
      | (_,_) -> false

let rec print_path p =
    match p with
      Path.Pident(i) -> Pretty.str(Ident.name i)
    | Path.Pdot(p, s, n) -> Pretty.blo(0, [print_path p; Pretty.str ".";
					   Pretty.str s])
    | Path.Papply(p1,p2) -> Pretty.blo(0, [print_path p1; Pretty.str "("; 
                                           print_path p2; Pretty.str ")"])

(* Print a type expression *)

let var_names = ref ([] : (type_desc * string) list)
let var_counter = ref 0

let reset_var_names () = var_names := []; var_counter := 0

let name_of_var v =
  try
    List.assq v !var_names
  with Not_found ->
    let name = 
      if !var_counter < 26
      then String.make 1 (Char.chr(97 + !var_counter)) 
      else String.make 1 (Char.chr(97 + !var_counter mod 26)) ^
           string_of_int(!var_counter / 26) in
    var_names := (v, name) :: !var_names;
    incr var_counter;
    name

let rec typexp sch prio ty = match ty.desc with
    Tvar as v ->
	let prefix = 
	    if (not sch) or ty.level = -1 (* generic *) then "'"
	    else "'_"
	in
	    Pretty.str(prefix ^ (name_of_var v))
  | Tarrow(ty1, ty2) ->
	let brackets l =
	    if prio >= 1 then 
		[Pretty.str "("] @ l @ [Pretty.str ")"]
	    else
		l
	in
	    Pretty.blo((if prio >= 1 then 1 else 0),
		       brackets[typexp sch 1 ty1;
				Pretty.str " ->";
				Pretty.brk 1;
				typexp sch 0 ty2])
  | Ttuple tyl ->
	let brackets l =
	    if prio >= 2 then 
		[Pretty.str "("] @ l @ [Pretty.str ")"]
	    else
		l
	in
	    Pretty.blo((if prio >= 2 then 1 else 0),
		       brackets(typlist sch 2 " *" tyl))
  | Tconstr(p, tyl, _) ->
	Pretty.blo(0,
		   (match tyl with
			[] -> []
		      | [ty1] -> [typexp sch 2 ty1; Pretty.str " "]
		      | tyl -> [Pretty.blo(1, [Pretty.str "("] @ (typlist sch 0 "," tyl) @ [Pretty.str ")"]); Pretty.str " "])
		   @ [print_path p])
  | Tobject _ -> Pretty.str "Tobject"
  | Tfield _ -> Pretty.str "Tfield"
  | Tnil -> Pretty.str "Tnil"
  | Tlink ty -> typexp sch prio ty

and typlist sch prio sep = function
    [] -> []
  | [ty] -> [typexp sch prio ty]
  | ty::tyl -> [typexp sch prio ty;
		Pretty.str sep;
		Pretty.brk 1] @ (typlist sch prio sep tyl)

let type_expr ty = typexp false 0 ty
and type_scheme ty = reset_var_names(); typexp true 0 ty

let print_type_expr = type_expr;;
	
let print_constant c =
    match c with
	(Const_int n) -> Pretty.str(string_of_int n)
      | (Const_string s)  -> Pretty.str("\"" ^ String.escaped(s) ^ "\"")
      | (Const_float s) -> Pretty.str(s)
      | (Const_char c) -> Pretty.str("'" ^ (String.make 1 c) ^ "'")

let print_field label_desc =
    Pretty.str ("f" ^ (string_of_int label_desc.lbl_pos))

let rec print_pattern_desc(p) =
    (match p with
        Tpat_var(ident) -> Pretty.str(Ident.name ident)
      | Tpat_any -> Pretty.str "_"
      | Tpat_construct(description, exprs) ->
	Pretty.blo(3,
		   print_constructor_description(description) ::
		   (List.fold_right (fun rand x -> Pretty.brk(1) :: print_pattern(rand) :: x) exprs []))
      | Tpat_tuple(exprs) ->
	Pretty.blo(1, [Pretty.str "("] @ (print_with_delimitors print_pattern [Pretty.str","; Pretty.brk 1] exprs) @ [Pretty.str ")"])
      | Tpat_alias(pat,ident) -> Pretty.blo(1, [Pretty.str "("; print_pattern pat; Pretty.str " as "; Pretty.str (Ident.name ident); Pretty.str ")"])
      | Tpat_constant c -> print_constant c
      | Tpat_or(pat1,pat2) -> Pretty.blo(1, [Pretty.str "("; print_pattern pat1; Pretty.str " | "; print_pattern pat2; Pretty.str ")"])
      | Tpat_record fieldlist -> Pretty.blo(1, [Pretty.str "{"] @ (print_with_delimitors print_field_pattern [Pretty.str"; "; Pretty.brk 1] fieldlist) @ [Pretty.str "}"]))

and print_pattern{pat_desc=pat_desc; pat_loc=pat_loc; pat_type=pat_type} =
    Pretty.blo(1, [Pretty.str "(";
		   print_pattern_desc(pat_desc);
		   Pretty.brk 0;
		   Pretty.str(": ");
		   type_expr(pat_type);
		   Pretty.str ")"])

and print_field_pattern = function (label_desc, pat) ->
     Pretty.blo(0, [print_field label_desc;
                    print_pattern pat])

and print_bindings(bindings) =
    Pretty.blo(0,
	       print_with_delimitors (print_clause' "=") [Pretty.nl] bindings)


and print_clause' sep = print_clause [Pretty.str(sep); Pretty.brk 1]

and print_clause sep (pattern, expression) =
    Pretty.blo(1,
	       [print_pattern(pattern); Pretty.str " "] @ sep @ [print_expression expression])

and print_expression(e) = print_expression2(e)

and print_expression1 x = generic_print_expression print_expression_desc1 x

and print_expression2 x = generic_print_expression print_expression_desc2 x

and generic_print_expression print_expression_desc e =
    match e.exp_desc with
        Texp_ident(path, {val_type=ident_type}) ->
            if equals e.exp_type ident_type then
                print_expression_desc e.exp_desc
            else
	        let name = "PolymorphicInstance(" in
		Pretty.blo(String.length name,
			   [Pretty.str name;
			    type_expr ident_type;
			    Pretty.str ",";
			    Pretty.brk 1;
			    type_expr e.exp_type;
			    Pretty.str ",";
			    Pretty.brk 1;
			    print_expression_desc1 e.exp_desc;
			    Pretty.str ")"])
      | _ -> print_expression_desc e.exp_desc

and print_expression_desc1(e) =
    match e with
      Texp_ident(path, _) -> print_path(path)
    | Texp_constant(c) -> print_constant(c)
    | Texp_tuple(exprs) ->
	Pretty.blo(1, [Pretty.str "("] @ (print_with_delimitors print_expression [Pretty.str","; Pretty.brk 1] exprs) @ [Pretty.str ")"])

    | Texp_array(exprs) ->
	Pretty.blo(1, [Pretty.str "[|"] @ (print_with_delimitors print_expression [Pretty.str";"; Pretty.brk 1] exprs) @ [Pretty.str "|]"])

    | _ ->
      Pretty.blo(0, [Pretty.str "("] @ [print_expression_desc2(e)] @ [Pretty.str ")"])

and print_expression_desc2(e) =
    match e with
      Texp_let(rec_flag, stack, bindings, body) ->
      Pretty.blo(0,
		 [Pretty.str (match stack with Nostack | Anystack -> "let"
                                  | Stack -> "letstack"
                                  | StackTail -> "letstack'");
                  Pretty.str (match rec_flag with Recursive -> " rec "
                                                | _ -> " ");
		  Pretty.blo(3, print_with_delimitors (print_clause' "=") [Pretty.brk 1; Pretty.str"and "] bindings);
		  Pretty.str(" in");
		  Pretty.nl;
		  Pretty.blo(0,
			     [Pretty.str(match body.exp_desc with (Texp_let(_,_,_,_)) -> "" | _ -> " ");
			      print_expression(body)])])
    | Texp_function(patterns) ->
	  Pretty.blo(1,
		     (Pretty.str "(function")::(Pretty.brk 1)::
		     (print_with_delimitors (print_clause [Pretty.str("->"); Pretty.nl])
			                    [Pretty.str(" |"); Pretty.nl] patterns) @
		     [Pretty.str ")"])

     | Texp_apply(rator, rands) ->
	   Pretty.blo(3,
		      print_expression2(rator) :: (List.fold_right (fun rand x -> Pretty.brk(1) :: print_expression1(rand) :: x) rands []))

     | Texp_match(expr, clauses) ->
       Pretty.blo(0,
		   (Pretty.str "match ")::
		   (print_expression expr)::
		   (Pretty.str " with")::
		   Pretty.nl::
		  (print_with_delimitors (print_clause' "->") [Pretty.nl; Pretty.str("| ")] clauses))

     | Texp_try(expr, clauses) ->
       Pretty.blo(0,
		  [Pretty.str "try ";
		   print_expression(expr);
		   Pretty.str " with";
		   Pretty.nl] @
		  (print_with_delimitors (print_clause' "->") [Pretty.nl; Pretty.str("| ")] clauses))

    | Texp_construct(description, exprs) ->
	Pretty.blo(3,
		   print_constructor_description(description) ::
		   (List.fold_right (fun rand x -> Pretty.brk(1) :: print_expression1(rand) :: x) exprs []))

    | Texp_ifthenelse(predicate, consequent, alternative) ->
      Pretty.blo(0,
		 [Pretty.str "(if ";
		  print_expression predicate;
		  Pretty.str " then"; Pretty.nl;
		  Pretty.blo(0, [Pretty.str("   "); print_expression consequent]);
		  Pretty.nl] @
		 (match alternative with
		     None -> []
		   | Some(alternative) ->
		     [Pretty.str(" else"); Pretty.nl;
		      Pretty.blo(0, [Pretty.str("   "); print_expression alternative])]))
    | Texp_sequence(e1, e2) ->
      let rec print_sequence_desc e =
	  match e with
	      Texp_sequence(e1, e2) -> 
	      (print_expression e1) :: Pretty.str(";"):: Pretty.nl :: (print_sequence_desc e2.exp_desc)
	    | _ -> [print_expression_desc2 e]
      in
	  Pretty.blo(0, 
		     [Pretty.str "begin"; Pretty.nl;
		      Pretty.str("   "); Pretty.blo(0, print_sequence_desc(e));
		      Pretty.nl;
		      Pretty.str "end"])

    | Texp_while(e1, e2) ->
      Pretty.blo(0,
		 [Pretty.str "while "; 
		  print_expression e1; Pretty.str(" do"); Pretty.nl;
		  Pretty.blo(0, [Pretty.str("   "); print_expression e2]);
                  Pretty.str " done"])
    | Texp_for(ident, e1, e2, dir, body) ->
      Pretty.blo(0,
		 [Pretty.str "for "; 
		  Pretty.blo(0, [Pretty.str(Ident.name(ident)); Pretty.str(" = "); print_expression e1;
				 Pretty.str " to "; print_expression e2; Pretty.str(" do")]);
		  Pretty.nl;
		  Pretty.blo(0, [Pretty.str("   "); print_expression body]);
                  Pretty.str " done"])
    | Texp_ident(_, _) -> print_expression_desc1(e)
    | Texp_constant(_) -> print_expression_desc1(e)
    | Texp_tuple(_) -> print_expression_desc1(e)
    | Texp_array(_) -> print_expression_desc1(e)
    | Texp_setfield(lhs, fielddescr, rhs) ->
       Pretty.blo(0,
                  [print_expression lhs;
                   Pretty.str ".";
                   print_field fielddescr;
                   Pretty.str "<-";
                   print_expression rhs])
    | Texp_field(expr,desc) ->
       Pretty.blo(0,
                  [print_expression expr;
                   Pretty.str ".";
                   print_field desc])
    | Texp_record(field_descriptions) ->
       Pretty.blo(0, (Pretty.str "{") :: (print_with_delimitors
print_field_expr [Pretty.str "; "] field_descriptions) @ [Pretty.str "}"])
    | Texp_when(cond,expr) ->
       Pretty.blo(0, [Pretty.str " when ";
                      print_expression cond;
                      Pretty.str " -> ";
                      print_expression expr])
    | Texp_send(expr,message) ->
      Pretty.blo(0, [print_expression expr;
                     Pretty.str ("#" ^ message)])
    | Texp_new(object_path) ->
      Pretty.blo(0, [Pretty.str "new ";
                     print_path object_path])
    | Texp_instvar(_, instvar_path) ->
      print_path instvar_path
    | Texp_setinstvar(_, instvar_path, expr) ->
      Pretty.blo(0, [print_path instvar_path;
                     Pretty.str "<-";
                     print_expression expr])
    | Texp_override(_, override_list) ->
      Pretty.blo(0, (Pretty.str "{<") :: (print_with_delimitors
        print_override_expr [Pretty.str "; "] override_list) @ 
        [Pretty.str ">}"])

and print_field_expr (desc,expr) =
     Pretty.blo(0, [print_field desc;
                    Pretty.str " = ";
                    print_expression expr])

and print_override_expr (over_path, expr) =
     Pretty.blo(0, [print_path over_path;
                    Pretty.str " = ";
                    print_expression expr])

and print_constructor_description(c) =
    Pretty.blo(0,
	       [begin
		    match (c.cstr_tag) with
			Cstr_constant(n) -> Pretty.str("Inject")
		      | Cstr_block(n) -> Pretty.str("Inject")
		      | Cstr_exception(p) -> Pretty.str("Exception")
		end;
		Pretty.str("(");
		type_expr(c.cstr_res);
		Pretty.str(",");
		begin
		    match (c.cstr_tag) with
			Cstr_constant(n) -> Pretty.str("C" ^ string_of_int(n))
		      | Cstr_block(n) -> Pretty.str("B" ^ string_of_int(n))
		      | Cstr_exception(p) -> print_path(p)
	        end;
		Pretty.str(")")])

let print_class_field =
   function Cf_inher _ -> Pretty.str "inherit ..." (* TO DO *)
      |     Cf_val(name, _, private_flag, expr_opt) ->
               Pretty.blo(0, [Pretty.str "val "] @ 
               (match private_flag with Private -> [Pretty.str "private "] 
                                      | Public -> []) @
               [Pretty.str name] @
               (match expr_opt with None -> []
                                  | Some expr -> [print_expression expr]))
      |     Cf_meth(name, expr) ->
               Pretty.blo(0, [Pretty.str ("method " ^ name);
                              print_expression expr]) 

let print_class (class_name, class_def) = 
    Pretty.blo(0, (Pretty.str (Ident.name class_name)) ::
       (List.map print_pattern class_def.cl_args) @ [Pretty.str " = "] @
       (List.map print_class_field class_def.cl_field))

let rec revfold f l zero =
    match l with
        [] -> zero
      | (x::xs) -> revfold f xs (f x zero)

let oc = stdout
let col = 128

let rec print_structure (items: Typedtree.structure)  =
    output_string oc "struct\n";
    List.iter (fun item -> print_structure_item item; output_string oc "\n")
         items;
    output_string oc "end\n"

and print_structure_item (s: Typedtree.structure_item) =
  match s with
    Tstr_eval(e) -> Pretty.pr(oc, print_expression(e), col)
  | Tstr_value(recursive, bindings) ->
    print_definition (match recursive with
        Asttypes.Recursive -> "let rec"
      | Asttypes.Nonrecursive -> "let") bindings
  | Tstr_primitive(_,_) -> output_string oc "#<primitive>\n"
  | Tstr_type(_) -> output_string oc "#<type>\n"
  | Tstr_exception(_,_) -> output_string oc "#<exception>\n"
  | Tstr_module(name, body) ->
        output_string oc ("module " ^ Ident.name(name) ^ "=\n");
	print_module_expr body
  | Tstr_modtype(_,_) -> output_string oc "#<modtype>\n"
  | Tstr_open(_) -> output_string oc "#<open ...>\n"
  | Tstr_class(class_list) -> Pretty.pr(oc, Pretty.blo(0, 
        (Pretty.str "class ") :: (print_with_delimitors print_class 
        [Pretty.str " and "] class_list)), col) 

and print_module_expr {mod_desc=mod_desc} = print_module_expr_desc mod_desc

and print_module_expr_desc d =
  match d with
    Tmod_ident(path) -> output_string oc "#<mod_ident>"
  | Tmod_structure(str) -> print_structure str
  | Tmod_functor(name, sign, body) ->
         output_string oc ("functor " ^ Ident.name(name) ^ "(...) =\n");
	 print_module_expr body
  | Tmod_apply(rator, rand, coercion) -> output_string oc "#<Tmod_apply>"
  | Tmod_constraint(module_expr, sign, coercion) ->
         print_module_expr module_expr;
         output_string oc ": #<constraint>"

and print_definition keyword bindings =
    Pretty.pr(oc, Pretty.blo(3, [Pretty.str keyword;
		   Pretty.brk 1;
		   print_bindings bindings;
		   Pretty.nl]), col)

let print(str: Typedtree.structure) = print_structure str

